
import dbWriter
import sys
#from datetime import datetime
from mx.DateTime import *
from logger import Logger
#------------------
import psycopg
#-----------
datetime = DateTime


def quote_identifier(ident):
    return '"' + '""'.join(ident.split('"')) + '"'

type_map = dbWriter.type_map


class Column:
	def __init__(self,name,datatype,width,index):
		self.name=name
		self.quoted_name=quote_identifier(name)
		self.datatype=datatype
		self.width=width
		self.index=index


		
		
class AbstractTable:
	def __init__(self,tablename,db,logger=None):
		self.logger=logger
		self.db=db
		self.primaryKey=[]
		self.quoted_pk=[]
		if not self.__class__.__dict__.has_key("primaryKey"):
			#print "first call for %s " %self.__class__.__name__
			self.__class__.primaryKey=[]
			self.__class__.quoted_pk=[]
			self.__class__.tablename=tablename
			self.__class__.column_map={}
			self.__class__.columns = []
			self.__class__.quoted_tablename=quote_identifier(tablename)
			self._buildColumns()
			self._findoutPK()
			
		self.tablename=tablename
		
		self.columns=[]
		self.column_map = {}
		self.record={}
		self.quoted_tablename=quote_identifier(tablename)
		self.new=0
		
	def _log(self,message):
		self.logger.log(self.__class__.__name__,message)
		
	
	def _buildColumns(self):
		cursor = self.db.cursor()
		print "build cloumns for %s"%self.__class__.quoted_tablename
		cursor.execute("SELECT * FROM %s WHERE 0=1" % self.__class__.quoted_tablename)
		description = cursor.description
		for i in range(len(description)):
			for pgtyp, tabletyp in type_map:
				if pgtyp == description[i][1]:
					col = Column(description[i][0], tabletyp,len(self.__class__.columns),i)
					break
			else:
				if description[i][1] == self.db.geometry_type:
					self.__class__.geometry_column = description[i][0]
					self.__class__.quoted_geo_col =quote_identifier(self.geometry_column)
				
				continue
			self.__class__.columns.append(col)
	
		for col in self.__class__.columns:
		    self.__class__.column_map[col.name]=col
		    self.__class__.column_map[col.index]=col
		
		self.query_stmt = ("SELECT %s from %s" % (", ".join([col.quoted_name for col in self.columns]),self.quoted_tablename))
	
	def _findoutPK(self):
		curs  = self.db.cursor()
		list =["geompoint","geommultipoly","geompoly","geommultilinestring"]
		geomtable = False
		for i in list:
			if self.__class__.tablename == i:
				geomtable = True
		if geomtable==True:
			stmt = "select indexdef from pg_indexes where tablename='geom' and indexdef like 'CREATE UNIQUE INDEX%' and indexname like('PK_%')"
		else:
			stmt = "select indexdef from pg_indexes where tablename='"+self.__class__.tablename+"' and indexdef like 'CREATE UNIQUE INDEX%' and indexname like('PK_%')"
		curs.execute(stmt)
		result =  curs.fetchone()
		
		if result != None:
			index1=result[0].index("(")
			index2=result[0].rindex(")")
			result = result[0][index1+1:index2]
			primaryKeys=[]
			if result.find(",") != -1:
				keys = result.split(",")
				primaryKeys=keys
			else:
				primaryKeys=[result]
				
			
			self.quoted_pk=[]
			self.__class__.quoted_pk=[]
			for i in primaryKeys:
				self.quoted_pk.append( quote_identifier(i.strip()) )
				self.primaryKey.append( quote_identifier(i.strip()) )
				self.__class__.quoted_pk.append( quote_identifier(i.strip()) )
				self.__class__.primaryKey.append( quote_identifier(i.strip()) )
				
		else:
			self._log("NOT Found PK")	
		
	def _existRecord(self,id):	#id must be list of ids in order of primary key components
		curs = self.db.cursor()
		where =""
		for i in range(len(self.__class__.quoted_pk)):
			where += " %s=%d and"%(self.__class__.quoted_pk[i],id[i])
		where= where[:-3]	
		stmt = "select * from %s where %s" %(self.__class__.quoted_tablename,where)
		curs.execute(stmt)
		result = curs.fetchone()
		if result == None:
			return 0
		return 1
	
	def fillFromDB(self):
		
		curs = self.db.cursor()
		where =""
		for i in  range(len(self.__class__.quoted_pk)):
			temp =  " %s = %d and"%(self.__class__.quoted_pk[i],self.pkid[i])
			where += temp
		where= where[:-3]
		stmt = "select * from %s where %s" %(self.__class__.quoted_tablename,where)
		count = 0
		curs.execute(stmt)
		
		result =  curs.fetchone()
		if  result == None :
			self.new=1
			return
		string = ""
		for i in result:
			key = self.__class__.column_map[count].name
			
			type = self.__class__.column_map[count].datatype
			if type=="datetime":
				i = self._parseDatum(i)

			if i == None:
				i = "NULL"			

			self.record[ key ] = i
			type = self.__class__.column_map[count].datatype
			name = self.__class__.dbmap[key]
			
			if self.__dict__[name] != i:			
					self.__dict__[name]=i
					
			count +=1
	
	def _checkDirty(self):
		curs = self.db.cursor()
		where =""
		for i in  range(len(self.__class__.quoted_pk)):
			where += " %s=%d and"%(self.__class__.quoted_pk[i],self.pkid[i])
		where= where[:-3]
		stmt = "select * from %s where %s" %(self.__class__.quoted_tablename,where)
		count = 0
		curs.execute(stmt)
		
		result =  curs.fetchone()
		if  result == None :
			self.new=1
			return
		for i in result:
			if i == None:
				i= "NULL"			
			self.record[ self.__class__.column_map[count].name ] = i
			type = self.__class__.column_map[count].datatype
			if type=="datetime":
				i = self._parseDatum(i)
			
			
			name = self.__class__.dbmap[self.__class__.column_map[count].name]
			
			if self.__dict__[name] != i:
				return 1
			count +=1


	def save(self):
		#all parameters present?
		for i in self.__class__.columns:
			name = self.__class__.dbmap[i.name]
			type = i.datatype
			if self.__dict__[name] == None:
				#param not present
				self._log("Cannot save: %s not all parameters present"%name)
				return 0
		
		
		#is it a new Record ?
		if self._existRecord(self.pkid) == 0:
			return self._insert()
		else:
			if self._checkDirty() == 1:
				return self._update()
			else:
				self._log("nothing to update. Record is up to date")
		
	def _insert(self):
		# build the sqlInsertString
		stmtroot= "insert into %s (%s) values (%s);"
		keys=""
		values=""
		for i in self.__class__.columns:
			internname = self.__class__.dbmap[i.name]
			dbname=i.quoted_name
			type = i.datatype
			keys += "%s,"%dbname 
			if ( (type == "string" )or(type=="geometry") or (type=="datetime") ):
				values += "'%s',"%self.__dict__[internname]
			else:
				values += "%s," %self.__dict__[internname]
		keys = keys[:-1]
		values = values[:-1]
		stmt = stmtroot %(self.__class__.quoted_tablename,keys,values)
		
		curs=self.db.cursor()
		
		try:
			curs.execute(stmt)
		except:
			self.db.rollback()
			self._log("record could not be stored:\n\t%s"%stmt)
			return 0
		self._log("record stored:\n\t%s"%stmt)
		return 1

	def _update(self):
		#build update sql and fire it
		self._log("update: Recordid:%s"%self.pkid)
		
		where =""
		for i in  range(len(self.__class__.quoted_pk)):
			temp =  " %s = %d and"%(self.__class__.quoted_pk[i],self.pkid[i])
			where += temp
		where = where[:-4]
		stmtroot = "update %s set %s where %s"
		pairs=""
		for i in self.__class__columns:
			internname = self.__class__.dbmap[i.name]
			dbname=i.quoted_name
			type = i.datatype
			
			if ( (type == "string" )or(type=="geometry") or (type=="datetime") ):
				pairs += "%s = '%s'," %(dbname,self.__dict__[internname])
			else:
				pairs += "%s = %s ," %(dbname,self.__dict__[internname])
		pairs = pairs[:-1]
		stmt = stmtroot %(self.__class__.quoted_tablename,pairs,where)
		
		curs=self.db.cursor()
		
		try:
			curs.execute(stmt)
		except:
			self.db.rollback()
			self._log("record could not be updated:\n\t%s"%stmt)
			return 0
		self._log("record updated: ")
		return 1
		

	def _parseDatum(self,datum):
		if datum == None:
			return None
		string = str(datum).split()
		datestr=string[0]
		timestr=string[1]
		if datestr.find(".") != -1:
			thedate = datestr.split(".")
		if datestr.find("-") != -1:
			thedate = datestr.split("-")
			
		thetime = timestr.split(":")
		lastelem = len(thetime)-1
		if thetime[lastelem].find("."):
			microsec=thetime[lastelem].split(".")
		list = thedate + thetime[:lastelem]+[microsec[0]]
		y,m,d,h,mm,s=list	
		ret =  datetime(int(y),int(m),int(d),int(h),int(mm),int(s))
		return ret
	
	def getNewID(self,tablename):
		if tablename[:4]=="geom":
			tablename="geom"
		curs = self.db.cursor()
		stmt = "select getnextid('%s')"%tablename
		curs.execute(stmt)
		result = curs.fetchone()[0]
		return result
			 
	
	def __str__(self):
		classname = self.__class__
		string=""
		count=0
		for i in self.columns :
			name = self.__class__.dbmap[self.column_map[count].name]
			
			string += "%s:%s "%(i.name, self.__dict__[name])
			count += 1
		return string

