from Thuban.Model.layer import *

from data_fake import MemoryShapeStore, ShapefileShape_dynamic

from sharedStuffForLoadNSave import *
from shapeSpeicherer import *
from shapeStoreBuilder import PriolShapeStore
from Thuban.Model.save import unify_filename, relative_filename

from utilities import kante, polygon

from psyco.classes import __metaclass__

import os

NumCustomLayer = 0

#================================================================================
#    $Revision: 1.30 $
#    $Source: D:/CVS/cvsrepo/thuban/Extensions/precisionFarming/customLayer.py,v $
#    $Id: customLayer.py,v 1.30 2004/10/27 09:52:17 orahn Exp $
#================================================================================


#================================================================================
#  customLayer und dessen Rollen
#================================================================================

class customLayer(Layer):
    """
    Klasse die lediglich zur Unterscheidung zwischen normalen Layern und Layern mit MemoryShapeStores dient, offeriert
    allerdings auch eine Identifikation eines Layers, die nicht Titel-abhaengig ist und ein Rollen-Schema, das
    allen angeschlossenen Klassen eine rollen-spezifische Behandlung eines customLayer-Objektes erlaubt.
    """
    
    def __init__(self, title, data, typ = None, 
                 projection = None,
                 fill = Transparent,
                 stroke = Black,
                 lineWidth = 1,
                 visible = True):
        Layer.__init__(self, title, data, projection,
                 fill,
                 stroke,
                 lineWidth,
                 visible)

        if ( typ == None ):
            typ = typenMap.getEinfacheShapesTyp()

        global NumCustomLayer
        NumCustomLayer += 1

        self.eindeutigeID = NumCustomLayer
        self.rolle = None
        
        self.istGaussKrueger = False
        
        self._bestimmeRolle( typ )
        
        if ( isinstance( self.ShapeStore(), MemoryShapeStore ) or isinstance( self.ShapeStore(), PriolShapeStore) ):
            # Shapestores werden beim Speichern zuerst 
            # bearbeitet - daher sollte die abfragbar sein 
            # (soll das Ding ueberhaupt gespeichert werden?)
            self.ShapeStore().rolle = self.rolle
            self.ShapeStore().setTitle( title )
            
    def hatVerwertbareProjektion( self ):
        return self.istGaussKrueger
        
    def _bestimmeRolle( self, typ ):
        if ( typ == typenMap.getEinfacheShapesTyp() ):
            print typenMap.getEinfacheShapesTyp()
            self.rolle = customLayerRollen_base()
        elif ( typ == typenMap.getShapefileTyp() ):
            print typenMap.getShapefileTyp()
            self.rolle = shapefile_base_rolle(self.ShapeStore())
        elif ( typ == typenMap.getPirolDB() ):
            print typenMap.getPirolDB()
            self.rolle = pirolDb_rolle()
        elif ( typ == typenMap.getRasterTyp() ):
            print typenMap.getRasterTyp()
            self.rolle = rasterEinteilungs_rolle(self.ShapeStore())
        elif ( typ == typenMap.getIdLayer() ):
            print typenMap.getIdLayer()
            self.rolle = id_rolle()
        elif ( typ == typenMap.getCSVTyp() ):
            print typenMap.getCSVTyp()
            self.rolle = csvLayer_rolle( self.ShapeStore() )
        elif ( typ == typenMap.getProjezierterTyp() ):
            print typenMap.getProjezierterTyp()
            self.rolle = umprojeziertesLayer_rolle( self.ShapeStore() )
        elif ( typ == typenMap.getExRasterTyp() ):
            print typenMap.getExRasterTyp()
            self.rolle = ehemRaster_rolle( self.ShapeStore() )
        else:
            print "nicht registrierter layertyp:", typ
            self.rolle = customLayerRollen_base()

    def write_layer(self, sessionSaver, attrs = None, dir=None):
        self.rolle.write_layer( self, sessionSaver, attrs, dir )

    def getId(self):
        return self.eindeutigeID
        
    def getRollenName( self ):
        return self.rolle.getRollenName()
        
    def getRolle( self ):
        return self.rolle
        
    def enthaeltAusgangsdaten( self ):
        return self.rolle.enthaeltAusgangsdaten()
        
    def istDrehbar( self ):
        return self.rolle.istDrehbar()
        
    def hatDbConnection( self ):
        return self.rolle.hatDbConnection()
        
    def wirdAlsShapefileGespeichert( self ):
        return self.rolle.wirdAlsShapefileGespeichert()
        
    def sollGespeichertWerden( self ):
        return self.rolle.sollGespeichertWerden()
        
    def enthaeltRasterDaten( self ):
        return self.rolle.enthaeltRasterDaten()
        
    def getPolygone( self ):
        if ( self.enthaeltRasterDaten() ):
            return self.rolle.getPolygone(self.ShapeStore())
        else:
            return None
        
class typenMap:
    
    def getIdLayer():
        return "id"
        
    getIdLayer = staticmethod( getIdLayer )
    
    def getPirolDB():
        return "pirolDb"
        
    getPirolDB = staticmethod( getPirolDB )
    
    def getVoronoiTyp():
         return typenMap.getShapefileTyp()
   
    getVoronoiTyp = staticmethod( getVoronoiTyp )
 
    def getDelaunayTyp():
         return typenMap.getShapefileTyp()
   
    getDelaunayTyp = staticmethod( getDelaunayTyp )

    
    def getEinfacheShapesTyp():
         return "simpel"
   
    getEinfacheShapesTyp = staticmethod( getEinfacheShapesTyp )

    def getRasterTyp():
         return "raster"
   
    getRasterTyp = staticmethod( getRasterTyp )
    
    def getShapefileTyp():
        return "shape"

    getShapefileTyp = staticmethod( getShapefileTyp )
    
    def getCSVTyp():
        return "cvs"
        
    getCSVTyp = staticmethod( getCSVTyp )
    
    def getProjezierterTyp():
        return "projeziert"
        
    getProjezierterTyp = staticmethod( getProjezierterTyp )
    
    def getExRasterTyp():
        return "ehemRaster"
        
    getExRasterTyp = staticmethod( getExRasterTyp )
        
class customLayerRollen_base:
    """
    default rolle fuer ein customLayer - definiert Standard-Verhalten
    """
    def __init__( self ):
        self.rollenName = typenMap.getEinfacheShapesTyp()
        
    def getRollenName( self ):
        return self.rollenName
        
    def write_layer( self, layer, sessionSaver, attrs = None, dir=None ):
        if ( not self.sollGespeichertWerden() ):
            return
            
        if attrs is None:
            attrs = {}

        attrs["title"]   = layer.Title()
        attrs["visible"] = ("false", "true")[int(layer.Visible())]

        if isinstance(layer, customLayer):
            attrs["shapestore"]   = kuerzeId( repr( layer.ShapeStore() ) )

            lc = layer.GetClassification()
            attrs["stroke"] = lc.GetDefaultLineColor().hex()
            attrs["stroke_width"] = str(lc.GetDefaultLineWidth())
            attrs["fill"] = lc.GetDefaultFill().hex()
            attrs["rolle"] = self.getRollenName()
            global pirolNamespace
            attrs["xmlns"] = pirolNamespace

            sessionSaver.open_element("customlayer", attrs)
            sessionSaver.write_projection(layer.GetProjection())
            sessionSaver.write_classification(layer)
            sessionSaver.close_element("customlayer")
    
    def enthaeltAusgangsdaten( self ):
        return False
        
    def istDrehbar( self ):
        return False
        
    def hatDbConnection( self ):
        return False
    
    def drehe( self, winkel ):
        if ( not self.istDrehbar() ):
            return False
        else:
            return True

    def sollGespeichertWerden( self ):
        """
        soll das layer gespeichert werden, wenn die session gespeichert wird?
        """
        return True
        
    def wirdAlsShapefileGespeichert( self ):
        # default speicher methode soll das Speichern als XML sein
        return False
        
    def enthaeltRasterDaten( self ):
        return False

class pirolDb_rolle(customLayerRollen_base):
    """
    Layer Rolle, die ein Layer darstellt, das aus der PIROL Datenbank geladen wurde
    """
    def __init__( self ):
        self.rollenName = typenMap.getPirolDB()
        self.dbParam = {}
        self.schlagid=None
        self.tiefenid=None
        self.gruppenid=None

    def enthaeltAusgangsdaten( self ):
        return True
        
    def hatDbConnection( self ):
        return True
        
    def _prepareDict( self ):
        self.dbParam = {"schlagId": repr(self.schlagid),
                        "gruppenId":repr(self.gruppenid),
                        "tiefenId": repr(self.tiefenid) }
                        
    def gibLayerParams( self ):
        self._prepareDict()
        return self.dbParam
        

class id_rolle(customLayerRollen_base):
    """
    Layer Rolle, die ein Layer darstellt, das nur dazu dient Daten z.B. aus einem Histogramm zu identifizieren 
    (welche Punkte gehoeren zu welcher Saeule, Clusterzentren, etc.)
    """
    def __init__( self ):
        self.rollenName = typenMap.getIdLayer()
        
    def sollGespeichertWerden( self ):
        return False

        
class shapefile_base_rolle(customLayerRollen_base):
    """
    basisklasse fuer Daten, die als Shapefile gespeichert werden sollen
    """
    def __init__( self, store ):
        self.rollenName = typenMap.getShapefileTyp()
        self.titel = kuerzeId( repr(store) )
    
    def wirdAlsShapefileGespeichert( self ):
        return True

    def write_layer( self, layer, sessionSaver, attrs = None, dir=None ):
            
        if attrs is None:
            attrs = {}

        attrs["title"]   = layer.Title()
        attrs["visible"] = ("false", "true")[int(layer.Visible())]

        if isinstance(layer, customLayer):
            titel = self.titel + ".shp"
            dirname = sessionSaver.file.name[ :sessionSaver.file.name.rfind(".") ]
            path = unify_filename( dirname )
            
            print "------------------"
            print "Speichere als Shapefile:", dir, ",", titel, ",", path
            
            if ( not os.path.exists(path) ):
                os.mkdir( path )
            
            speicherOnkel = shapeFileSpeicherer( path+"/"+titel, layer )
            speicherOnkel.speichern()
            
            attrs["shapestore"]   = self.titel

            lc = layer.GetClassification()
            attrs["stroke"] = lc.GetDefaultLineColor().hex()
            attrs["stroke_width"] = str(lc.GetDefaultLineWidth())
            attrs["fill"] = lc.GetDefaultFill().hex()

            sessionSaver.open_element("layer", attrs)
            sessionSaver.write_projection(layer.GetProjection())
            sessionSaver.write_classification(layer)
            sessionSaver.close_element("layer")    

class csvLayer_rolle(shapefile_base_rolle):
    def __init__( self, store ):
        shapefile_base_rolle.__init__( self, store )
        self.ausgDaten = True
        if ( store.ShapeType() != SHAPETYPE_POINT ):
            self.ausgDaten = False
    
    def enthaeltAusgangsdaten( self ):
        return self.ausgDaten

class umprojeziertesLayer_rolle( csvLayer_rolle ):
    def __init__( self, store ):
        csvLayer_rolle.__init__( self, store )

class ehemRaster_rolle( csvLayer_rolle ):
    def __init__( self, store ):
        csvLayer_rolle.__init__( self, store )

        
class rasterEinteilungs_rolle(shapefile_base_rolle):
    """
    Layer Rolle, die ein Layer darstellt, das eine Einteilung der Flaeche in quadratische Teilflaechen und 
    entsprechende Verschneidung der Punkte und Zuordnungen von Werten erlaubt.
    """
    def istDrehbar( self ):
        return False
        
    def enthaeltRasterDaten( self ):
        return True
        
    def getPolygone( self, store ):
        polygone = []
        letzesTupel = None
        erstesTupel = None
        for shp in store.AllShapes():
            letzesTupel = None
            erstesTupel = None
            kanten = []
            for tpl in shp.Points()[0]:
                if ( letzesTupel != None ):
                    x1, y1 = letzesTupel
                    x2, y2 = tpl
                    kanten.append( kante( x1,y1,x2,y2, -1,-1 ) )
                else:
                    erstesTupel = tpl
                letzesTupel = tpl
            # Polygon schliessen:
#            x1, y1 = erstesTupel
#            x2, y2 = letzesTupel
#            kanten.append( kante( x1,y1,x2,y2, -1,-1 ) )
            
            nPoly = polygon(kanten)
            nPoly.setIndex( shp.ShapeID() )
            polygone.append( nPoly )
                
        return polygone
        
        
