import os
import shutil

from Thuban.Model.save import SessionSaver, save_session, sort_data_stores
from Thuban.Model.load import SessionLoader, load_session, parse_color

from Thuban.Model.xmlreader import XMLReader
import xml.sax
import xml.sax.handler

from customLayer import customLayer, typenMap
#from Thuban.Model.layer import RasterLayer
from sharedStuffForLoadNSave import *

from Thuban.Model.transientdb import AutoTransientTable, TransientJoinedTable
from Thuban.Model.table import DBFTable, MemoryTable, FIELDTYPE_STRING
from Thuban.Model.data import DerivedShapeStore, ShapefileStore
from Thuban.Model.xmlwriter import XMLWriter
from Thuban.Model.postgisdb import PostGISConnection, PostGISShapeStore

from data_fake import MemoryShapeStore, ShapefileShape_dynamic

# Priol DB kram
from shapeStoreBuilder import *
import dbWriter
from dbFactory import *
# ---------------------

# schleifen beschleunigen
try:
    from psyco.classes import __metaclass__
except:
    pass

from string2nestedlist import *

pirolNamespace = "http://www.pirol.fh-osnabrueck.de/pirol-0.1.dtd"

#================================================================================
#    $Revision: 1.9 $
#    $Source: D:/CVS/cvsrepo/thuban/Extensions/precisionFarming/customSaveAndLoad.py,v $
#    $Id: customSaveAndLoad.py,v 1.9 2004/11/04 10:02:17 orahn Exp $
#================================================================================


class customSessionLoader( SessionLoader ):
    def __init__(self, db_connection_callback = None):
        SessionLoader.__init__(self, db_connection_callback)
        additionalDispatchers = {
            'pirolshapesource'    : ("start_pirolshapesource", None),
            'memshapestore'       : ("start_memshapestore", "end_memshapestore"),
            'customlayer'         : ("start_customlayer",          "end_customlayer"),
            'columns'             : ("start_columns",          "end_columns"),
            'spalte'              : ("start_spalte",          None),
            'memdata'             : ("start_content",          None),
            'memtable'            : ("start_memtable",          "end_memtable"),
            'shapedata'           : ("start_shapedata",          "end_shapedata"),
            'shape'               : ("start_shape",          None)            
        }
        
        global pirolNamespace
        xmlns = pirolNamespace
        for key, value in additionalDispatchers.items():
            additionalDispatchers[(xmlns, key)] = value

        XMLReader.AddDispatchers(self, additionalDispatchers)
        
        self.aStore = None
        
        self.aColSet = []
        self.aDataSet = []
        self.aShapeSet = []
        
    def skippedEntity( self, name):
        print "skipped: ", name


    def start_pirolshapesource(self, name, qname, attrs):
        print "start_pirolshapesource()"
        
        
        dbfact = None
        
        connections = self.theSession.DBConnections()
        if len (connections) == 0:
             print('Fuegen sie die Pirol - Datenbankverbindung hinzu')
             return False
        
        db = self.encode(attrs.get((None, 'dbconn'), None))
        conn = self.idmap[db]
        
        writer = dbWriter.dbWriter.getDbWriter(conn)
        dbfact = DbFactory.getDbFactory( writer )
        
        ID = self.encode(attrs.get((None, 'id'), "none"))
        schlagId = int(self.encode(attrs.get((None, 'schlagId'), None)))
        gruppenId = int(self.encode(attrs.get((None, 'gruppenId'), None)))
        
        parser = string2nestedlist( self.encode(attrs.get((None, 'tiefenId'), None)) )
        geparst = parser.parseString2nestedList()
        tiefenId = geparst

        builder = ShapeStoreBuilder(dbfact,None,schlagid=schlagId,gruppenid=gruppenId,tiefen=tiefenId )
        self.aStore = builder.getShapeStore()
        
        self.idmap[ID] = self.aStore
        self.theSession.AddShapeStore( self.aStore )
        
        self.aColSet = []
        self.aDataSet = []
        self.aStore = None

        
    def startElementNS( self, name, qname, attrs):
        XMLReader.startElementNS( self, name, qname, attrs)
    
    def start_shape(self, name, qname, attrs):
        pointsString = self.encode(attrs.get((None, 'points'), "[]"))
        index = int( attrs.get((None, 'id'), "-1") )
        parser = string2nestedlist( pointsString )
        geparst = parser.parseString2nestedList()
        
        for shape in geparst:
            xs = []
            ys = []
            for punkt in shape:
                x, y = punkt
                xs.append(x)
                ys.append(y)

            self.aStore.addShape( xs, ys, index )
        
    
    def start_shapedata(self, name, qname, attrs):
        pass

    def end_shapedata(self, name, qname):
        pass
        
    def start_memtable(self, name, qname, attrs):
        print "start_memtable"
        
    def end_memtable(self, name, qname ):
        self.aStore.setTable( MemoryTable(self.aColSet,self.aDataSet) )
        
    def start_content(self, name, qname, attrs):
        tabellenString = self.encode(attrs.get((None, 'content'), "[]"))
        parser = string2nestedlist( tabellenString )
        geparst = parser.parseString2nestedList()
        self.aDataSet = geparst
    
    def start_spalte(self, name, qname, attrs):
        index = int(attrs.get((None, 'index'), "0"))
        type = self.encode(attrs.get((None, 'type'), "string"))
        name = self.encode(attrs.get((None, 'name'), "spalte"))
        
        self.aColSet.append( ( name, type ) )
        
    def start_columns(self, name, qname, attrs):
        pass
        
    def end_columns(self, name, qname ):
        pass
        
    def start_memshapestore(self, name, qname, attrs):
        print "start_memshapestore()"
        
        ID = self.encode(attrs.get((None, 'id'), "none"))

        self.aStore = MemoryShapeStore()
        self.aStore.shapetype = self.encode(attrs.get((None, 'shapetyp'), "arc"))
        
        parser = string2nestedlist( self.encode(attrs.get((None, 'bbox'), "")) )
        geparst = parser.parseString2nestedList()
        self.aStore.bbox = geparst
        
        self.idmap[ID] = self.aStore
        self.theSession.AddShapeStore( self.aStore )

    def end_memshapestore(self, name, qname):
        self.aColSet = []
        self.aDataSet = []
        self.aStore = None
 
    def start_customlayer(self, name, qname, attrs, layer_class = customLayer):
        """Start a layer

        Instantiate a layer of class layer_class from the attributes in
        attrs which may be a dictionary as well as the normal SAX attrs
        object and bind it to self.aLayer.
        """
        title = self.encode(attrs.get((None, 'title'), ""))

        visible  = self.encode(attrs.get((None, 'visible'), "true")) != "false"
        fill = parse_color(attrs.get((None, 'fill'), "None"))
        stroke = parse_color(attrs.get((None, 'stroke'), "#000000"))
        stroke_width = int(attrs.get((None, 'stroke_width'), "1"))
        typus = self.encode(attrs.get((None, 'rolle'), typenMap.getEinfacheShapesTyp()))
        if attrs.has_key((None, "shapestore")) and self.idmap.has_key(attrs.get((None, "shapestore"),"")):
            store = self.idmap[attrs.get((None, "shapestore"),"")]
        else:
            if attrs.has_key((None, "shapestore")):
                print "konnte shapeStore <", attrs.get((None, "shapestore"),""), "> nicht finden!"
                print self.idmap
            else:
                print "konnte shapeStore fuer layer <", title, "> nicht finden!"
            store = MemoryShapeStore()
            
        self.aLayer = layer_class(title, store,
                                  typ = typus,
                                  fill = fill, stroke = stroke,
                                  lineWidth = stroke_width,
                                  visible = visible)

    def end_customlayer(self, name, qname):
        self.aMap.AddLayer(self.aLayer)
        self.aLayer = None


## ----------------

class customSessionSaver(SessionSaver):
    """
    Klasse die im Grunde alles genauso macht wie die SessionSaver Klasse, aber bestimmte Layer beim Speichern ueberspringt...
    """
    def __init__(self, session):
        SessionSaver.__init__(self, session)

    def write_layer(self, layer, attrs = None):
        if isinstance(layer, customLayer):
            layer.write_layer( self, attrs, dir=self.dir )
            return
        else:
            SessionSaver.write_layer(self, layer, attrs)

    def write_data_containers(self, session):
        global pirolNamespace
        containers = sort_data_stores(session.DataContainers())
        for container in containers:
            if isinstance(container, AutoTransientTable):
                # AutoTransientTable instances are invisible in the
                # thuban files. They're only used internally. To make
                # sure that containers depending on AutoTransientTable
                # instances refer to the right real containers we give
                # the AutoTransientTable instances the same id as the
                # source they depend on.
                self.define_id(container,
                               self.get_id(container.Dependencies()[0]))
                continue

            idvalue = self.define_id(container)
            if isinstance(container, ShapefileStore):
#================================================================================
#                self.define_id(container.Table(), idvalue)
#                filename = self.prepare_filename(container.FileName())
#                self.write_element("fileshapesource",
#                                   {"id": idvalue, "filename": filename,
#                                    "filetype": "shapefile"})
#================================================================================
                                    
                 self.define_id(container.Table(), idvalue)
                 
                 dir, bareFileName = os.path.split( container.FileName() )
                 
                 #self.prepare_filename( bareFileName )               
                 
                 #fileNamePointless = self.prepare_filename( container.FileName() )
                 fileNamePointless = container.FileName()
                 if ( fileNamePointless.endswith(".shp") > -1 ):
                     print "vorher: ", fileNamePointless
                     fileNamePointless, ext1 = os.path.splitext( fileNamePointless )
                     print "nachher: ", fileNamePointless
                 

                 if ( bareFileName.endswith(".thuban") > -1 ):
                     bareFileName, shpExt = os.path.splitext( bareFileName )
 
                     
                 dirname, sessionExt = os.path.splitext( self.file.name )
                 sessionDir = os.path.split( self.file.name )

                 print  "dirs:", sessionDir, ",", os.getcwd()
                 
                 newFileName = os.path.join( dirname,bareFileName )
                 
                 print "----------------------"
                 print "kopiere Datei <", container.FileName(), "> nach <", newFileName+".shp", "> ..."
                 if ( not os.path.exists(dirname) ):
                     print "- lege Verzeichnis an:", dirname 
                     os.mkdir( dirname )
                 
                 print "- pruefe, ob Dateien mit gleichen Namen bereits existieren..."
                 if ( os.path.exists( newFileName + ".shp" )
                         or os.path.exists( newFileName + ".shx" )
                         or os.path.exists( newFileName + ".dbf" )):
                     nummer = 1
                     anhang = "("+str(nummer)+")"
                     while ( os.path.exists( newFileName + anhang + ".shp" )
                             or os.path.exists( newFileName + anhang + ".shx" )
                             or os.path.exists( newFileName + anhang + ".dbf" )):
                         nummer += 1
                         anhang = "("+str(nummer)+")"
                     newFileName = newFileName + anhang
                     print "kopiere Datei <", container.FileName(), "> nach <", newFileName+".shp", "> ..."
                     
                 shutil.copy( fileNamePointless + ".shp", newFileName + ".shp" )
                 shutil.copy( fileNamePointless + ".shx", newFileName + ".shx" )
                 shutil.copy( fileNamePointless + ".dbf", newFileName + ".dbf" )
                 print "fertig."
                 
                 filename = self.prepare_filename(newFileName + ".shp")
                 self.write_element("fileshapesource",
                                   {"id": idvalue, "filename": filename,
                                    "filetype": "shapefile"})
                                    
            elif isinstance(container, DerivedShapeStore):
                shapesource, table = container.Dependencies()
                self.write_element("derivedshapesource",
                                   {"id": idvalue,
                                    "shapesource": self.get_id(shapesource),
                                    "table": self.get_id(table)})
            elif isinstance(container, PostGISShapeStore):
                conn = container.DBConnection()
                self.write_element("dbshapesource",
                                   {"id": idvalue,
                                    "dbconn": self.get_id(conn),
                                    "tablename": container.TableName()})
            elif isinstance(container, DBFTable):
                filename = self.prepare_filename(container.FileName())
                self.write_element("filetable",
                                   {"id": idvalue,
                                    "title": container.Title(),
                                    "filename": filename,
                                    "filetype": "DBF"})
            elif isinstance(container, TransientJoinedTable):
                left, right = container.Dependencies()
                left_field = container.left_field
                right_field = container.right_field
                self.write_element("jointable",
                                   {"id": idvalue,
                                    "title": container.Title(),
                                    "right": self.get_id(right),
                                    "rightcolumn": right_field,
                                    "left": self.get_id(left),
                                    "leftcolumn": left_field,
                                    "jointype": container.JoinType()})
            elif isinstance(container, MemoryShapeStore):
                print "----------------------"
                print "write_data_containers: MemoryShapeStore"
                if ( not container.rolle.sollGespeichertWerden() ):
                    print "wird nicht gespeichert:", container.rolle.getRollenName()
                    break
                elif ( container.rolle.wirdAlsShapefileGespeichert() ):
                    contId = kuerzeId( repr(container) )
                    
                    dirname = self.file.name[ :self.file.name.rfind(".") ]
                    filename = self.prepare_filename(dirname+"/"+contId+".shp")
                    self.write_element("fileshapesource",
                                   {"id": contId, "filename": filename,
                                    "filetype": "shapefile"})
                else:
                    attrs = {}
                    contId = kuerzeId( repr(container) )
                    attrs["id"] = contId
                    attrs["rolle"] = container.rolle.getRollenName()
                    attrs["shapetyp"] = str( container.ShapeType() )
                    attrs["bbox"] = repr( container.BoundingBox() )
                    
                    attrs["xmlns"] = pirolNamespace
    # -----------------------
                    self.open_element("memshapestore", attrs)
    #    ----------------------- Tabelle
                    memtabattrs = {}
                    memtabattrs["xmlns"] = pirolNamespace
                    self.open_element("memtable", memtabattrs)
    
    #        ----------------------- TabellenSpalten
                    colAttr = {}
                    colAttr["xmlns"] = pirolNamespace
                    self.open_element("columns",colAttr)
                    tabelle = container.Table()
                    spalten = tabelle.Columns()
                    nummer = 0
                    
                    for spalte in spalten:
                        spaltenAttr = {}
                        spaltenAttr["xmlns"] = pirolNamespace
                        spaltenAttr["name"] = spalte.name
                        spaltenAttr["type"] = str( spalte.type )
                        spaltenAttr["index"] = str( spalte.index )
                        self.write_element("spalte", spaltenAttr)
                        
                    self.close_element("columns")
    #        -----------------------
    
    #        ----------------------- TabellenDaten (Zeilen)
                    dataAttr = {}
                    tabellenString = self.encode( repr(tabelle.data) )
                    dataAttr["content"] = tabellenString
                    self.write_element("memdata", dataAttr)
    #        -----------------------
                    self.close_element("memtable")
    #    -----------------------
    
    #    ----------------------- Shapes
                    self.open_element("shapedata")
                    
                    for shape in container.AllShapes():
                        shapeAttrs = {}
                        shapeAttrs["id"] = str( shape.ShapeID() )
                        shapeAttrs["points"] = repr( shape.Points() )
                        self.write_element("shape", shapeAttrs)
                    
                    self.close_element("shapedata")
    #    -----------------------
    
                    self.close_element("memshapestore")
# -----------------------
            elif isinstance(container, PriolShapeStore):
                print "----------------------"
                print "write_data_containers:", container.rolle.getRollenName()
                if ( container.rolle.sollGespeichertWerden() ):
                    conn = container.DBConnection()
                    contId = kuerzeId( repr(container) )
                    pirolParams = container.rolle.gibLayerParams()
                    
                    print "attrs: ", self.get_id(conn), conn, contId, pirolParams
                    
                    self.write_element("pirolshapesource",
                                   {"xmlns": pirolNamespace,
                                   "id": contId, 
                                   "dbconn": self.get_id(conn),
                                   "schlagId": pirolParams["schlagId"],
                                   "gruppenId": pirolParams["gruppenId"],
                                   "tiefenId": pirolParams["tiefenId"]})

            else:
                print "unbekannter Kontainer ", isinstance(container)
                raise ValueError("Can't handle container %r" % container)
    

