# Author: Ole Rahn
# Project: PIROL

#    __version__ = '$Revision: 1.44 $'
#    $Source: D:/CVS/cvsrepo/thuban/Extensions/precisionFarming/data_fake.py,v $
#    $Id: data_fake.py,v 1.44 2004/11/04 14:03:55 orahn Exp $

from __future__ import generators

from Thuban.Model.data import *
from Thuban.Model.table import *
from Thuban.Model.label import Label

# schleifen beschleunigen
try:
    from psyco.classes import __metaclass__
except:
    pass

from utilities import punkt

class ShapefileShape_extension_interface:

    """
    Repraesentation eines Shapes, das dynamisch - also unabhaengig von z.B. einem Shapefile - angelegt werden kann
    """

    def __init__(self, shapeid):
        self.shapeid = shapeid

    def compute_bbox(self):
        """
        Return the bounding box of the shape as a tuple (minx,miny,maxx,maxy)
        """
        return ()

    def ShapeID(self):
        """
        Rueckgabe: die ID (int) des Shapes, das durch das jeweilige Objekt dargestellt wird.
        """
        return self.shapeid

    def Points(self):
        """
        Return the coordinates of the shape as a list of lists of pairs.
        Z.B. [[(x0,y0),(x1,y1)]]
        """
        return []

    def RawData(self):
        """
        Return the shape id to use with the shapefile - wichtig fuer den Datentypus RAW_SHAPEFILE, wir nutzen
        allerdings RAW_PYTHON
        """
        return self.shapeid

    def Shapefile(self):
        """
        Return the shapefile object - aber wir haben kein Shapefile, daher: None
        """
        return None
        


class ShapefileShape_dynamic(ShapefileShape_extension_interface):

    """
    Jedes Objekt dieser Klasse repraesentiert ein dynamisch generiertes Shape
    """

    def __init__(self, x, y, shapeid):
        ShapefileShape_extension_interface.__init__(self,shapeid)
        self.minX = None
        self.maxX = None
        self.minY = None
        self.maxY = None
        self.pointsList = []
        for i in range(len(x)):
            if self.minX == None or x[i] < self.minX:
                self.minX = x[i]
            if self.maxX == None or x[i] > self.maxX:
                self.maxX = x[i]
            if self.minY == None or y[i] < self.minY:
                self.minY = y[i]
            if self.maxY == None or y[i] > self.maxY:
                self.maxY = y[i]
            self.pointsList.append( ( x[i], y[i] ) )

    def SetShapeID( self, id ):
        self.shapeid = id

    def compute_bbox(self):
        """
        Return the bounding box of the shape as a tuple (minx,miny,maxx,maxy)
        """
        return (self.minX, self.minY, self.maxX, self.maxY)

    def Points(self):
        """Return the coordinates of the shape as a list of lists of pairs"""
        return [self.pointsList]
        
    def SetPoints( self, newPoints ):
        self.pointsList = newPoints[0]
        
        self.minX = None
        self.maxX = None
        self.minY = None
        self.maxY = None
        for tupel in newPoints[0]:
            x, y = tupel
            if self.minX == None or x < self.minX:
                self.minX = x
            if self.maxX == None or x > self.maxX:
                self.maxX = x
            if self.minY == None or y < self.minY:
                self.minY = y
            if self.maxY == None or y > self.maxY:
                self.maxY = y
        

    def isInRegion( self, bbox ):
        ThisMinX, ThisMinY, ThisMaxX, ThisMaxY = bbox
        minX, minY, maxX, maxY = self.compute_bbox()

        if minX >= ThisMinX and minX <= ThisMaxX and minY >= ThisMinY and minY <= ThisMaxY:
            return True
        if maxX >= ThisMinX and maxX <= ThisMaxX and maxY >= ThisMinY and maxY <= ThisMaxY:
            return True

        isInRegion = False

        for i in self.pointsList:
            x, y = i
            if (x >= minX and x <= maxX) and (y >= minY and y <= maxY):
                isInRegion = True
                break

        return isInRegion

    def __str__(self):
        return "shape "+repr(self.pointsList)


class MemoryShapeStore_Interface:

    """Combine a shapefile and the corresponding DBF file into one object"""

    def __init__(self):
        self.numshapes = 0
        self.table = None
        self.shapetype = "undefiniert"
        self.rolle = None
        self.bbox = None
        
    def setTitle( self, titel ):
        if ( self.table != None ):
            self.table.SetTitle( titel )
            
    def Title( self ):
        return self.table.Title()

    def _open_shapefile(self):
        print "_open_shapefile is not implemented"

    def Table(self):
        """Return the table containing the attribute data"""
        return self.table

    def setTable( self, Table ):
        self.table = Table
        
    def SetTable( self, table ):
        """
        gepfuscht: ich schreib das immer gross...
        """
        self.setTable( table )

    def Shapefile(self):
        """Return the shapefile object"""
        return None

    def FileName(self):
        """Return the filename used to open the shapefile"""
        return None

    def FileType(self):
        """Return the filetype. This is always the string 'shapefile'"""
        return "shapefile"

    def ShapeType(self):
        """Return the type of the shapes in the shapestore.

        This is either SHAPETYPE_POINT, SHAPETYPE_ARC or SHAPETYPE_POLYGON.
        """
        return self.shapetype

    def RawShapeFormat(self):
        """
        Return the raw data format of the shape data, i.e. RAW_SHAPEFILE
        RAW_PYTHON ist wichtig, um dynamisch generierte Shapes auch rendern zu koennen!
        """
        return RAW_PYTHON

    def NumShapes(self):
        """Return the number of shapes in the shape store"""
        return self.numshapes

    def Dependencies(self):
        """
        Return the empty tuple.

        The ShapefileStore doesn't depend on anything else.
        """
        return ()

    def OrigShapeStore(self):
        """Return None.

        The ShapefileStore was not derived from another shapestore.
        """
        return None

    def BoundingBox(self):
        """Return the bounding box of the shapes in the shapestore.

        The coordinate system used is whatever was used in the shapefile.
        If the shapefile is empty, return None.
        """
        return None

    def ShapesInRegion(self, bbox):
        """Return an iterable over the shapes that overlap the bounding box.

        The bbox parameter should be the bounding box as a tuple in the
        form (minx, miny, maxx, maxy) in the coordinate system of the
        shape store.
        """
        return None


    def AllShapes(self):
        """Return an iterable over the shapes in the shape store."""
        return None

    def Shape(self, index):
        """Return the shape with index index"""
        return []

class MemoryShapeStore(MemoryShapeStore_Interface):

    """shapeFileStore that works without a shapefile or a dbf file"""

    def __init__(self):
        MemoryShapeStore_Interface.__init__(self)
        self.shapesList = {}
        self.removed = 0
        self.maxId = -1
        
    def RemoveShape( self, shape ):
        index = shape.ShapeID()
        if ( shape in self.shapesList.values() ):
            if ( isinstance( self.Table(), MemoryTable ) ):
                if ( self.Table().removeRow( index ) ):
                    #print "RemoveShape():", repr(index), self.shapesList[ index ].ShapeID()
                    self.removed += 1
                else:
                    print "1) kann Attributdaten des Shapes nicht loeschen;", isinstance( self.Table(), MemoryTable )
                    return
            else:
                print "2) kann Attributdaten des Shapes nicht loeschen;", isinstance( self.Table(), MemoryTable )
                return
            #self.shapesList.remove( shape )
            self.shapesList[ index ] = None
            self.numshapes = self.NumShapes()
        
    def getMemCopyOfShapeStore( store ):
        if (store == None):
            return None
            
        origTable = store.Table()
        origColumns = origTable.Columns()
        newStore = MemoryShapeStore()
        
        newStore.bbox = None
        
        newCols = []
        numCols = 0
        for column in origColumns:
            newCols.append( ( column.name, column.type ) )
            
            if ( numCols != column.index ):
                print "WARNUNG: Moegliche Attribut-Datenkorruption!"
                
            numCols += 1
        
        newTable = MemoryTable( newCols, [] )
        newTable.SetTitle( origTable.Title() )
        rowDict = None
        
        newStore.shapetype = store.ShapeType()
        
        for shape in store.AllShapes():
            newRow = []
            rowDict = origTable.ReadRowAsDict( shape.ShapeID() )
            points = shape.Points()
            xs = []
            ys = []
            
            for point in points[0]:
                x,y = point
                xs.append(x)
                ys.append(y)
                
            #newStore.addShape( xs, ys, shape.ShapeID() )
            newStore.addShape( xs, ys )
            
            for column in newTable.Columns():
                newRow.append( rowDict[column.name] )
                
            newTable.appendRow( newRow )
            
        newStore.setTable( newTable )
        
        return newStore

    getMemCopyOfShapeStore = staticmethod( getMemCopyOfShapeStore )

    def addPoint(self, x, y):
        self.shapetype = SHAPETYPE_POINT
        x_ = [x]
        y_ = [y]
        return self.addShape(x_,y_)

    def addPolygon(self, x, y):
        self.shapetype = SHAPETYPE_POLYGON
        return self.addShape(x,y)

    def addLine(self, x, y):
        self.shapetype = SHAPETYPE_ARC
        return self.addShape(x,y)

    def addShape(self, x_, y_, index = -1):
        newIndex = self.maxId + 1
        self.maxId = newIndex
        
        # experimentell
#        self.maxId = max(newIndex, index)
#        if ( index > -1 ):
#            newIndex = index
        
        newPoint = ShapefileShape_dynamic( x_, y_, newIndex )
        self.shapesList[newIndex] = newPoint
        self.numshapes = self.NumShapes()
        return newIndex

    def appendAttributeData( self, data ):
        self.table.appendRow( data )
        
    def appendAttributeDataColumn( self, name, type, defaultValue ):
        index = self.table.NumColumns()
        col = MemoryColumn( name, type, index )
        
        self.table.columns.append(col)
        self.table.column_map[name] = col
        self.table.column_map[index] = col
        
        for row in self.table.data:
            row.append( defaultValue )
            
    def setAttributeValue( self, shapeId, colStr, value ):
        col = self.Table().Column(colStr)
        self.Table().data[self.Table().RowIdToOrdinal(shapeId)][col.index] = value

    def calculateBbox( self ):
        minx, miny, maxx, maxy = (None, None, None, None)
        if ( self.NumShapes()>0 ):
            for shape in self.shapesList.keys():
                if (self.shapesList[shape]==None):
                    continue 
                mnx, mny, mxx, mxy = self.shapesList[shape].compute_bbox()
                
                if ( minx == None or mnx < minx ):
                    minx = mnx
                if ( miny == None or mny < miny ):
                    miny = mny
                if ( maxx == None or mxx > maxx ):
                    maxx = mxx
                if ( maxy == None or mxy > maxy ):
                    maxy = mxy
        
            self.bbox = ( minx, miny, maxx, maxy )
        else:
            self.bbox = None

    def BoundingBox(self):
        """Return the bounding box of the shapes in the shapestore.
        """
        if (self.bbox == None):
            self.calculateBbox()

        return self.bbox

    def NumShapes( self ):
        return len( self.shapesList.keys() ) - self.removed

    def ShapesInRegion(self, bbox):
        """Return an iterable over the shapes that overlap the bounding box.

        The bbox parameter should be the bounding box as a tuple in the
        form (minx, miny, maxx, maxy) in the coordinate system of the
        shape store.
        """
        for i in self.shapesList.keys():
            if ( self.shapesList[i] != None ):
                if self.shapesList[i].isInRegion( bbox ):
                    yield (self.shapesList[i])


    def AllShapes(self):
        """Return an iterable over the shapes in the shape store."""
        for i in self.shapesList.keys():
            if ( self.shapesList[i] != None ):
                yield (self.shapesList[i])

    def Shape(self, index):
        """Return the shape with index index"""
#        for shp in self.shapesList:
#            if ( shp.ShapeID() == index ):
#                return shp
        if ( self.shapesList.has_key( index ) ):
            #print "Shape(", repr(index), "); --->",repr(self.shapesList[ index ].ShapeID())
            return self.shapesList[ index ]

        print "MemoryShapeStore.Shape(): gesuchter index nicht vorhanden!"
        return None
