import math
from feldsortierer import *
from utilities import *

#import threading
import sys

from diagramm_oberklasse import *

# schleifen beschleunigen
try:
    from psyco.classes import __metaclass__
except:
    pass


class delaunay(diagramm):

    def __init__(self):
        """
        sortier_Kriterium - wenn 0 wird nach x sortiert, wenn 1 nach y
        """
        diagramm.__init__(self)

        self.minimalerAbstand = sys.maxint
        self.punkt1 = None
        self.punkt2 = None
        self.sortiert = False
        self.sortiertNach = x_komponente
        self.dreieckZaehler = 0

        self.geschaetzteKantenMenge = 0
        self.kantenErstellt = 0

        pi = math.pi
        self.umrechnungsFaktor = 1.0/(2.0*pi)*360.0
        self.grenzWinkel = pi/18.0
        self.standardGrenzWinkel = self.grenzWinkel

        self.sortier_Kriterium = x_komponente

    def getProzentFertig( self ):
        if self.fehlerIstAufgetreten() or self.berechnungAbgeschlossen == True:
            return 100
        prozent = int( round(100.0 * self.kantenErstellt / self.geschaetzteKantenMenge))
        return prozent

    def start(self ):
        self.erstelleDelaunayNetz()
        return 0

    def fuelleDiagrammMitDatenbankDaten( self, kantenfeld ):
        diagramm.fuelleDiagrammMitDatenbankDaten( self, kantenfeld )


        for i in self.kantenfeld:
            # endpunkte der kanten in korrekt gefuellte delaunayPunkte ueberfuehren
            if not isinstance(i.anfangspunkt, delaunayPunkt):
                tempPunkt1 = i.anfangspunkt
                i.anfangspunkt = delaunayPunkt( tempPunkt1.getKoordinaten(), tempPunkt1.index )
                i.anfangspunkt.normiert = tempPunkt1.normiert
                i.anfangspunkt.multipliziert = tempPunkt1.multipliziert
            #i.anfangspunkt.verbindenMit( i.datenPunktLinks )
            #print "Verbunden %s und %s " %(i.anfangspunkt,i.endpunkt.index)
            if not isinstance(i.endpunkt, delaunayPunkt):
                tempPunkt1 = i.endpunkt
                i.endpunkt = delaunayPunkt( tempPunkt1.getKoordinaten(), tempPunkt1.index )
                i.endpunkt.normiert = tempPunkt1.normiert
                i.endpunkt.multipliziert = tempPunkt1.multipliziert
            #i.endpunkt.verbindenMit( i.datenPunktRechts )

            self._findePunktInPunktefeld(i.datenPunktLinks).verbindenMit( i.datenPunktRechts )
            self._findePunktInPunktefeld(i.datenPunktRechts).verbindenMit( i.datenPunktLinks )
            #print "Verbunden %s und %s " %(i.endpunkt,i.anfangspunkt.index)
        self.sortiert = True
        self.sortiertNach = x_komponente

    def _findePunktInPunktefeld( self, shapeId ):
        for punkt in self.punktefeld:
            if ( punkt.index == shapeId ):
                return punkt
        return None


    def erstelleDelaunayNetz(self):
        if self.punktefeld == []:
            self.fehler = fehler("Es wurden keine Punkte bzw. kein Ziel fuer erstellte Linien gefunden")
            self.berechnungAbgeschlossen = True
            return
        else:
            print "- erstelle Delaunay Netz fuer", len(self.punktefeld), "punkte"
            print "- finde Punktepaar mit kleinstem Abstand"
            self.finde_naechsten_Nachbarn()
            # das punktefeld ist nun nach x koordinaten geordnet
            # rendern der ersten Basislinie
            self.kantenfeld.append( kante(self.punkt1.x(), self.punkt1.y(), self.punkt2.x(), self.punkt2.y(), self.punkt1.index, self.punkt2.index) )
            self.kantenErstellt += 1

            ind1 = self.loesePunktInFeldIndexAuf( self.punkt1 )
            ind2 = self.loesePunktInFeldIndexAuf( self.punkt2 )

            self.punktefeld[ ind1 ].schonVerbundenMit.append( self.punkt2.index )
            self.punktefeld[ ind2 ].schonVerbundenMit.append( self.punkt1.index )

            print "- starte Algorithmus"
            self.errechneDiagramm_Iterativ( ind1, ind2, -1, 0 )
            self.berechnungAbgeschlossen = True

    def berechneNormalenFaktor( self, punkt1, punkt2, punkt3 ):
        #           (C)
        #          / |  \
        #        /   |     \
        #      /     |y*N     \
        #   (A)______|_________(B)
        # ab ist vektor punkt1 -> punkt2 = a -> b
        ab = punkt( [punkt2.x()-punkt1.x(), punkt2.y()-punkt1.y()], -1 )
        # N ist normale zu ab
        N = punkt( [ab.y(), -ab.x()], -2 )
        # berechne faktor y mit dem N multipliziert werden muss, um punkt3 zu erreichen

        try:
            ergebnis = (((punkt1.y()- punkt3.y())-( (punkt1.x() - punkt3.x()) * ab.y() / ab.x() ))/ (N.x() * ab.y() / ab.x() - N.y()))
        except:
            # faengt division durch null fehler ab, die hier aber nicht auftauchen koennen sollten
            ergebnis = 1
            self.setFehler( "Es wurde eine unmoegliche Konstellation der Daten gefunden - ist die Projektion Gauss-Krueger?" )


        return ergebnis

    def bildeDreiecke(self, index1, index2, alterPunkt, nummerDreieck):
        resultat = []

        if self.punktefeld[index1].gehortZuMehrAlsEinemDreieckMit(self.punktefeld[index2]):
            return resultat

        # experimentell!!!
        #if self.punktefeld[index1].endpunkt and self.punktefeld[index2].endpunkt:
            #return resultat

        maximalerWinkel = 0
        aktuellerWinkel = 0
        grenzWinkel = self.grenzWinkel
        umrechnungsFaktor = self.umrechnungsFaktor
        index3 = -1
        normFaktor = 0

        c = self.punktefeld[index1].abstandZu( self.punktefeld[index2] )

        if alterPunkt != -1:
            normFaktor = self.berechneNormalenFaktor( self.punktefeld[index1], self.punktefeld[index2], self.punktefeld[alterPunkt] )

        for i in range( len(self.punktefeld) ):
            if i == index1 or i == index2 or i == alterPunkt:
                continue

            aktuellerNormFaktor = self.berechneNormalenFaktor( self.punktefeld[index1], self.punktefeld[index2], self.punktefeld[i] )
            if aktuellerNormFaktor == 0:
                continue
            elif normFaktor>0 and aktuellerNormFaktor>0:
                continue
            elif normFaktor<0 and aktuellerNormFaktor<0:
                continue

            #           (C)
            #          / |  \
            #       b/   |     \a
            #      /     |h       \
            #   (A)______|_________(B)
            #               c
            aktuellerWinkel = 0
            b = self.punktefeld[index1].abstandZu( self.punktefeld[i] )
            a = self.punktefeld[index2].abstandZu( self.punktefeld[i] )

            # a^2=b^2+c^2-2*b*c*cos( Gamma ) (Kosinussatz).
            # => gamma = acos( (a^2+b^2-c^2)/(2ab) )
            cosGamma = (a*a+b*b-c*c)/(2*b*a)
            if cosGamma >= -1.0 and cosGamma <= 1.0:
                aktuellerWinkel = math.acos( cosGamma )
            else:
                if self.debug == True:
                    print "cosGamma:", repr(cosGamma)
                continue

            if aktuellerWinkel > maximalerWinkel:
                maximalerWinkel = aktuellerWinkel
                index3 = i


        if index3 >= 0 and (self.punktefeld[index3].istVerbundenMit( self.punktefeld[index1].index )==False or \
                           self.punktefeld[index3].istVerbundenMit( self.punktefeld[index2].index )==False):

            # minimum 10 grad
            if maximalerWinkel < grenzWinkel:
                self.punktefeld[index1].endpunkt = True
                self.punktefeld[index2].endpunkt = True
                self.punktefeld[index3].endpunkt = True
                #print "-------------> Rand:", index1, ", ", index2, ", ", index3
                return resultat

            if self.debug == True:
                print "maximalerWinkel:", maximalerWinkel, (maximalerWinkel*umrechnungsFaktor)

            self.punktefeld[index1].gehoertZuDreieck.append(nummerDreieck)
            self.punktefeld[index2].gehoertZuDreieck.append(nummerDreieck)
            self.punktefeld[index3].gehoertZuDreieck.append(nummerDreieck)

            if self.punktefeld[index3].istVerbundenMit( self.punktefeld[index1].index )==False:
                self.punktefeld[index1].verbindenMit( self.punktefeld[index3].index )
                self.punktefeld[index3].verbindenMit( self.punktefeld[index1].index )
                self.kantenfeld.append( kante( self.punktefeld[index1].x(), self.punktefeld[index1].y(), self.punktefeld[index3].x(), self.punktefeld[index3].y(), self.punktefeld[index1].index, self.punktefeld[index3].index ) )
                resultat.append( [index1, index3, index2] )
                self.kantenErstellt += 1

            if self.punktefeld[index3].istVerbundenMit( self.punktefeld[index2].index )==False:
                self.punktefeld[index2].verbindenMit( self.punktefeld[index3].index )
                self.punktefeld[index3].verbindenMit( self.punktefeld[index2].index )
                self.kantenfeld.append( kante( self.punktefeld[index2].x(), self.punktefeld[index2].y(), self.punktefeld[index3].x(), self.punktefeld[index3].y(), self.punktefeld[index2].index, self.punktefeld[index3].index ) )
                resultat.append( [index2, index3, index1] )
                self.kantenErstellt += 1

            return resultat
        else:
            if self.debug == True:
                print "- index3:", self.punktefeld[index3].istVerbundenMit(self.punktefeld[index1].index ), self.punktefeld[index3].istVerbundenMit(self.punktefeld[index2].index )
            self.punktefeld[index1].endpunkt = True
            self.punktefeld[index2].endpunkt = True
        return resultat

    def errechneDiagramm_Iterativ( self, ind1, ind2, ap, dreieckZaehler ):
        nummer = dreieckZaehler
        aktuelleListe = [[ ind1, ind2, ap, nummer ]]

        while aktuelleListe!=[] and self.fehlerIstAufgetreten()==False:
            neueListe = []
            wiederholen = True

            while ( wiederholen ):
                wiederholen = False
                for i in aktuelleListe:
                    if i == []:
                        continue
                    nummer += 1
                    neueErgebnisse = self.bildeDreiecke( i[0], i[1], i[2], nummer )

                    neueListe.extend( neueErgebnisse )
                if (neueListe==[] and self.kantenErstellt==1 and self.grenzWinkel > 0):
                    self.grenzWinkel -= 0.017
                    print "self.grenzWinkel -= 0.017", self.grenzWinkel
                    wiederholen = True
                elif (self.grenzWinkel != self.standardGrenzWinkel):
                    self.grenzWinkel = self.standardGrenzWinkel
                    print "self.grenzWinkel = self.standardGrenzWinkel", self.grenzWinkel


            del aktuelleListe
            aktuelleListe = neueListe

    def loesePunktInFeldIndexAuf( self, punkt ):
        try:
            return self.punktefeld.index(punkt)
        except:
            print "FEHLER: punkt in feld nicht gefunden"
            return -1


    def nutzeVorhandenesPunkteFeld(self, feld):
        for i in feld:
            neuerPunkt = delaunayPunkt( [i.x(), i.y()], i.index )
            neuerPunkt.multipliziert = i.multipliziert
            neuerPunkt.normiert = i.normiert
            self.punktefeld.append( neuerPunkt )
        self.geschaetzteKantenMenge = len(feld)*3

    def finde_naechsten_Nachbarn(self):
        """
        finde neachsten Nachbarn in der gesamten Punktmenge
        """
        if self.sortiert==False:
            sortierer = feld_sortierer( self.punktefeld )
            self.punktefeld = sortierer.sortiertesFeld()
            self.sortiert = True
            self.sortiertNach = x_komponente
        self.finde_Nachbarn_rekursiv(0, len(self.punktefeld)-1)

    def finde_Nachbarn_rekursiv(self, von, bis):
        if von < bis:
            trenner = int(math.floor( (bis + von)/2 ))
            self.finde_Nachbarn_rekursiv( von, trenner )
            self.finde_Nachbarn_rekursiv( trenner+1, bis )

            if self.debug == True:
                print "von", von, "bis", bis

            trennenderXWert = self.punktefeld[trenner].x()
            untererGrenzIndex = trenner
            obererGrenzIndex = trenner
            gefunden = False
            while gefunden == False and untererGrenzIndex > von :
                if abs(trennenderXWert - self.punktefeld[untererGrenzIndex-1].x()) > self.minimalerAbstand:
                    gefunden = True
                    break
                untererGrenzIndex-=1

            gefunden = False
            while gefunden == False and obererGrenzIndex < bis :
                if abs(self.punktefeld[obererGrenzIndex+1].x() - trennenderXWert) > self.minimalerAbstand:
                    gefunden = True
                    break
                obererGrenzIndex+=1

            grenzPunkte = []

            # Uebertragen der Punkte im Grenzstreifen in neues Hilfsfeld
            for i in range( untererGrenzIndex, obererGrenzIndex+1 ):
                grenzPunkte.append( self.punktefeld[i] )

            sortierer = feld_sortierer( grenzPunkte )
            sortierer.sortier_Kriterium = y_komponente
            grenzPunkte = sortierer.sortiertesFeld()


            for i in range( len(grenzPunkte)):
                for j in range( len(grenzPunkte)):
                    if i != j and not grenzPunkte[i].istVerbundenMit(grenzPunkte[j].index):
                        if abs( grenzPunkte[i].y() - grenzPunkte[j].y() ) < self.minimalerAbstand:
                            minAbstand = grenzPunkte[i].abstandZu( grenzPunkte[j] )
                            if minAbstand < self.minimalerAbstand :
                                if minAbstand <= 0:
                                    self.setFehler( code = FEHLER_GLEICHE_PUNKTE )
                                    print "- Achtung: zwei identische Punkte gefunden!!"
                                self.minimalerAbstand = minAbstand
                                self.punkt1 = grenzPunkte[i]
                                self.punkt2 = grenzPunkte[j]
                        elif j>i:
                            # die Abstaende sind bereits groesser, da der index j im nach y-koordinaten
                            # geordneten feld groesser ist...
                            break



