# Author: Ole Rahn
# Project: PIROL

from voronoi import *
from wxPython.wx import *
from Thuban import _
from utilities import punkt

from Thuban.Model.color import Color
from Thuban.Model.classification import ClassGroupRange, ClassGroupSingleton, ClassGroupProperties

class malPanel(wxPanel):
    def __init__( self, parent, size):
        wxPanel.__init__(self,parent,-1, style=wxMINIMIZE_BOX|wxHSCROLL|wxVSCROLL)
        self.BufferBmp = None
        self.parent = parent
        
        x, y = size
        self.SetSizeHints(minW=320, minH=240, maxW=1024, maxH=768)
        self.SetVirtualSizeHints(minW=640, minH=480, maxW=1024, maxH=768)
        self.SetVirtualSize( wxSize(640, 480))
        self.sizer = wxFlexGridSizer(rows=1, cols=1 )
        self.sizer.Add( self, flag=wxEXPAND )
        self.sizer.SetMinSize( wxSize(320, 240) )
        
        self.sizer.SetItemMinSize( self, 520, 390 )
        self.sizer.AddGrowableCol( 0 )
        self.sizer.AddGrowableRow( 0 )
        self.sizer.SetDimension( 0,0, x, y )
        
        EVT_SIZE(self, self.OnSize)
        EVT_PAINT(self, self.OnPaint)
        EVT_LEFT_DOWN(self, self.OnClick)
        
    def getSizer( self ):
        return self.sizer
        
    # OnSize is fired at the application start or when the frame is resized.
    # The OnSize event is fired BEFORE the OnPaint event. wxWindows
    # handles the events in this order. This is not due to the fact,
    # that the code line EVT_SIZE(...) is placed before the line
    # EVT_PAINT(...).
    # The procedure OnSize() is the right place to define the
    # BufferBmp and its size. self.BufferBmp is the picture in memory,
    # which contains your drawing. self.BufferBmp is also used as a flag,    
    # a None value indicates no picture.
    #
    def OnSize(self, event):
        # Get the size of the drawing area in pixels.
        self.wi, self.he = self.GetSizeTuple()
        #print self.GetSizeTuple()
        # Create BufferBmp and set the same size as the drawing area.
        self.BufferBmp = wxEmptyBitmap(self.wi, self.he)
        memdc = wxMemoryDC()
        memdc.SelectObject(self.BufferBmp)
        # Drawing job
        ret = self.zeichneAufBuffer(memdc)
        #error
        if not ret:
            self.BufferBmp = None
            wxMessageBox('Error in drawing','CommentedDrawing', wxOK | wxICON_EXCLAMATION)
        #memdc.SelectObject(wxNullBitmap) 
        #is this important?


    # OnPaint is executed at the app start, when resizing or when
    # the application windows becomes active. OnPaint copies the
    # buffered picture, instead of preparing (again) the drawing job.
    # This is the trick, copying is very fast.
    # Note: if you forget to define the dc in this procedure,
    # (no dc = ... code line), the application will run in
    # an infinite loop. This is a common beginner's error. (I think)
    # BeginDrawing() and EndDrawing() are for windows platforms (see doc).
    
    def OnPaint(self, event):
        if self.BufferBmp == None:
            # Windows ruft das Size Event nicht dann auf, wenn linux dies tun wuerde
            self.OnSize(None)
        dc = wxPaintDC(self)
        dc.BeginDrawing()
        if self.BufferBmp != None:
            dc.DrawBitmap(self.BufferBmp, 0, 0, True)
        else:
            print '...nothing to draw'
        dc.EndDrawing()
        
    def zeichneAufBuffer( self, memoryDC ):
        return self.parent.zeichneAufBuffer(memoryDC)
        
    def OnClick(self, event):
        return self.parent.OnClickInsDiagramm( event )


class diagramZeichner(wxFrame):
    def __init__( self, parent, titel ):
        wxFrame.__init__( self, parent=parent, id=-1, pos=(-1, -1), size=(-1,-1), title=_(titel), style = wxCAPTION | wxMINIMIZE_BOX | wxRESIZE_BORDER )
        
        self.parent = parent
        self.SetBackgroundColour(wxWHITE)
        #self.SetCursor(wxCROSS_CURSOR)
        self.BufferBmp = None
                
        top = wxBoxSizer(wxVERTICAL)
        panelSizer = wxBoxSizer(wxVERTICAL)
        
        box = wxBoxSizer(wxHORIZONTAL)

        self.malflaeche = malPanel(self, (520,390))
        self.malflaeche.SetBackgroundColour(wxWHITE)
        box.Add(self.malflaeche.getSizer(), 1, wxALIGN_CENTER_VERTICAL|wxEXPAND, 1)
        top.Add(box, 1, wxEXPAND, 4)
        
        buttons = wxBoxSizer(wxHORIZONTAL)
        button = wxButton(self, wxID_OK, _("OK"))
        buttons.Add(button, 0, wxALL|wxALIGN_CENTER_VERTICAL|wxEXPAND, 1)
        top.Add(buttons, 0, wxALIGN_RIGHT, 4)
        
        self.SetAutoLayout(True)
        self.SetSizer(top)
        top.Fit(self)
        #top.SetSizeHints(self)

        EVT_BUTTON(self, wxID_OK, self.OnOK)
        EVT_CLOSE(self, self.OnCancel)
        
        
    def RunDialog(self):
        self.Show(True)
        return

    def end_dialog(self):
        self.Show(False)
        self.Destroy()
        
    def OnCancel(self, event):
        self.end_dialog()

    def OnOK(self, event):
        self.end_dialog()
        
    def zeichneAufBuffer( self, memoryDC ):
        # muss von einer abgeleiteten klasse ueberschrieben werden!!
        return false
        
    def OnClickInsDiagramm(self, event):
        self.identifiziereObjektAn( event.m_x, event.m_y )
        
    def identifiziereObjektAn( self, x, y ):
        pass
    

class balken_diagramZeichner(diagramZeichner):

    def __init__( self, parent, classLayer, titel, Range, balkenwerte, diskreteWerte=False, spaltenNamen=[] ):
        diagramZeichner.__init__( self, parent, titel )
        
        self.diskret = diskreteWerte
        self.diskreteSpalten = spaltenNamen
        self.diskreteSpaltenOrig = list(spaltenNamen)
        
        self.classLayer = classLayer
        self.classLayer.SetClassification( None )
        
        self.identified = []

        # ziemlich gepfuscht: man sollte lieber den Attributnamen uebergeben...        
        colName = titel[ titel.rfind( "fuer " )+5: ]
        
        self.classLayer.SetClassificationColumn(colName)
        
        if self.diskreteSpalten != []:
            for i in range( len(self.diskreteSpalten)):
                if len( self.diskreteSpalten[i] ) > 9:
                    tmpStr = self.diskreteSpalten[i]
                    teilStrings = tmpStr.split( "-" )
                    tmpStr = ""
                    for j in teilStrings:
                        j = j[:4] + "."
                        if len(tmpStr) != 0:
                            tmpStr = tmpStr + "-" + j
                        else:
                            tmpStr = j
                    self.diskreteSpalten[i] = tmpStr
                
        
        self.malflaeche.parent = self
        
        self.hoehe = 0
        self.breite = 0
        
        self.menge = 0
        
        self.xRange, self.yRange = Range
        self.ySteps = 10.0
        
        self.xSteps = float(len( balkenwerte ))
        self.balkenwerte = balkenwerte
        
        self.balkenbreite = 0
        self.diagrammBreite = 0
        
        self.nullPunkt = (0,0)
        
    def setzteNullpunkt( self, breite, hoehe ):
        self.nullPunkt = ( int(breite*0.1), int(hoehe*0.83) )
        self.diagrammBreite = breite*0.78
        self.diagrammHoehe = hoehe*0.78
        self.balkenbreite = int( (self.diagrammBreite - (self.xSteps * 10))/self.xSteps  )
        if self.balkenbreite > 100:
            self.balkenbreite = 100
        elif self.balkenbreite < 3:
            self.balkenbreite = 3
            
        for i in self.balkenwerte:
            self.menge += i
        
    def rechnePunktUm( self, punkt ):
        x, y = punkt
        nullX, nullY = self.nullPunkt
        return ( nullX + x, nullY - y )
        
    def rechneXKoordUm( self, x ):
        nullX, nullY = self.nullPunkt
        return int(nullX + x)
        
    def rechneYKoordUm( self, y ):
        nullX, nullY = self.nullPunkt
        return int(nullY - y)
        
    def identifiziereObjektAn( self, x, y ):
        mouse = punkt([x,y])
        xAnfang, xEnde = self.xRange
        yAnfang, yEnde = self.yRange
        xSpanne = xEnde - xAnfang
        ySpanne = yEnde - yAnfang
        
        xStep = 1.0*xSpanne/self.xSteps
        yStep = 1.0*ySpanne/self.ySteps
        
        x_Step = 1.0 * self.diagrammBreite/self.xSteps
        y_Step = 1.0 * self.diagrammHoehe/self.ySteps
        
        gesuchterBalken = -1
        menge = float( len( self.balkenwerte ) )
        
        for i in range( int(menge) ):
            
            x_ul = self.rechneXKoordUm(x_Step * i)
            x_ur = self.rechneXKoordUm(x_Step * i + self.balkenbreite)
            
            if ( mouse.x() >= x_ul and mouse.x() <= x_ur ):
                gesuchterBalken = i
                print "gesuchterBalken (", i, "): ", str( self.balkenwerte[i] )
                break
        if ( gesuchterBalken == -1 or gesuchterBalken in self.identified):
            return None
        
        self.identified.append( gesuchterBalken )
        classific = self.classLayer.GetClassification()
        props = ClassGroupProperties()
        props.SetFill( Color( float(1.0/menge*gesuchterBalken), 0.0, float(1.0-(1.0/menge*gesuchterBalken)) ) )
        
        if ( self.diskret == False ):
            startWert = xAnfang + xStep * gesuchterBalken - 0.0000001
            endWert = xAnfang + xStep * (gesuchterBalken+1) + 0.0000001
            print "werte ", startWert, endWert
            classific.AppendGroup( ClassGroupRange( (startWert, endWert), props, repr((startWert, endWert)) ) )
            
        else:
            wert = self.diskreteSpaltenOrig[gesuchterBalken]
            classific.AppendGroup( ClassGroupSingleton( wert, props, wert ) )
            print "wert ", wert
        
        
    def zeichneBalken( self, memoryDC ):
        for i in range( len( self.balkenwerte ) ):
            self.zeichneEinenBalken( memoryDC, i, self.balkenwerte[i] )
        
        
    def zeichneEinenBalken( self, memoryDC, balkenNr, balkenYWert ):
        xAnfang, xEnde = self.xRange
        yAnfang, yEnde = self.yRange
        xSpanne = xEnde - xAnfang
        ySpanne = yEnde - yAnfang
        xStep = 1.0*xSpanne/self.xSteps
        yStep = 1.0*ySpanne/self.ySteps
        
        x_Step = 1.0 * self.diagrammBreite/self.xSteps
        y_Step = 1.0 * self.diagrammHoehe/self.ySteps
        
        punkt_ul = ( self.rechneXKoordUm(x_Step * balkenNr), self.rechneYKoordUm(0) )
        punkt_ur = ( self.rechneXKoordUm(x_Step * balkenNr + self.balkenbreite), self.rechneYKoordUm(0) )
        
        punkt_ol = ( self.rechneXKoordUm(x_Step * balkenNr), self.rechneYKoordUm( balkenYWert/ySpanne*self.diagrammHoehe ) )
        punkt_or = ( self.rechneXKoordUm(x_Step * balkenNr + self.balkenbreite), self.rechneYKoordUm( balkenYWert/ySpanne*self.diagrammHoehe ) )
        
        memoryDC.DrawPolygon( [ punkt_ul, punkt_ur, punkt_or, punkt_ol, punkt_ul ] )
        
    def zeichneAchsen( self, memoryDC ):
        xAnfang, xEnde = self.xRange
        yAnfang, yEnde = self.yRange
        xSpanne = xEnde - xAnfang
        ySpanne = yEnde - yAnfang
        xStep = 1.0*xSpanne/self.xSteps
        yStep = 1.0*ySpanne/self.ySteps
        
        
        x_Step = 1.0 * self.diagrammBreite/self.xSteps
        y_Step = 1.0 * self.diagrammHoehe/self.ySteps
            
        memoryDC.SetBrush(wxBrush(wxCYAN, wxTRANSPARENT))
        memoryDC.SetPen(wxPen(wxBLACK, 2, wxSOLID))
        # X
        memoryDC.DrawLine( self.rechneXKoordUm(0), self.rechneYKoordUm(0), self.rechneXKoordUm(self.diagrammBreite), self.rechneYKoordUm(0))
        
        for i in range( int(self.xSteps) ):
            memoryDC.DrawLine( self.rechneXKoordUm(i*x_Step), self.rechneYKoordUm(-3), self.rechneXKoordUm(i*x_Step), self.rechneYKoordUm(3))
            if self.diskret!=True:
                xZahl = repr(1.0*xAnfang+i*xStep)
                xZahl = xZahl[:5]
                xText = xZahl
            else:
                xText = str(self.diskreteSpalten[i])
                xText = xText[:9]
            memoryDC.DrawRotatedText( xText, self.rechneXKoordUm(i*x_Step+5), self.rechneYKoordUm(-10), 270 )
        
        endPunkte = [ (self.rechneXKoordUm(self.diagrammBreite), self.rechneYKoordUm(-3)),  (self.rechneXKoordUm(int(self.diagrammBreite)), self.rechneYKoordUm(3)), (self.rechneXKoordUm(self.diagrammBreite+10), self.rechneYKoordUm(0)), (self.rechneXKoordUm(self.diagrammBreite), self.rechneYKoordUm(-3))]
        memoryDC.DrawPolygon( endPunkte )
        
        # Y
        memoryDC.SetBrush(wxBrush(wxCYAN, wxTRANSPARENT))
        memoryDC.SetPen(wxPen(wxBLACK, 2, wxSOLID))
        memoryDC.DrawLine( self.rechneXKoordUm(0), self.rechneYKoordUm(0), self.rechneXKoordUm(0), self.rechneYKoordUm(self.diagrammHoehe))
        
        for i in range( int(self.ySteps) ):
            memoryDC.DrawLine( self.rechneXKoordUm(-3), self.rechneYKoordUm(i*y_Step), self.rechneXKoordUm(3), self.rechneYKoordUm(i*y_Step))
            yZahl = repr(1.0*yAnfang+i*yStep)
            yZahl = yZahl[:5]
            yText = yZahl
            memoryDC.DrawText( yText, self.rechneXKoordUm(-40), self.rechneYKoordUm(i*y_Step+7) )
            
        endPunkte = [ (self.rechneXKoordUm(-3), self.rechneYKoordUm(self.diagrammHoehe)),  (self.rechneXKoordUm(3), self.rechneYKoordUm(self.diagrammHoehe)), (self.rechneXKoordUm(0), self.rechneYKoordUm(self.diagrammHoehe+10)), (self.rechneXKoordUm(-3), self.rechneYKoordUm(self.diagrammHoehe))]
        memoryDC.DrawPolygon( endPunkte )
        
        memoryDC.SetBrush(wxBrush(wxCYAN, wxTRANSPARENT))
        memoryDC.SetPen(wxPen(wxBLACK, 2, wxSOLID))
        
    def zeichneDaten( self, memoryDC ):
        schriftHoehe = 15
        trennerSchritt = int(schriftHoehe+schriftHoehe/2)
        memoryDC.DrawText( "Menge:", self.rechneXKoordUm(self.diagrammBreite), self.rechneYKoordUm(self.diagrammHoehe) )
        text = str(int(self.menge))
        memoryDC.DrawText( text, self.rechneXKoordUm(self.diagrammBreite), self.rechneYKoordUm(self.diagrammHoehe-schriftHoehe) )
        
        memoryDC.DrawText( "Num.Kat.:", self.rechneXKoordUm(self.diagrammBreite), self.rechneYKoordUm(self.diagrammHoehe-(1*schriftHoehe+1*trennerSchritt)) )
        text = str(self.xSteps)
        text = text[:5]
        memoryDC.DrawText( text, self.rechneXKoordUm(self.diagrammBreite), self.rechneYKoordUm(self.diagrammHoehe-(2*schriftHoehe+1*trennerSchritt)) )
        
        if self.diskret!=True:
            xAnfang, xEnde = self.xRange
            memoryDC.DrawText( "min.Wert:", self.rechneXKoordUm(self.diagrammBreite), self.rechneYKoordUm(self.diagrammHoehe-(2*schriftHoehe+2*trennerSchritt)) )
            text = str(xAnfang)
            text = text[:5]
            memoryDC.DrawText( text, self.rechneXKoordUm(self.diagrammBreite), self.rechneYKoordUm(self.diagrammHoehe-(3*schriftHoehe+2*trennerSchritt)) )
            
            memoryDC.DrawText( "max.Wert:", self.rechneXKoordUm(self.diagrammBreite), self.rechneYKoordUm(self.diagrammHoehe-(3*schriftHoehe+3*trennerSchritt)) )
            text = str(xEnde)
            text = text[:5]
            memoryDC.DrawText( text, self.rechneXKoordUm(self.diagrammBreite), self.rechneYKoordUm(self.diagrammHoehe-(4*schriftHoehe+3*trennerSchritt)) )
        
    def zeichneAufBuffer( self, memoryDC ):
        try:
            if memoryDC==None:
                print "keine DC!!"
                return False
            memoryDC.BeginDrawing()

            # Clear everything
            memoryDC.SetBrush(wxBrush(wxWHITE, wxSOLID))
            memoryDC.Clear()

            # zeichne Rahmen
            self.breite, self.hoehe = memoryDC.GetSizeTuple()
            self.setzteNullpunkt( self.breite, self.hoehe )
            memoryDC.SetBrush(wxBrush(wxCYAN, wxTRANSPARENT))
            memoryDC.SetPen(wxPen(wxBLACK, 2, wxSOLID))
            memoryDC.DrawRectangle(5, 5, self.breite-10, self.hoehe-10)
            
            self.zeichneAchsen( memoryDC )
            self.zeichneBalken( memoryDC )
            self.zeichneDaten( memoryDC )

            memoryDC.EndDrawing()
            return True

        except:
            return False
            
            
# ***********************************************************************
# ***********************************************************************

class punkt_diagramZeichner(diagramZeichner):

    def __init__( self, parent, titel, Range, punktefeld, clusterZentren, achsenNamen=[], tool=None, fuzzy=0 ):
        diagramZeichner.__init__( self, parent, titel )
        
        self.malflaeche.parent = self
        self.tool = tool
        self.bboxLayer = self.tool.layer
        
        self.hoehe = 0
        self.breite = 0
        
        self.menge = len( punktefeld )
        self.fuzzyStaerke = fuzzy
        
        self.xRange, self.yRange = Range
        self.ySteps = 10.0
        
        self.xSteps = 10.0
        self.punktefeld = punktefeld
        self.clusterZentren = clusterZentren
        
        self.maxPunktDurchm = 10
        self.diagrammBreite = 0
        
        self.nullPunkt = (0,0)
        
        # Bezeichnungen
        self.xAchse = "X"
        self.yAchse = "Y"
        
        self.farben = []
        
        if len(self.clusterZentren) > 1:
            colorStep = int( 255 / (len(self.clusterZentren)-1) )
        else:
            colorStep = 255
        for i in range( len(self.clusterZentren) ):
            self.farben.append( wxColour( colorStep * i, 0, 255 - (colorStep * i) ) )
        
        if achsenNamen!=[]:
            self.xAchse = achsenNamen[0][:10]
            self.yAchse = achsenNamen[1][:7]
        
    def setzteNullpunkt( self, breite, hoehe ):
        #self.nullPunkt = ( int(breite*0.9), int(hoehe*0.9) )
        self.nullPunkt = ( int(breite*0.1), int(hoehe*0.83) )
        self.diagrammBreite = breite*0.78
        self.diagrammHoehe = hoehe*0.78
        self.maxPunktDurchm = int( round(self.diagrammBreite / 35)  )
        if self.maxPunktDurchm > 30:
            self.maxPunktDurchm = 30
        elif self.maxPunktDurchm < 4:
            self.maxPunktDurchm = 4
            
    def rechnePunktUm( self, punkt ):
        x, y = punkt
        nullX, nullY = self.nullPunkt
        return ( nullX + x, nullY - y )
        
    def rechneXKoordUm( self, x ):
        nullX, nullY = self.nullPunkt
        return int(nullX + x)
        
    def rechneYKoordUm( self, y ):
        nullX, nullY = self.nullPunkt
        return int(nullY - y)
        
    def zeichnePunkte( self, memoryDC ):
        for i in range( len( self.punktefeld ) ):
            self.zeichneEinenPunkt( memoryDC, self.punktefeld[i] )
            
        for i in range( len( self.clusterZentren ) ):
            self.zeichneEinZentrum( memoryDC, self.clusterZentren[i], i )
            
    def identifiziereObjektAn( self, x, y ):
        #print "identifiziereObjektAn", x, y
        mouse = punkt([x,y])
        xAnfang, xEnde = self.xRange
        yAnfang, yEnde = self.yRange
        xSpanne = xEnde - xAnfang
        ySpanne = yEnde - yAnfang
        
        zutreffendesZentrum = None
        
        for zentrum in self.clusterZentren:
            zentrumX = self.rechneXKoordUm( (1.0* (zentrum.koord[0] - xAnfang) /  xSpanne) * self.diagrammBreite)
            zentrumY = self.rechneYKoordUm( (1.0* (zentrum.koord[1] - yAnfang) /  ySpanne) * self.diagrammHoehe)
            tmp = punkt( [zentrumX, zentrumY] )
            abstand = tmp.abstandZu( mouse )
            if ( abstand <= self.maxPunktDurchm/2 ):
                zutreffendesZentrum = zentrum
                print "clusterZentrum (", str( self.clusterZentren.index(zentrum) ), "):", zentrum, ", mouse an:", mouse
                break
        if ( zutreffendesZentrum == None ):
            return None
        
        indexeUndZugehoerigePunkte = {}
        
        for i in range( len( self.punktefeld ) ):
            if ( self.punktefeld[i].GetPrimaereZugehoerigkeit() == zutreffendesZentrum.index ):
                indexeUndZugehoerigePunkte[ i ] = self.punktefeld[i]
            
        self.tool.identifizierePunkte( zutreffendesZentrum.index, indexeUndZugehoerigePunkte, self.farben[zutreffendesZentrum.index], self.bboxLayer.getId() )
        
    def zeichneEinZentrum( self, memoryDC, punkt, nummer ):
        xAnfang, xEnde = self.xRange
        yAnfang, yEnde = self.yRange
        xSpanne = xEnde - xAnfang
        ySpanne = yEnde - yAnfang

        x = (1.0* (punkt.koord[0] - xAnfang) /  xSpanne) * self.diagrammBreite
        y = (1.0* (punkt.koord[1] - yAnfang) /  ySpanne) * self.diagrammHoehe
        
        #memoryDC.SetBrush( wxBrush(wxColour(255,0,0), wxSOLID) )
        hoechsterWert = 0
        farbWert = self.farben[nummer]
        
        alteFont = memoryDC.GetFont()
        alterSchriftBg = memoryDC.GetTextBackground()
        neueFont = wxFont( weight = wxNORMAL, pointSize = 9, family = wxDEFAULT, style = wxNORMAL )
        punktKoordinaten = str(punkt)
        
        
        
        alterPen = memoryDC.GetPen()
        neuerPen = wxPen( farbWert, 3 )
        dickerSchwarzerPen = wxPen( wxColour(0,0,0), 5 )
        dickerWeisserPen = wxPen( wxColour(255,255,255), 7 )

        memoryDC.SetPen( alterPen )
        memoryDC.SetFont( neueFont )
        memoryDC.SetTextBackground(farbWert)
        memoryDC.SetTextForeground(wxColour(255,255,255) )
        memoryDC.SetBackgroundMode( wxSOLID )
        
        breite, hoehe = memoryDC.GetTextExtent(punktKoordinaten)
        memoryDC.DrawText( punktKoordinaten, self.rechneXKoordUm(self.diagrammBreite-breite), self.rechneYKoordUm(self.diagrammHoehe-(nummer*hoehe)) )
        
        memoryDC.SetFont( alteFont )
        memoryDC.SetTextForeground(wxColour(0,0,0) )
        memoryDC.SetTextBackground(alterSchriftBg)
        memoryDC.SetBackgroundMode( wxTRANSPARENT )
        
        if self.menge > 100:
            memoryDC.SetPen( dickerWeisserPen )
            memoryDC.DrawLine( self.rechneXKoordUm(x), self.rechneYKoordUm(y+self.maxPunktDurchm/2), self.rechneXKoordUm(x), self.rechneYKoordUm(y-self.maxPunktDurchm/2) )
            memoryDC.DrawLine( self.rechneXKoordUm(x+self.maxPunktDurchm/2), self.rechneYKoordUm(y), self.rechneXKoordUm(x-self.maxPunktDurchm/2), self.rechneYKoordUm(y) )
            
        
        memoryDC.SetPen( dickerSchwarzerPen )
        memoryDC.DrawLine( self.rechneXKoordUm(x), self.rechneYKoordUm(y+self.maxPunktDurchm/2), self.rechneXKoordUm(x), self.rechneYKoordUm(y-self.maxPunktDurchm/2) )
        memoryDC.DrawLine( self.rechneXKoordUm(x+self.maxPunktDurchm/2), self.rechneYKoordUm(y), self.rechneXKoordUm(x-self.maxPunktDurchm/2), self.rechneYKoordUm(y) )
        
        memoryDC.SetPen( neuerPen )
        memoryDC.DrawLine( self.rechneXKoordUm(x), self.rechneYKoordUm(y+self.maxPunktDurchm/2), self.rechneXKoordUm(x), self.rechneYKoordUm(y-self.maxPunktDurchm/2) )
        memoryDC.DrawLine( self.rechneXKoordUm(x+self.maxPunktDurchm/2), self.rechneYKoordUm(y), self.rechneXKoordUm(x-self.maxPunktDurchm/2), self.rechneYKoordUm(y) )

        
    
    def zeichneEinenPunkt( self, memoryDC, punkt ):
        xAnfang, xEnde = self.xRange
        yAnfang, yEnde = self.yRange
        xSpanne = xEnde - xAnfang
        ySpanne = yEnde - yAnfang

        #print "zeichne", punkt

        x = (1.0* (punkt.koord[0] - xAnfang) /  xSpanne) * self.diagrammBreite
        y = (1.0* (punkt.koord[1] - yAnfang) /  ySpanne) * self.diagrammHoehe
        
        #memoryDC.SetBrush( wxBrush(wxColour(255,0,0), wxSOLID) )
        hoechsterWert = 0
        farbindex = 0
        for i in range( len(self.clusterZentren) ):
            if hoechsterWert < punkt.GetZugehoerigkeit( i ):
                hoechsterWert = punkt.GetZugehoerigkeit( i )
                farbindex = i
        
        neuerPen = wxPen( wxColour(0,0,0), 1 )        
        memoryDC.SetPen( neuerPen )
                
        memoryDC.SetBrush( wxBrush(self.farben[farbindex], wxSOLID) )
        
        memoryDC.DrawCircle( self.rechneXKoordUm(x), self.rechneYKoordUm(y), int(self.maxPunktDurchm/2) )
        
        
        startWinkel = 0
        endWinkel = 90 
        for i in range( len(self.clusterZentren) ):
            if i == farbindex:
                continue
            if punkt.GetZugehoerigkeit( i ) >= 0.01:
                startWinkel = endWinkel
                endWinkel += int(360 * punkt.GetZugehoerigkeit( i ))
                memoryDC.SetBrush( wxBrush(self.farben[i], wxSOLID) )
                memoryDC.DrawEllipticArc( self.rechneXKoordUm(x-self.maxPunktDurchm/2), self.rechneYKoordUm(y+self.maxPunktDurchm/2), self.maxPunktDurchm, self.maxPunktDurchm, startWinkel, endWinkel )
        
        memoryDC.SetBrush(wxBrush(wxColour(0,0,0), wxTRANSPARENT))

    def zeichneAchsen( self, memoryDC ):
        xAnfang, xEnde = self.xRange
        yAnfang, yEnde = self.yRange
        xSpanne = xEnde - xAnfang
        ySpanne = yEnde - yAnfang
        xStep = 1.0*xSpanne/self.xSteps
        yStep = 1.0*ySpanne/self.ySteps
        
        
        x_Step = 1.0 * self.diagrammBreite/self.xSteps
        y_Step = 1.0 * self.diagrammHoehe/self.ySteps
            
        memoryDC.SetBrush(wxBrush(wxCYAN, wxTRANSPARENT))
        memoryDC.SetPen(wxPen(wxBLACK, 2, wxSOLID))
        # X
        memoryDC.DrawText( self.xAchse, self.rechneXKoordUm(self.diagrammBreite-20), self.rechneYKoordUm(-20) )
        memoryDC.DrawLine( self.rechneXKoordUm(0), self.rechneYKoordUm(0), self.rechneXKoordUm(self.diagrammBreite), self.rechneYKoordUm(0))
        
        for i in range( int(self.xSteps) ):
            memoryDC.DrawLine( self.rechneXKoordUm(i*x_Step), self.rechneYKoordUm(-3), self.rechneXKoordUm(i*x_Step), self.rechneYKoordUm(3))
            xZahl = repr(1.0*xAnfang+i*xStep)
            xZahl = xZahl[:5]
            xText = xZahl

            memoryDC.DrawRotatedText( xText, self.rechneXKoordUm(i*x_Step+5), self.rechneYKoordUm(-10), 270 )
        
        endPunkte = [ (self.rechneXKoordUm(self.diagrammBreite), self.rechneYKoordUm(-3)),  (self.rechneXKoordUm(int(self.diagrammBreite)), self.rechneYKoordUm(3)), (self.rechneXKoordUm(self.diagrammBreite+10), self.rechneYKoordUm(0)), (self.rechneXKoordUm(self.diagrammBreite), self.rechneYKoordUm(-3))]
        memoryDC.DrawPolygon( endPunkte )
        
        # Y
        memoryDC.DrawText( self.yAchse, self.rechneXKoordUm(-45), self.rechneYKoordUm(self.diagrammHoehe) )
        memoryDC.SetBrush(wxBrush(wxCYAN, wxTRANSPARENT))
        memoryDC.SetPen(wxPen(wxBLACK, 2, wxSOLID))
        memoryDC.DrawLine( self.rechneXKoordUm(0), self.rechneYKoordUm(0), self.rechneXKoordUm(0), self.rechneYKoordUm(self.diagrammHoehe))
        
        for i in range( int(self.ySteps) ):
            memoryDC.DrawLine( self.rechneXKoordUm(-3), self.rechneYKoordUm(i*y_Step), self.rechneXKoordUm(3), self.rechneYKoordUm(i*y_Step))
            yZahl = repr(1.0*yAnfang+i*yStep)
            yZahl = yZahl[:5]
            yText = yZahl
            memoryDC.DrawText( yText, self.rechneXKoordUm(-40), self.rechneYKoordUm(i*y_Step+7) )
            
        endPunkte = [ (self.rechneXKoordUm(-3), self.rechneYKoordUm(self.diagrammHoehe)),  (self.rechneXKoordUm(3), self.rechneYKoordUm(self.diagrammHoehe)), (self.rechneXKoordUm(0), self.rechneYKoordUm(self.diagrammHoehe+10)), (self.rechneXKoordUm(-3), self.rechneYKoordUm(self.diagrammHoehe))]
        memoryDC.DrawPolygon( endPunkte )
        
        memoryDC.SetBrush(wxBrush(wxCYAN, wxTRANSPARENT))
        memoryDC.SetPen(wxPen(wxBLACK, 2, wxSOLID))
        
    def zeichneDaten( self, memoryDC ):
        schriftHoehe = 15
        trennerSchritt = int(schriftHoehe+schriftHoehe/2)
        
        xAnfang, xEnde = self.xRange
        yAnfang, yEnde = self.yRange
        
        bezeichner = [ "Menge:", "Num.Cluster:", "min.Wert X:", "max.Wert X:", "min.Wert Y:", "max.Wert Y:" ]
        werte = [ self.menge, len(self.clusterZentren), xAnfang, xEnde, yAnfang, yEnde ]
        
        if self.fuzzyStaerke >= 1:
            bezeichner.append( "Fuzzyfiz.:")
            werte.append( self.fuzzyStaerke )
        
        for i in range( min( len(bezeichner), len(werte) ) ):
            memoryDC.DrawText( bezeichner[i], self.rechneXKoordUm(self.diagrammBreite), self.rechneYKoordUm(self.diagrammHoehe-(i*2*schriftHoehe)-(int(i/2)*trennerSchritt)) )
            text = str(repr(werte[i]))
            text = text[:5]
            memoryDC.DrawText( text, self.rechneXKoordUm(self.diagrammBreite), self.rechneYKoordUm(self.diagrammHoehe-((i*2+1)*schriftHoehe)-(int(i/2)*trennerSchritt)) )
        
        
    def zeichneAufBuffer( self, memoryDC ):
        try:
            if memoryDC==None:
                print "keine DC!!"
                return False
            memoryDC.BeginDrawing()

            # Clear everything
            memoryDC.SetBrush(wxBrush(wxWHITE, wxSOLID))
            memoryDC.Clear()

            # zeichne Rahmen
            self.breite, self.hoehe = memoryDC.GetSizeTuple()
            self.setzteNullpunkt( self.breite, self.hoehe )
            memoryDC.SetBrush(wxBrush(wxCYAN, wxTRANSPARENT))
            memoryDC.SetPen(wxPen(wxBLACK, 2, wxSOLID))
            memoryDC.DrawRectangle(5, 5, self.breite-10, self.hoehe-10)
            
            self.zeichnePunkte( memoryDC )
            self.zeichneAchsen( memoryDC )
            self.zeichneDaten( memoryDC )

            memoryDC.EndDrawing()
            return True

        except:
            return False

    def end_dialog(self):
        if self.tool != None:
            self.tool.RemoveLayer(self.bboxLayer)
        diagramZeichner.end_dialog(self)
        
# ---------------------------------------------------------------