from Thuban.UI.viewport import Tool
from lineTool import LineTool
import os
from Thuban import _
from Thuban.UI.command import registry, ToolCommand

from Thuban.Model.messages import LAYER_PROJECTION_CHANGED, LAYER_VISIBILITY_CHANGED, \
     LAYER_CHANGED, LAYER_SHAPESTORE_REPLACED, CLASS_CHANGED

from utilities import punkt
from customLayer import customLayer
from Thuban.Model.table import *
from data_fake import MemoryShapeStore

#import Projection
import math

from time import sleep
from thread import *

class polyTool( LineTool ):
    def __init__( self, view, context, findePunktLayer, appAnpasser ):
        LineTool.__init__( self, view )

        self._findePunktLayer = findePunktLayer
        self.appAnpasser = appAnpasser
        self.context = context

        self.layerName = "Polygon Nr. "

        self.storageObjekt = MemoryShapeStore()
        self.storageObjekt.setTable( self._erstelleLeereTabelle() )

        nummer = self._sucheLayerNummer( self.context.mainwindow.canvas.Map() )
        self.layer = customLayer( _(self.layerName+str(nummer)), self.storageObjekt)

        self.teilstuecke = 0
        self.gesamtLaenge = 0.0
        
        self.closing = 0

        self.punkte = []
        
        self.dragCounter = 0
        self.drawing = 0
        
        self.lastEvent = None
        
#        self.layerMitPunktdaten = -1
#        
#        if (map.HasLayers() == True):
#            if ( self.layerMitPunktdaten < 0 ):
#                self.layerMitPunktdaten = self._findePunktLayer(self.context)
#
#            if self.layerMitPunktdaten < 0:
#                main_window.RunMessageBox(_('missing base data layer'), _('Please add a layer consisting of point data, first'))
#                LineTool.drag_stop(self, event.m_x, event.m_y)
#                return

    def Name(self):
        return "PolyWerkzeug"

    def ausgangsPunkt(self):
        if ( len(self.punkte) != 0 ):
            return self.punkte[0]
        return None

    def Punkt1( self ):
        return self.ersterPunkt

    def Punkt2( self ):
        return self.zweiterPunkt


    def _berechneEntfernung( self, p1, p2 ):
        distanz = 0.0
        if ( True ):
            # Gauss-Krueger: 1.0 == 1m
            distanz = p1.abstandZu(p2)

        return distanz
        
    def _sucheLayerNummer( self, map, name=None ):
        if (name == None):
            name = self.layerName
        if map.HasLayers() == True:
            hoechsteBisher = -1
            for layer in map.layers:
                if ( layer.Title().startswith(name) and isinstance(layer, customLayer) ):
                    zuPruefen = layer.Title()
                    zuPruefen = zuPruefen.replace( name, "" )

                    if ( zuPruefen.isdigit() ):
                        neueNummer = int(zuPruefen)
                        if ( neueNummer > hoechsteBisher ):
                            hoechsteBisher = neueNummer

            return hoechsteBisher+1

        return 0

    def _aktualisiereTeilstuecke( self, neueDist ):
        self.gesamtLaenge += neueDist
        self.teilstuecke += 1

    def _fuegeLinieEinZwischen( self, event, punkt1, punkt2 ):
        main_window = self.context.mainwindow
        map = main_window.canvas.Map()
        self.appAnpasser.passeAn( self.context.application )

        if map.HasLayers() == True and punkt1 != None and punkt2 != None:

            distanz = self._berechneEntfernung( punkt1, punkt2 )

            if ( distanz <= 0.005 ):
                if ( self.closing == 1 ):
                    return False
                if ( self.teilstuecke > 0 ):
                    # Schliessen des LinienZuges
                    self.closing = 1
                    self._fuegeLinieEinZwischen( None, self.punkte[ len(self.punkte)-1 ], self.ausgangsPunkt() )
                    self._resetTool()
                    return True
                else:
                    return False
            else:
                self._aktualisiereTeilstuecke(distanz)

                self.storageObjekt.addLine( [punkt1.x(),punkt2.x()], [punkt1.y(),punkt2.y()] )
                self.storageObjekt.appendAttributeData(  [ self.teilstuecke ] )

                if ( self.teilstuecke == 1 ):
                    map.AddLayer(self.layer)
                    self.context.session.AddShapeStore( self.storageObjekt )
                    #self.storageObjekt.bbox = map.layers[self.layerMitPunktdaten].BoundingBox()
                    
                self.layer.issue(LAYER_CHANGED, self.layer)
                return True

    def _erstelleLeereTabelle( self ):
        table = MemoryTable([("Index", FIELDTYPE_INT)],[])
        table.SetTitle( self.layerName )
        return table


    def _resetTool( self ):
        self.dragging = 0

        self.teilstuecke = 0
        self.gesamtLaenge = 0.0

        self.ersterPunkt = None
        self.zweiterPunkt = None

        self.punkte = []

        self.storageObjekt = MemoryShapeStore()
        self.storageObjekt.setTable( self._erstelleLeereTabelle() )

        nummer = self._sucheLayerNummer( self.context.mainwindow.canvas.Map() )
        self.layer = customLayer( _(self.layerName+str(nummer)), self.storageObjekt)
        
        self.closing = 0

    def MouseDown(self, event):
        main_window = self.context.mainwindow
        map = main_window.canvas.Map()
        self.appAnpasser.passeAn( self.context.application )
        
        LineTool.drag_start(self, event.m_x, event.m_y)
        

    def MouseUp(self, event):
        if ( event == self.lastEvent or self.closing ):
            return
        
        self.ersterPunkt = self.zweiterPunkt
        LineTool.MouseUp( self, event)
        
        self.punkte.append( self.zweiterPunkt )
        
        if ( len(self.punkte) > 1 ):
            self._fuegeLinieEinZwischen( event, self.punkte[ len(self.punkte)-2 ], self.punkte[ len(self.punkte)-1 ] )
        if ( self.teilstuecke > 0 ):
            while( self.view._do_redraw() ):
                pass
            self.view.OnLeftDown( event )

    def MouseDown(self, event):
        if ( self.closing ):
            return
        else:
            self.drag_start(event.m_x, event.m_y)

    def drag_start(self, _x, _y):
        if ( not self.dragging ):
            LineTool.drag_start(self, _x, _y)        
            if ( len(self.punkte) == 0 ):
                self.punkte.append( self.ersterPunkt )
            
    def drag_stop(self, x, y):
        if self.view.HasCapture():
            self.view.ReleaseMouse()
        try:
            self.Hide(self.view.drag_dc)
        finally:
            self.view.drag_dc = None
            self.view.dragging = 0

        LineTool.drag_stop( self, x, y )            
# ---------------------------
            
    def drag_move(self, _x, _y):
        if ( self.dragging != 0 ):
            self.dragCounter += 1
            LineTool.drag_move( self, _x, _y )
       
      

