"""
Extend Thuban with a sample Hello World to demonstrate simple
extensions.
"""

__version__ = '$Revision: 1.77 $'
#    $Source: D:/CVS/cvsrepo/thuban/Extensions/precisionFarming/precisionFarming.py,v $
#    $Id: precisionFarming.py,v 1.77 2004/11/04 10:02:16 orahn Exp $

import math
import os

# use _() already now for all strings that may later be translated
from Thuban import _

# Thuban has named commands which can be registered in the central
# instance registry.
from Thuban.UI.command import registry, Command, ToolCommand

# The instance of the main menu of the Thuban application
# See Thuban/UI/menu.py for the API of the Menu class
from Thuban.UI.mainwindow import *
from Thuban.Model.color import *

from Thuban.Model.map import Map

from Thuban.Model.data import *
from data_fake import *
from voronoi import *
from statTests import *
from delaunay import *
from diagram import *
from histogram_berechner import *
from diagrammDialog import *
from shapeStoreVorbereiter import *

from clusteringTool import *
from hoehenlinientool import *
from teilschlagTool import *
from varianzAnalyse import *
from distanzTool import *
from flaechenTool import *
from rasterTool import *
from rasterAttributTool import *
from rasterZellenLoeschenTool import *
from linRegTool import *

from fehler import *
from diagrammHalter import diagrammHalter

from shapeStoreZerleger import *
from shapeStoreBuilder import *
import dbWriter
from dbFactory import *
from logger import *
from time import sleep


from Thuban.Model.table import MemoryTable, \
     FIELDTYPE_DOUBLE, FIELDTYPE_INT, FIELDTYPE_STRING

from Thuban.UI.viewport import Tool, RectTool

from customLayer import *
from customSaveAndLoad import *
from applicationAnpasser import *
from csvReader import *
from shapeSpeicherer import *
from ebenenProjezierer import *

from csv_dialog import *

#from version import versions
from Thuban.version import versions

# schleifen beschleunigen
try:
    from psyco.classes import __metaclass__
except:
    pass

# --------------------------------------------------------

class precisionFarming:
    def __init__(self, registry, main_toolbar, main_menu):
        self.registry = registry
        self.main_toolbar = main_toolbar
        self.main_menu = main_menu

        # menue namen
        self.menuVisual_name = "Anzeigen"
        self.menuProcess_name = "Auswerten"
        self.menuMeasure_name = "Ausmessen"
        self.menuData_name = "Datenhaltung"

        # menue objekte
        self.pirolVisual_menu = self.main_menu.find_menu( self.menuVisual_name)
        self.pirolProcess_menu = self.main_menu.find_menu(self.menuProcess_name)
        self.pirolMeasure_menu = self.main_menu.find_menu(self.menuMeasure_name)
        self.pirolDaten_menu = self.main_menu.find_menu(self.menuData_name)

        # datenhaltung
        self.dbfact = None
        self.diagramSpeicher = diagrammHalter.getDiagrammHalter()
        self.appAnpasser = applicationAnpasser.getApplicationAnpasser()

        # befehle
        self.command_Anpassen = "c_anpassen"

        self.command_Delaunay = "c_delaunay"
        self.command_Voronoi  = "c_voronoi"
        self.command_Farbflaechen = "c_farbFlaechen"
        self.command_Farbbereiche = "c_farbBereiche"
        self.command_hoehenlinien = "c_hoehenlinien"

        self.command_Histogramm = "c_histogram"
        self.command_Chi = "c_chi"
        self.command_Ausreisser = "c_ausreisser"

        self.command_SaveShape = "c_saveShape"
        self.command_Save = "c_save"
        self.command_Load = "c_load"
        self.command_LoadCsv = "c_loadcsv"

        self.command_projEbene = "c_proj"

        self.command_rasterladen = "c_loadRaster"
        self.command_AddAttr = "c_+attr"
        self.command_raster2punkt = "c_r2p"
        self.command_punkt2raster = "c_p2r"
        # werkzeuge
        self.tool_Hoehenlinien = "hoehenlinien_tool"
        self.tool_Teilschlag = "teilschlag_tool"
        self.tool_Cluster = "clustering_tool"
        self.tool_Varianz = "varianz_tool"
        self.tool_LinReg = "linreg_tool"

        self.tool_distanz = "distanz_tool"
        self.tool_flaeche = "flaeche_tool"
        self.tool_raster = "raster_tool"
        self.tool_raster_attribut = "raster_attr_tool"
        self.tool_raster_loeschen = "raster_loeschen_tool"

        self.tool_name_hoehenlinie = "N.A."
        self.tool_name_teilschlag  = "N.A."
        self.tool_name_cluster     = "N.A."
        self.tool_name_varianz     = "N.A."
        self.tool_name_linreg      = "N.A."
        self.tool_name_distanz     = "N.A."
        self.tool_name_flaeche     = "N.A."
        self.tool_name_raster      = "N.A."
        self.tool_name_raster_attribut = "N.A."
        self.tool_name_raster_loeschen = "N.A."


        self._registriereFunktionen()
        self._fuegeFunktionenInDieOberflaecheEin()

        self.versionierungsZusatz = "a"
        global __version__
        versions[ "pirol" ] = __version__[ __version__.find(" ")+1: __version__.rfind(" ") ] + self.versionierungsZusatz

    def _registriereFunktionen( self ):
        # create a new command and register it

        self.registry.Add(Command(self.command_Anpassen, _('Pirol DB Verbindung eintragen'), self.passeApplikationAn,
                             helptext = _('Einfuegen der Standard Pirol Datenbank Verbindung')))

        self.registry.Add(Command(self.command_Delaunay, _('Delaunay'), self._delaunay_func,
                             helptext = _('Zeichne Delaunay Raster')))
        self.registry.Add(Command(self.command_Voronoi, _('Voronoi'), self._voronoi,
                             helptext = _('Zeichne Voronoi Diagram')))

        self.registry.Add(Command(self.command_Farbflaechen, _('Farb-Flaechen'), self._farbFlaechen,
                             helptext = _('Zeichne Farb-Flaechen fuer ein Attribut')))
        self.registry.Add(Command(self.command_Farbbereiche, _('Farb-Bereiche'), self._farbBereiche,
                             helptext = _('Zeichne Farb-Bereiche fuer ein Attribut')))

        self.registry.Add(Command(self.command_hoehenlinien, _('Iso-Linien'), self._hoehenlinien_command,
                             helptext = _('Zeichne \"Hoehenlinien\" zu einem bestimmten Attribut mit einem bestimmten Wert.')))

        self.registry.Add(Command(self.command_Histogramm, _('Histogramm'), self._histogram,
                             helptext = _('Zeichne Histogram fuer ein Attribut')))

        self.registry.Add(Command(self.command_Chi, _('Chi^2 Unabhaengigkeitstest'), self._chiQuadrat_func,
                             helptext = _('Pruefe Unabhaengigkeit zweier Attribute anhand einer chi^2 Verteilung')))

        self.registry.Add(Command(self.command_Ausreisser, _('Ausreisser identifizieren'), self._ausreisserIdentifizieren,
                             helptext = _('Zeige Punkte die mehr als 3mal Standardabweichung vom Mittelwert abweichen')))

        # Datenhaltung

        self.registry.Add(Command(self.command_SaveShape, _('Ebene als .shp speichern'), self._speichereEbeneAlsShapefile,
                     helptext = _('Eine Ebene als Shape-Datei abspeichern, z.B. um die Daten dieser Ebene in anderen Programmen verfuegbar zu machen. (Shape-Dateien koennen von vielen anderen Programmen eingelesen werden.)')))

        self.registry.Add(Command(self.command_projEbene, _('Projektion einer Ebene aendern'), self._projeziereEbene,
                     helptext = _('Die Projektion der Koordinaten einer Ebene aendern')))

        self.registry.Add(Command(self.command_Save, _('Save-Pirol'), self._save,
                     helptext = _('Layer in PirolDB speichern')))
        self.registry.Add(Command(self.command_Load, _('Load-Pirol'), self._load,
                     helptext = _('Layer aus Pirol holen')))
        self.registry.Add(Command(self.command_LoadCsv, _('Load-Pirol aus CSV-Datei'), self._loadcsv,
                     helptext = _('Layer aus CSV-Datei holen')))

        self.registry.Add(Command(self.command_rasterladen, _('Raster laden'), self._rasterEbeneLaden,
                     helptext = _('Laedt Rasterung aus einem Shapefile.')))

        self.registry.Add(Command(self.command_AddAttr, _('Attribut hinzufuegen'), self._fuegeAttributHinzu,
                     helptext = _('Fuegt einem Raster-Layer eines Attribut mit einem initialen Werte hinzu')))

        self.registry.Add(Command(self.command_raster2punkt, _('Punkt- aus Raster-Ebene'), self._PunktEbeneAusRasterEbene,
                     helptext = _('Erstellt eine Punkt-Ebene aus einer Raster-Ebene. Es werden alle Attribute etc. uebernommen.')))

        self.registry.Add(Command(self.command_punkt2raster, _('Punkt- auf Raster-Ebene verdichten'), self._verdichtePunktEbeneAufRasterZellen,
                     helptext = _('Uebertraegt die Daten einer Punkt-Ebene auf die Zellen einer Raster-Ebene')))

        # ----------------------------------------------------------

        iconfile = os.path.join(os.path.abspath(Thuban.__path__[0]),
                                "..", "Extensions", "precisionFarming", "hoehenlinie")

        self.registry.Add(ToolCommand(self.tool_Hoehenlinien, "Iso-Linien Werkzeug",
                                 self._hoehenlinien_tool,
                                 "Zeichnet Aeqipotential-Linen zu einem gewaehlten Attribut", iconfile,
                                 checked = self._check_hoehenlinien_tool,
                                 sensitive = self._sindPunktToolsAnwendbar))

        iconfile = os.path.join(os.path.abspath(Thuban.__path__[0]),
                                "..", "Extensions", "precisionFarming", "teilschlag")

        self.registry.Add(ToolCommand(self.tool_Teilschlag, "Teilschlag Werkzeug",
                                 self._teilschlag_tool,
                                 "Teilschlag auswerten - gibt Ueberblick ueber min./durchn./max. Werte im gewaehlten Bereich", iconfile,
                                 checked = self._check_teilschlag_tool,
                                 sensitive = self._sindPunktToolsAnwendbar))

        iconfile = os.path.join(os.path.abspath(Thuban.__path__[0]),
                                "..", "Extensions", "precisionFarming", "clustering")

        self.registry.Add(ToolCommand(self.tool_Cluster, "Clustering Werkzeug",
                                 self._clustering_tool,
                                 "Clustering", iconfile,
                                 checked = self._check_clustering_tool,
                                 sensitive = self._sindPunktToolsAnwendbar))

        iconfile = os.path.join(os.path.abspath(Thuban.__path__[0]),
                                "..", "Extensions", "precisionFarming", "varianz")

        self.registry.Add(ToolCommand(self.tool_Varianz, "Varianz Analyse",
                                 self._varianz_tool,
                                 "Varianzanalyse", iconfile,
                                 checked = self._check_varianz_tool,
                                 sensitive = self._sindPunktToolsAnwendbar))
        #self.tool_LinReg
        iconfile = os.path.join(os.path.abspath(Thuban.__path__[0]),
                                "..", "Extensions", "precisionFarming", "linreg")

        self.registry.Add(ToolCommand(self.tool_LinReg, "lineare Regression",
                                 self._linreg_tool,
                                 "Eine lineare Regression auf ein gewaehltes Attribut ausfuehren, um evtl. einen Trend ueber die Flaeche zu erkennen.", iconfile,
                                 checked = self._check_linreg_tool,
                                 sensitive = self._sindPunktToolsAnwendbar))

        iconfile = os.path.join(os.path.abspath(Thuban.__path__[0]),
                                "..", "Extensions", "precisionFarming", "distanz")


        self.registry.Add(ToolCommand(self.tool_distanz, "Distanz Werkzeug",
                                 self._distanz_tool,
                                 "Entfernung zwischen zwei Punkten messen", iconfile,
                                 checked = self._check_distanz_tool,
                                 sensitive = self._sindPunktToolsAnwendbar))

        iconfile = os.path.join(os.path.abspath(Thuban.__path__[0]),
                                "..", "Extensions", "precisionFarming", "flaeche")

        self.registry.Add(ToolCommand(self.tool_flaeche, "Flaechen Werkzeug",
                                 self._flaeche_tool,
                                 "Entfernung zwischen zwei oder mehr Punkten messen und Flaecheninhalt des resultierenden Polygons errechnen\n(Schliessen d. Polygons durch Doppel-Klick)", iconfile,
                                 checked = self._check_flaechen_tool,
                                 sensitive = self._sindPunktToolsAnwendbar))

        iconfile = os.path.join(os.path.abspath(Thuban.__path__[0]),
                                "..", "Extensions", "precisionFarming", "raster")

        self.registry.Add(ToolCommand(self.tool_raster, "Raster Werkzeug",
                                 self._raster_tool,
                                 "Legt ein Raster mit der gegebenen Spurbreite ueber eine gewaehlte Punktebene.\nForm und Ausrichtung des Rasters orientieren sich dabei an der, mit diesem Werkzeug zu erstellenden, Flaeche.\nDie erste Linie dieser Flaeche entscheidet ueber die Ausrichtung des Rasters.", iconfile,
                                 checked = self._check_raster_tool,
                                 sensitive = self._sindPunktToolsAnwendbar))

        iconfile = os.path.join(os.path.abspath(Thuban.__path__[0]),
                                "..", "Extensions", "precisionFarming", "raster_attr")

        self.registry.Add(ToolCommand(self.tool_raster_attribut, "Raster Attribut Werkzeug",
                                 self._raster_attribut_tool,
                                 "Aendert die Attribut Werte innerhalb eines Raster-Layers", iconfile,
                                 checked = self._check_raster_attribut_tool,
                                 sensitive = self._sindRasterToolsAnwendbar))

        iconfile = os.path.join(os.path.abspath(Thuban.__path__[0]),
                                "..", "Extensions", "precisionFarming", "rasterLoeschen")

        self.registry.Add(ToolCommand(self.tool_raster_loeschen, "Raster Zellen Loeschen Werkzeug",
                                 self._raster_loeschen_tool,
                                 "Loescht einzelne Raster-Zellen.", iconfile,
                                 checked = self._check_raster_loeschen_tool,
                                 sensitive = self._sindRasterToolsAnwendbar))

    def _fuegeFunktionenInDieOberflaecheEin(self):
        # Add the command to the toolbar
        self.main_toolbar.InsertSeparator()
        self.main_toolbar.InsertItem(self.tool_Hoehenlinien)
        self.main_toolbar.InsertSeparator()
        self.main_toolbar.InsertItem(self.tool_Teilschlag)
        self.main_toolbar.InsertItem(self.tool_Cluster)
        self.main_toolbar.InsertItem(self.tool_Varianz)
        self.main_toolbar.InsertItem(self.tool_LinReg)
        self.main_toolbar.InsertSeparator()
        self.main_toolbar.InsertItem(self.tool_distanz)
        self.main_toolbar.InsertItem(self.tool_flaeche)
        self.main_toolbar.InsertSeparator()
        self.main_toolbar.InsertItem(self.tool_raster)
        self.main_toolbar.InsertItem(self.tool_raster_attribut)
        self.main_toolbar.InsertItem(self.tool_raster_loeschen)


        if self.pirolVisual_menu is None:
            self.pirolVisual_menu = main_menu.InsertMenu(self.menuVisual_name, _(self.menuVisual_name))

        if self.pirolProcess_menu is None:
            self.pirolProcess_menu = main_menu.InsertMenu(self.menuProcess_name,_(self.menuProcess_name))

        if self.pirolMeasure_menu is None:
            self.pirolMeasure_menu = main_menu.InsertMenu(self.menuMeasure_name,_(self.menuMeasure_name))

        if self.pirolDaten_menu is None:
            self.pirolDaten_menu = main_menu.InsertMenu(self.menuData_name,_(self.menuData_name))

        # finally bind the new command with an entry in the extensions menu

        # ------------------------------------------
        self.pirolVisual_menu.InsertItem(self.command_Delaunay)
        self.pirolVisual_menu.InsertItem(self.command_Voronoi)
        self.pirolVisual_menu.InsertSeparator()
        # ------------------------------------------
        self.pirolVisual_menu.InsertItem(self.command_Farbflaechen)
        #self.pirolVisual_menu.InsertItem(self.command_Farbbereiche)
        self.pirolVisual_menu.InsertSeparator()
        # ------------------------------------------
        self.pirolVisual_menu.InsertItem(self.command_hoehenlinien)
        self.pirolVisual_menu.InsertItem(self.tool_Hoehenlinien)
        # ------------------------------------------
        # ------------------------------------------
        self.pirolProcess_menu.InsertItem(self.command_Histogramm)
        self.pirolProcess_menu.InsertItem(self.command_Ausreisser)
        self.pirolProcess_menu.InsertSeparator()
        # ------------------------------------------
        self.pirolProcess_menu.InsertItem(self.command_Chi)
        self.pirolProcess_menu.InsertItem(self.tool_Varianz)
        self.pirolProcess_menu.InsertSeparator()
        # ------------------------------------------
        self.pirolProcess_menu.InsertItem(self.tool_Teilschlag)
        self.pirolProcess_menu.InsertItem(self.tool_Cluster)
        self.pirolProcess_menu.InsertItem(self.tool_LinReg)
        #----------------------------------------------
        #----------------------------------------------
        self.pirolMeasure_menu.InsertItem(self.tool_distanz)
        self.pirolMeasure_menu.InsertItem(self.tool_flaeche)
        #----------------------------------------------
        #----------------------------------------------
        self.pirolDaten_menu.InsertItem(self.command_Anpassen)
        self.pirolDaten_menu.InsertSeparator()
        #----------------------------------------------
        self.pirolDaten_menu.InsertItem(self.command_projEbene)
        self.pirolDaten_menu.InsertSeparator()
        #----------------------------------------------
        self.pirolDaten_menu.InsertItem(self.command_SaveShape)
        self.pirolDaten_menu.InsertSeparator()
        #----------------------------------------------
        self.pirolDaten_menu.InsertItem(self.command_Save)
        self.pirolDaten_menu.InsertItem(self.command_Load)
        self.pirolDaten_menu.InsertItem(self.command_LoadCsv)
        self.pirolDaten_menu.InsertSeparator()
        #----------------------------------------------
        self.pirolDaten_menu.InsertItem(self.command_rasterladen)
        self.pirolDaten_menu.InsertItem(self.tool_raster)
        self.pirolDaten_menu.InsertItem(self.command_AddAttr)
        self.pirolDaten_menu.InsertItem(self.tool_raster_attribut)
        self.pirolDaten_menu.InsertItem(self.command_raster2punkt)
        self.pirolDaten_menu.InsertItem(self.command_punkt2raster)
        self.pirolDaten_menu.InsertItem(self.tool_raster_loeschen)



# ----------------------------------------------------------

    def _hoehenlinien_tool(self, context):
        canvas = context.mainwindow.canvas
        neuesTool = HoehenlinienTool(canvas, context, self._findePunktLayer, self.appAnpasser, self.diagramSpeicher )
        self.tool_name_hoehenlinie = neuesTool.Name()
        canvas.SelectTool(neuesTool)

    def _varianz_tool(self, context):
        canvas = context.mainwindow.canvas
        neuesTool = VarianzAnalysenTool(canvas, context, self._findePunktLayer)
        self.tool_name_varianz = neuesTool.Name()
        canvas.SelectTool(neuesTool)

    def _linreg_tool(self, context):
        canvas = context.mainwindow.canvas
        neuesTool = linRegTool(canvas, context, self._findePunktLayer)
        self.tool_name_linreg = neuesTool.Name()
        canvas.SelectTool(neuesTool)

    def _teilschlag_tool(self, context):
        canvas = context.mainwindow.canvas
        neuesTool = TeilschlagTool(canvas, context, self._findePunktLayer)
        self.tool_name_teilschlag = neuesTool.Name()
        canvas.SelectTool(neuesTool)

    def _clustering_tool(self, context):
        canvas = context.mainwindow.canvas
        neuesTool = ClusteringTool(canvas, context, self._findePunktLayer)
        self.tool_name_cluster = neuesTool.Name()
        canvas.SelectTool(neuesTool)

    def _distanz_tool(self, context):
        canvas = context.mainwindow.canvas
        neuesTool = distanzTool(canvas, context, self._findePunktLayer, self.appAnpasser)
        self.tool_name_distanz = neuesTool.Name()
        canvas.SelectTool(neuesTool)

    def _flaeche_tool(self, context):
        canvas = context.mainwindow.canvas
        neuesTool = flaechenTool(canvas, context, self._findePunktLayer, self.appAnpasser)
        self.tool_name_flaeche = neuesTool.Name()
        canvas.SelectTool(neuesTool)

    def _raster_tool(self, context):
        canvas = context.mainwindow.canvas
        neuesTool = rasterTool(canvas, context, self._findePunktLayer, self.appAnpasser, self._meldeFehler)
        self.tool_name_raster = neuesTool.Name()
        canvas.SelectTool(neuesTool)

    def _raster_attribut_tool(self, context):
        canvas = context.mainwindow.canvas
        neuesTool = rasterAttributTool(canvas, context, self._findeRasterLayer, self.appAnpasser, self._versucheWertZuParsen)
        self.tool_name_raster_attribut = neuesTool.Name()
        canvas.SelectTool(neuesTool)

    def _raster_loeschen_tool(self, context):
        canvas = context.mainwindow.canvas
        neuesTool = rasterZellenLoeschenTool(canvas, context, self._findeRasterLayer, self.appAnpasser )
        self.tool_name_raster_loeschen = neuesTool.Name()
        canvas.SelectTool(neuesTool)

    def _check_tool( self, context, toolName ):
        return context.mainwindow.canvas.CurrentTool() == toolName

    def _check_varianz_tool(self, context):
        return self._check_tool( context, self.tool_name_varianz )

    def _check_linreg_tool(self, context):
        return self._check_tool( context, self.tool_name_linreg )

    def _check_teilschlag_tool(self, context):
        return self._check_tool( context, self.tool_name_teilschlag )

    def _check_hoehenlinien_tool(self, context):
        return self._check_tool( context, self.tool_name_hoehenlinie )

    def _check_clustering_tool(self, context):
        return self._check_tool( context, self.tool_name_cluster )

    def _check_distanz_tool(self, context):
        return self._check_tool( context, self.tool_name_distanz )

    def _check_flaechen_tool(self, context):
        return self._check_tool( context, self.tool_name_flaeche )

    def _check_raster_tool(self, context):
        return self._check_tool( context, self.tool_name_raster )

    def _check_raster_attribut_tool(self, context):
        return self._check_tool( context, self.tool_name_raster_attribut )

    def _check_raster_loeschen_tool(self, context):
        return self._check_tool( context, self.tool_name_raster_loeschen )
# ----------------------------------------------------------

    def _sindPunktToolsAnwendbar(self, context ):
        resultat = self._findePunktLayer( context, True )

        return ( resultat > -1 )

    def _sindRasterToolsAnwendbar(self, context ):
        resultat = self._findeRasterLayer( context, True )

        return ( resultat > -1 )

    def _findePunktLayer(self, context, justCheck = False ):
        return self._findeLayer( context, justCheck, justPointLayer = True )

    def _findeLayer(self, context, justCheck = False, justPointLayer = True, noCustomLayer = True ):
        main_window = context.mainwindow
        map = main_window.canvas.Map()
        layerMitPunktdaten = []

        if map.HasLayers() == True:
            for i in range( len(map.layers)):
                if ( isinstance(map.layers[i],RasterLayer) ):
                    continue
                if (map.layers[i].ShapeStore().ShapeType() == SHAPETYPE_POINT or justPointLayer==False) and \
                    ( (not isinstance(map.layers[i], customLayer) or noCustomLayer==False )or \
                        (isinstance(map.layers[i], customLayer) and \
                         map.layers[i].enthaeltAusgangsdaten() ) \
                    ) and \
                    map.layers[i].Visible():
                    layerMitPunktdaten.append( i )

        if ( len(layerMitPunktdaten)==0 ):
            return -1
        elif ( justCheck or len(layerMitPunktdaten)==1):
            return layerMitPunktdaten[0]
        else:
            TitelZuIndexen = {}
            message = "In der aktuellen Session werden mehrere Ebenen angezeigt, die sich zur Weiterverarbeitung eingnen.\nBitte waehlen Sie eins aus der Liste!"
            text = "gewuenschte Ebene"
            for index in layerMitPunktdaten:
                TitelZuIndexen[map.layers[index].Title()] = index

            dialog_titel = _("Bitte waehlen Sie eine Ebene aus der Liste")
            v_dialog = attributAuswahl( main_window, TitelZuIndexen.keys(), dialog_titel, message = message, text = text )
            userEingabe = v_dialog.RunDialog()

            if userEingabe == None:
                main_window.RunMessageBox(_('Standardauswahl!'), _('Auswahl abgebrochen!\nNehme die erst-beste Ebene!'))
                return 0

            while not TitelZuIndexen.has_key(userEingabe["attr"]):
                v_dialog = attributAuswahl( main_window, colNames, dialog_titel, message = _("diese Ebene existiert nicht"), text = text )

            return TitelZuIndexen[ userEingabe["attr"] ]

    def _speichereEbeneAlsShapefile( self, context ):
        print "Ebene als .shp speichern..."
        main_window = context.mainwindow
        map = main_window.canvas.Map()

        self.appAnpasser.passeAn( context.application )

        if map.HasLayers() == True:
            layerMitDaten = self._findeLayer(context, justCheck=False, justPointLayer=False, noCustomLayer=False)
            fileChooser = wxFileDialog( main_window, message = "Speichern - Bitte geben Sie den Dateinamen an.", defaultDir = "", defaultFile = "", wildcard = "*.shp", style = wxSAVE | wxOVERWRITE_PROMPT , pos = (-1,-1) )
            whatHappened = fileChooser.ShowModal()

            if ( whatHappened == wxID_OK ):
                speicherOnkel = shapeFileSpeicherer( fileChooser.GetPath(), map.layers[layerMitDaten] )
                speicherOnkel.speichern()
            else:
                print "Abbruch"

            fileChooser.Destroy()

        else:
            self._keinPunktDatenLayer( main_window )
# ----------------------------------------------------------
    def _keinPunktDatenLayer( self, main_window ):
        main_window.RunMessageBox(_('missing base data layer'), _('Es wurde kein Layer, welches Punktdaten enthaelt gefunden.'))

    def _keinRasterDatenLayer( self, main_window ):
        main_window.RunMessageBox(_('missing base data layer'), _('Es wurde kein Layer, welches eine Rasterung vorgibt gefunden.'))
# ----------------------------------------------------------
# ---------------- RasterEbenen Funktionen -----------------
    def _findeRasterLayer(self, context, justCheck = False ):
        return self._findeRLayer( context, justCheck )

    def _findeRLayer(self, context, justCheck = False ):
        main_window = context.mainwindow
        map = main_window.canvas.Map()
        layerMitRasterdaten = []

        if map.HasLayers() == True:
            for i in range( len(map.layers)):
                if ( isinstance(map.layers[i],RasterLayer) ):
                    continue
                if (map.layers[i].ShapeStore().ShapeType() == SHAPETYPE_POLYGON and \
                        isinstance(map.layers[i], customLayer) and \
                         map.layers[i].enthaeltRasterDaten()  \
                    ) and \
                    map.layers[i].Visible():
                    layerMitRasterdaten.append( i )

        if ( len(layerMitRasterdaten)==0 ):
            return -1
        elif ( justCheck or len(layerMitRasterdaten)==1):
            return layerMitRasterdaten[0]
        else:
            TitelZuIndexen = {}
            message = "In der aktuellen Session werden mehrere Ebenen angezeigt, die sich zur Weiterverarbeitung eingnen.\nBitte waehlen Sie eins aus der Liste!"
            text = "gewuenschte Ebene"
            for index in layerMitRasterdaten:
                TitelZuIndexen[map.layers[index].Title()] = index

            dialog_titel = _("Bitte waehlen Sie eine Ebene aus der Liste")
            v_dialog = attributAuswahl( main_window, TitelZuIndexen.keys(), dialog_titel, message = message, text = text )
            userEingabe = v_dialog.RunDialog()

            if userEingabe == None:
                main_window.RunMessageBox(_('Standardauswahl!'), _('Auswahl abgebrochen!\nNehme die erst-beste Ebene!'))
                return 0

            while not TitelZuIndexen.has_key(userEingabe["attr"]):
                v_dialog = attributAuswahl( main_window, colNames, dialog_titel, message = _("diese Ebene existiert nicht"), text = text )

            return TitelZuIndexen[ userEingabe["attr"] ]


    def _fuegeAttributHinzu( self, context ):
        main_window = context.mainwindow
        map = main_window.canvas.Map()

        if map.HasLayers() == True:
            layerMitRasterdaten = self._findeRasterLayer(context)

            if layerMitRasterdaten < 0:
                self._keinRasterDatenLayer( main_window )
                return

            rasterLayer = map.layers[layerMitRasterdaten]

            #( parent, parameters, titel, message = "", txtZahl="", txtText1="", txtText2=""):
            dialog_titel = _("Angaben zum neuen Attribut")
            v_dialog = doppelAttributAuswahl( main_window, ["FIELDTYPE_INT","FIELDTYPE_DOUBLE","FIELDTYPE_STRING"], dialog_titel, txtZahl="Basis Wert", txtText2="Attribut Typ", txtText1="Attribut Name" )
            userEingabe = v_dialog.RunDialog()

            if userEingabe == None:
                return

            while self._versucheWertZuParsen( userEingabe["yattr"],userEingabe["alpha"] ) == None:
                v_dialog = doppelAttributAuswahl( main_window, ["FIELDTYPE_INT","FIELDTYPE_DOUBLE","FIELDTYPE_STRING"], dialog_titel, _("Der Typ stimmt nicht mit dem default-Wert ueberein!"), txtZahl="Basis Wert", txtText2="Attribut Typ", txtText1="Attribut Name" )

                userEingabe = v_dialog.RunDialog()
                if userEingabe == None:
                    return

            type = eval(userEingabe["yattr"])
            attrName = userEingabe["xattr"]

            defaultValue = self._versucheWertZuParsen( userEingabe["yattr"],userEingabe["alpha"] )
            rasterLayer.ShapeStore().appendAttributeDataColumn( attrName, type, defaultValue )

            rasterLayer.SetClassification( None )

            rasterLayer.SetClassificationColumn(attrName)

            classif = rasterLayer.GetClassification()

            properties = ClassGroupProperties()
            properties.SetFill( Color(0,1.0,0) )

            classif.AppendGroup( ClassGroupSingleton( value = defaultValue, props = properties, label = "Attr.: "+attrName+" "+str(defaultValue), group = None) )



    def _versucheWertZuParsen( self, typStr, wert ):
        if (typStr!=FIELDTYPE_INT and typStr!=FIELDTYPE_DOUBLE and typStr!=FIELDTYPE_STRING):
            typ = eval(typStr)
        else:
            typ = typStr
        if (typ==FIELDTYPE_INT and wert.isdigit()):
            try:
                retval = int( wert )
                return retval
            except:
                return None
        elif (typ==FIELDTYPE_DOUBLE and istZahl(wert)):
            try:
                retval = float( wert )
                return retval
            except:
                return None
        elif (typ==FIELDTYPE_STRING):
            # eigentlich unnoetig
            try:
                retval = str( wert )
                return retval
            except:
                return None

        return None

    def _verdichtePunktEbeneAufRasterZellen( self, context ):
        main_window = context.mainwindow
        map = main_window.canvas.Map()
        debug = False


        if map.HasLayers() == True:
            layerMitPunktdaten = self._findePunktLayer(context)

            if layerMitPunktdaten < 0:
                self._keinPunktDatenLayer( main_window )
                return

            punktLayer = map.layers[layerMitPunktdaten]

            layerMitRasterdaten = self._findeRasterLayer(context)

            if layerMitRasterdaten < 0:
                self._keinRasterDatenLayer( main_window )
                return

            rasterLayer = map.layers[layerMitRasterdaten]

            polygone = rasterLayer.getPolygone()

            if ( polygone == None ):
                print "FEHLER: polygon feld ist NULL"
                return

            print len(polygone),"polygone erhalten"

            # punkte polygonen zuordnen ...
            polyIdZuPunktefeld = {}
            allePunkte = punktLayer.ShapeStore().AllShapes()
            punktefeld = []

            for shp in allePunkte:
                x,y = shp.Points()[0][0]
                punktefeld.append( punkt( [x, y], shp.ShapeID() ) )

            mengePunkte = float(len(punktefeld))
            markiert = 0.
            for poly in polygone:
                minX,minY,maxX,maxY = poly.BoundingBox()
                pktefeld = []
                for pkt in punktefeld:
                    if ( pkt.markiert ):
                        continue
#                    if ( pkt.x() < minX or pkt.x() > maxX or pkt.y() < minY or pkt.y() > maxY ):
#                        continue
                    if ( poly.liegtPunktDrin( pkt ) ):
                        pktefeld.append( pkt )
                        pkt.markiere()
                        markiert += 1.0

                polyIdZuPunktefeld[ poly.getIndex() ] = pktefeld
                if ( debug ):
                    print repr(markiert/mengePunkte*100)[:4],"Prozent zugeordnet"

            print "punkte zugeordnet"

            # Attribute uebernehmen
            columns = punktLayer.ShapeStore().Table().Columns()
            rasterStore = rasterLayer.ShapeStore()

            dialog_titel = _("Angaben zum Histogramm")
            message = "Gesucht ist der Wert zu einem Attribut, der bei Zellen eingesetzt werden soll,\nin die kein Punkt der Punkt-Ebene faellt.\nKlicken Sie auf Abbrechen, um das betreffende Attribut nicht zu uebernehmen"
            text = "Null Wert:"

            gewuenschteAttribute = []

            for column in columns:
                attrString = "\n\nATTRIBUT: "+column.name
                defaultWert = None
                if ( column.type==FIELDTYPE_INT ):
                    defaultWert = "0"
                elif ( column.type==FIELDTYPE_DOUBLE ):
                    defaultWert = "0.0"
                elif ( column.type==FIELDTYPE_STRING ):
                    defaultWert = "NULL"

                v_dialog = einzelAbfrageDialog( main_window, dialog_titel, message=message+attrString, defaultWert=defaultWert, text=text )
                userEingabe = v_dialog.RunDialog()

                if userEingabe == None:
                    continue

                while ( not self._passtWertZuTyp( userEingabe["wert"], column.type ) ):
                    v_dialog = einzelAbfrageDialog( main_window, dialog_titel, message=message+"\n\nWert nicht zulaessig fuer Attribut"+attrString, defaultWert=defaultWert, text=text )
                    userEingabe = v_dialog.RunDialog()

                    if userEingabe == None:
                        continue

                defaultWert = self._versucheWertZuParsen( column.type, userEingabe["wert"] )
                gewuenschteAttribute.append( ( column.name, column.type, defaultWert ) )
                #rasterStore.appendAttributeDataColumn( column.name, column.type, defaultValue=defaultWert )

            for tupel in gewuenschteAttribute:
                name, typ, wert = tupel
                rasterStore.appendAttributeDataColumn( name, typ, wert )

            # werte mitteln, bzw. modus bilden
            werteZugeordnet = 0.
            for k in polyIdZuPunktefeld.keys():
                werteZugeordnet += 1.
                shpId = int(k)
                punktefeld = polyIdZuPunktefeld[k]

                if ( len(punktefeld) == 0 ):
                    continue

                werte = {}

                for pkt in punktefeld:
                    row = punktLayer.ShapeStore().Table().ReadRowAsDict( pkt.index )

                    for tupel in gewuenschteAttribute:
                        name, typ, wert = tupel
                        column = punktLayer.ShapeStore().Table().Column( name )
                        if ( column.type==FIELDTYPE_INT or column.type==FIELDTYPE_DOUBLE ):
                            if ( werte.has_key( column.name ) ):
                                werte[column.name] += row[column.name]
                            else:
                                werte[column.name] = row[column.name]
                        else:
                            if ( werte.has_key( column.name ) ):
                                if (werte[column.name].has_key( row[column.name] ) ):
                                    werte[column.name][row[column.name]] += 1
                                else:
                                    werte[column.name][row[column.name]] = 1
                            else:
                                werte[column.name] = {}
                                werte[column.name][row[column.name]] = 1

                for tupel in gewuenschteAttribute:
                    name, typ, wert = tupel
                    column = punktLayer.ShapeStore().Table().Column( name )
                    if ( column.type==FIELDTYPE_INT ):
                        werte[column.name] /= len(punktefeld)
                    elif ( column.type==FIELDTYPE_DOUBLE ):
                        werte[column.name] /= float(len(punktefeld))
                    else:
                        maxZahl = 0
                        aktGewinner = None
                        for wert in werte[column.name].keys():
                            if ( werte[column.name][wert] > maxZahl ):
                                maxZahl = werte[column.name][wert]
                                aktGewinner = wert

                        werte[column.name] = aktGewinner

                for colName in werte.keys():
                    rasterStore.setAttributeValue( shpId, colName, werte[colName] )

                if ( debug ):
                    print repr(werteZugeordnet/float(len(polyIdZuPunktefeld.keys())))[:4], "% der Zellen mit Werten gefuellt"

        print "fertig!"


    def _passtWertZuTyp( self, wert, typ ):
        if (typ==FIELDTYPE_STRING):
            return True
        elif ( istZahl(wert) ):
            if (typ==FIELDTYPE_DOUBLE):
                return True
            elif ( wert.isdigit() and wert.find(".") == -1 and typ==FIELDTYPE_INT ):
                return True

        return False

    def _rasterEbeneLaden( self, context ):
        main_window = context.mainwindow
        map = main_window.canvas.Map()

        self.appAnpasser.passeAn( context.application )

        fileChooser = wxFileDialog( main_window, message = "Speichern - Bitte geben Sie den Dateinamen an.", defaultDir = "", defaultFile = "", wildcard = "*.shp", style = wxOPEN , pos = (-1,-1) )
        whatHappened = fileChooser.ShowModal()

        if ( whatHappened != wxID_OK ):
            print "Abbruch"
            fileChooser.Destroy()
            return

        print fileChooser.GetPath()

        newStore = ShapefileStore( context.session, fileChooser.GetPath() )

        if ( newStore.ShapeType() != SHAPETYPE_POLYGON ):
            self._keinRasterDatenLayer( main_window )
            return

        memStore = MemoryShapeStore.getMemCopyOfShapeStore( newStore )

        rasterLayer = customLayer( _("Raster ("+fileChooser.GetFilename()+")"), memStore, typenMap.getRasterTyp() )

        rasterLayer.ShapeStore().Table().SetTitle( _("Raster ("+fileChooser.GetFilename()+")") )
        map.AddLayer(rasterLayer)
        context.session.AddShapeStore( rasterLayer.ShapeStore() )

        main_window.canvas.FitMapToWindow()

        fileChooser.Destroy()

    def _PunktEbeneAusRasterEbene( self, context ):
        main_window = context.mainwindow
        map = main_window.canvas.Map()

        if map.HasLayers() == True:
            layerMitRasterdaten = self._findeRasterLayer(context)

            if layerMitRasterdaten < 0:
                self._keinRasterDatenLayer( main_window )
                return

            rasterLayer = map.layers[layerMitRasterdaten]
            data = MemoryShapeStore.getMemCopyOfShapeStore(rasterLayer.ShapeStore())
            data.shapetype = SHAPETYPE_POINT

            for shape in data.AllShapes():
                points_alt = shape.Points()[0]
                x_quer = .0
                y_quer = .0
                for punkteTuple in points_alt[1:]:
                    x,y = punkteTuple
                    x_quer += x
                    y_quer += y

                x_quer /= len(points_alt)-1
                y_quer /= len(points_alt)-1

                shape.SetPoints( [ [(x_quer,y_quer)] ]  )

            layerTitel = rasterLayer.Title()+" (Punkt-Ebene)"

            data.setTitle(layerTitel)

            punktLayer = customLayer( layerTitel, data, typ = typenMap.getExRasterTyp(),
                                     projection = rasterLayer.GetProjection(),
                                     fill = rasterLayer.GetClassification().GetDefaultFill(),
                                     stroke = Black,
                                     lineWidth = rasterLayer.GetClassification().GetDefaultLineWidth(),
                                     visible = rasterLayer.Visible())

            punktLayer.SetClassification(rasterLayer.GetClassification())

            punktLayer.SetClassificationColumn(rasterLayer.GetClassificationColumn())

            context.session.AddShapeStore( punktLayer.ShapeStore() )
            map.AddLayer(punktLayer)

# ----------------------------------------------------------

    def _projeziereEbene( self, context ):
        main_window = context.mainwindow
        map = main_window.canvas.Map()


        if map.HasLayers() == True:
            layerMitPunktdaten = self._findePunktLayer(context)

            if layerMitPunktdaten < 0:
                self._keinPunktDatenLayer( main_window )
                return

            oldLayer = map.layers[layerMitPunktdaten]


            dialog_titel = _("Ausgangs-Projektion waehlen")
            v_dialog = diagramm_dialog( main_window, projmap.keys(), dialog_titel, message="Ziel Projektion ist Gauss-Krueger (German Grid Zone 3)\n(Hinweis: Ein zu geringer Mindest Abstand kann dazu fuehren, dass Voronoi- (z.B. Farbflaechen)\noder Delaunay- Diagramme (Iso-Linien) nicht berechnet werden koennen.\nDie Zahl -1 schaltet den Filter ab.)", textAttr="Ausgangs-Projektion", textKat="Mindest Abstand zweier Punkte (m)"  )
            userEingabe = v_dialog.RunDialog()

            if userEingabe == None:
                return

            while projmap.has_key( userEingabe["attr"] )==False or not istZahl(userEingabe["numKath"]):
                if (projmap.has_key( userEingabe["attr"] )==False):
                    v_dialog = diagramm_dialog( main_window, projmap.keys(), dialog_titel, _("Diese Projektion ist unbekannt"), textAttr="Ausgangs-Projektion", textKat="Mindest Abstand zweier Punkte (m)" )
                elif (not istZahl(userEingabe["numKath"])):
                    v_dialog = diagramm_dialog( main_window, projmap.keys(), dialog_titel, _("Bitte geben Sie eine ZAHL ein"), textAttr="Ausgangs-Projektion", textKat="Mindest Abstand zweier Punkte (m)" )
                userEingabe = v_dialog.RunDialog()
                if userEingabe == None:
                    return

            projIn = projmap[userEingabe["attr"]]

            projezierer = ebenenProjezierer(projIn , oldLayer, filterAbst=float(userEingabe["numKath"]))
            newLayer = projezierer.erstelleProjeziertesLayer()

            if projezierer.fehlerIstAufgetreten():
                self._meldeFehler( main_window, projezierer.getFehler() )
                return

            map.RemoveLayer( oldLayer )
            #map.SetProjection(projezierer.getOutProj())

            context.session.AddShapeStore( newLayer.ShapeStore() )
            map.AddLayer(newLayer)

            main_window.canvas.map_projection_changed(map, None)

            main_window.canvas.FitMapToWindow()

    def _histogram(self, context):
        """
        Zeichnen eines Histogramms
        """
        main_window = context.mainwindow
        map = main_window.canvas.Map()


        if map.HasLayers() == True:
            layerMitPunktdaten = self._findePunktLayer(context)

            if layerMitPunktdaten < 0:
                self._keinPunktDatenLayer( main_window )
                return

            attrTable = map.layers[layerMitPunktdaten].ShapeStore().Table()
            cols = attrTable.Columns()
            colNames = []

            for i in cols:
                #if i.type == FIELDTYPE_INT or i.type == FIELDTYPE_DOUBLE:
                colNames.append( i.name )

            werteSindDiskret = False
            dialog_titel = _("Angaben zum Histogramm")
            v_dialog = diagramm_dialog( main_window, colNames, dialog_titel )
            userEingabe = v_dialog.RunDialog()

            if userEingabe == None:
                return

            aktCol = attrTable.Column( userEingabe["attr"] )
            if aktCol.type != FIELDTYPE_INT and aktCol.type != FIELDTYPE_DOUBLE:
                werteSindDiskret = True

            while attrTable.HasColumn( userEingabe["attr"] )==False or (userEingabe["numKath"].isdigit()==False and werteSindDiskret != True):
                if attrTable.HasColumn( userEingabe["attr"] )==False:
                    v_dialog = diagramm_dialog( main_window, colNames, dialog_titel, _("dieses Attribut existiert nicht") )
                else:
                    aktCol = attrTable.Column( userEingabe["attr"] )
                    if aktCol.type != FIELDTYPE_INT and aktCol.type != FIELDTYPE_DOUBLE:
                        werteSindDiskret = True
                if userEingabe["numKath"].isdigit()==False and werteSindDiskret != True:
                    v_dialog = diagramm_dialog( main_window, colNames, dialog_titel, _("bitte geben Sie eine Zahl ein") )

                userEingabe = v_dialog.RunDialog()
                if userEingabe == None:
                    return

            werteTabelle = []

            storeVorbereiter = shapeStoreVorbereiter( map.layers[layerMitPunktdaten].ShapeStore() )
            punktefeld = storeVorbereiter.getRohPunktefeld()

            for i in punktefeld:
                werteTabelle.append( attrTable.ReadValue( i.index, userEingabe["attr"] ) )

            if werteSindDiskret == True:
                histBer = histogramBerechner( 2, werteTabelle, werteSindDiskret )
            else:
                histBer = histogramBerechner( int(userEingabe["numKath"]), werteTabelle )
            histBer.erstelleHistogram()

            if werteSindDiskret != True:
                diagramMaler = balken_diagramZeichner( main_window, map.layers[layerMitPunktdaten], _("Histogram/Mengenverteilung fuer "+userEingabe["attr"]), ((histBer.minWert,histBer.maxWert), (0,histBer.maxMenge)), histBer.getBalkenwerte()  )
            else:
                diagramMaler = balken_diagramZeichner( main_window, map.layers[layerMitPunktdaten], _("Histogram/Mengenverteilung fuer "+userEingabe["attr"]), ((histBer.minWert,histBer.maxWert), (0,histBer.maxMenge)), histBer.getBalkenwerte(), True, histBer.getColNames() )
            diagramMaler.RunDialog()
            print "- zeichne histogram"


        else:
            main_window.RunMessageBox(_('missing base data layer'), _('Please add a layer consisting of point data, first'))

# ----------------------------------------------------------

    def _hoehenlinien_command(self, context):
        """
        Zeichnen eines Histograms
        """
        main_window = context.mainwindow

        canvas = context.mainwindow.canvas
        neuesTool = HoehenlinienTool(canvas, context, self._findePunktLayer, self.appAnpasser, self.diagramSpeicher )
        neuesTool.useAsCommand()

    def _ausreisserIdentifizieren(self, context):
        """
        Zeichnen eines Histograms
        """
        main_window = context.mainwindow
        map = main_window.canvas.Map()


        if map.HasLayers() == True:
            layerMitPunktdaten = self._findePunktLayer(context)

            if layerMitPunktdaten < 0:
                self._keinPunktDatenLayer( main_window )
                return

            attrTable = map.layers[layerMitPunktdaten].ShapeStore().Table()
            cols = attrTable.Columns()
            colNames = []

            for i in cols:
                if i.type == FIELDTYPE_INT or i.type == FIELDTYPE_DOUBLE:
                    colNames.append( i.name )

            werteSindDiskret = False
            dialog_titel = _("Angaben zur Ausreisser-Identifikation")
            v_dialog = attributAuswahl( main_window, colNames, dialog_titel )
            userEingabe = v_dialog.RunDialog()

            if userEingabe == None:
                return

            aktCol = attrTable.Column( userEingabe["attr"] )
            if aktCol.type != FIELDTYPE_INT and aktCol.type != FIELDTYPE_DOUBLE:
                werteSindDiskret = True

            while attrTable.HasColumn( userEingabe["attr"] )==False or werteSindDiskret == True:
                if attrTable.HasColumn( userEingabe["attr"] )==False:
                    v_dialog = attributAuswahl( main_window, colNames, dialog_titel, _("dieses Attribut existiert nicht") )
                elif werteSindDiskret == True:
                    v_dialog = attributAuswahl( main_window, colNames, dialog_titel, _("bitte waehlen Sie ein numerisches Attribut") )
                    werteSindDiskret = False

                userEingabe = v_dialog.RunDialog()
                if userEingabe == None:
                    return
                else:
                    aktCol = attrTable.Column( userEingabe["attr"] )
                    if aktCol.type != FIELDTYPE_INT and aktCol.type != FIELDTYPE_DOUBLE:
                        werteSindDiskret = True

            werteTabelle = []

            storageObjekt = MemoryShapeStore()

            storageObjekt.bbox = map.layers[layerMitPunktdaten].BoundingBox()

            storeVorbereiter = shapeStoreVorbereiter( map.layers[layerMitPunktdaten].ShapeStore(), storageObjekt )

            gesuchtesAttribut = userEingabe["attr"]
            storeVorbereiter.setzeAuswertungsAttribut( gesuchtesAttribut )

            gibtAusreisser = storeVorbereiter.identifiziereAusreisser()

            if gibtAusreisser == False:
                main_window.RunMessageBox(_('keine Ausreisser'), _('unter den Attributwerten ('+gesuchtesAttribut+') zeigen sich keine Ausreisser.'))
                return

            layerName = "Ausreisser "+gesuchtesAttribut
            layer = customLayer( _(layerName), storageObjekt)
            layer.SetClassificationColumn( storeVorbereiter.auswertungsAttribut )

            layer.SetClassification(storeVorbereiter.classification)

            storageObjekt.setTitle( layerName )

            map.AddLayer(layer)
            context.session.AddShapeStore( storageObjekt )

            context.session.WasModified()

            print "- identifiziere Ausreisser"


        else:
            main_window.RunMessageBox(_('missing base data layer'), _('Please add a layer consisting of point data, first'))

    # ------------------------------------------------------------

    def _chiQuadrat_func(self, context ):
        main_window = context.mainwindow
        map = main_window.canvas.Map()

        if map.HasLayers() == True:
            layerMitPunktdaten = self._findePunktLayer(context)

            if layerMitPunktdaten < 0:
                self._keinPunktDatenLayer( main_window )
                return

            attrTable = map.layers[layerMitPunktdaten].ShapeStore().Table()
            cols = map.layers[layerMitPunktdaten].ShapeStore().Table().Columns()
            colNames = []

            for i in cols:
                #if i.type == FIELDTYPE_INT or i.type == FIELDTYPE_DOUBLE:
                colNames.append( i.name )

            diagram_titel = _("Angaben zum Chi-Unabhaengigkeitstest")
            werteSindDiskret = False

            v_dialog = doppelAttributAuswahl( main_window, colNames, diagram_titel )
            userEingabe = v_dialog.RunDialog()
            if userEingabe == None:
                return

            while (attrTable.HasColumn( userEingabe["xattr"] )==False or attrTable.HasColumn( userEingabe["yattr"] )==False) or userEingabe["alpha"].isdigit() == False:
                if attrTable.HasColumn( userEingabe["xattr"] )==False:
                    v_dialog = doppelAttributAuswahl( main_window, colNames, diagram_titel, _("das 1. Attribut existiert nicht") )
                elif attrTable.HasColumn( userEingabe["yattr"] )==False:
                    v_dialog = doppelAttributAuswahl( main_window, colNames, diagram_titel, _("das 2. Attribut existiert nicht") )
                else:
                    v_dialog = doppelAttributAuswahl( main_window, colNames, diagram_titel, _("unzulaessiger Wert fuer die Signifikanz") )
                userEingabe = v_dialog.RunDialog()
                if userEingabe == None:
                    return

            werte = []
            attribute = [ userEingabe["xattr"], userEingabe["yattr"] ]

            shapeStoreVorber = shapeStoreVorbereiter( map.layers[layerMitPunktdaten].ShapeStore() )
            punktefeld = shapeStoreVorber.getRohPunktefeld()

            for j in range( len(attribute)):
                werte.append([])

            for i in range( len(punktefeld) ):
                for j in range( len(attribute)):
                    werte[j].append( attrTable.ReadValue( punktefeld[i].index, attribute[j] ) )

            kathNamen1 = []
            aktCol = attrTable.Column( userEingabe["xattr"] )
            if aktCol.type != FIELDTYPE_INT and aktCol.type != FIELDTYPE_DOUBLE:
                for i in werte[0]:
                    if not ( i in kathNamen1 ):
                        kathNamen1.append(i)
                minAttr1, maxAttr1 = (0,0)
            else:
                minAttr1, maxAttr1 = attrTable.ValueRange( attribute[0] )

            kathNamen2 = []
            aktCol = attrTable.Column( userEingabe["yattr"] )
            if aktCol.type != FIELDTYPE_INT and aktCol.type != FIELDTYPE_DOUBLE:
                for i in werte[1]:
                    if not ( i in kathNamen2 ):
                        kathNamen2.append(i)
                minAttr2, maxAttr2 = (0,0)
            else:
                minAttr2, maxAttr2 = attrTable.ValueRange( attribute[1] )

            if kathNamen1 == [] and kathNamen2 == []:
                chiQuadratTest = ChiQuadratTest( werte, len(punktefeld), [minAttr1,minAttr2], [maxAttr1,maxAttr2] )
            else:
                chiQuadratTest = ChiQuadratTest( werte, len(punktefeld), [minAttr1,minAttr2], [maxAttr1,maxAttr2], {"0":kathNamen1,"1":kathNamen2} )

            D = chiQuadratTest.getPruefsumme()
            wahrsch = chiQuadratTest.signifikanzZurPruefsumme()
            print "D:", D, ", Wahrscheinlichk.", wahrsch
            print "freiheitsgrade:", chiQuadratTest.getFreiheitsgrade(), chiQuadratTest.getXYFreiheitsgrade()

            attribut_string = "("+userEingabe["xattr"]+", "+userEingabe["yattr"]+")"

            alpha = (float(userEingabe["alpha"])/100.0)

            if chiQuadratTest.fehlerIstAufgetreten():
                self._meldeFehler( main_window, chiQuadratTest.getFehler() )
                return

            istWiderlegt = "nein"
            if (chiQuadratTest.signifikanzZurPruefsumme() <= alpha):
                istWiderlegt = "ja"
            anzeiger = allgTabellenAnzeige( main_window, [[ alpha ],[ repr(chiQuadratTest.getPruefsumme())[:8] ], [repr(chiQuadratTest.signifikanzZurPruefsumme())[:8]], [chiQuadratTest.getFreiheitsgrade()], [istWiderlegt ]], [ "Signifikanz", "Pruefsumme D(n)", "P(D(n)<x)", "Freiheitsgrade", "Unabh. widerlegt" ], ["wert"], attribut_string+" Chi-Quadrat Unabhaengigkeitstest" )
            anzeiger.RunDialog()


    # ------------------------------------------------------------

    def _delaunay_func(self, context):
        """
        Zeichnen eines Delaunay-Rasters
        """
        main_window = context.mainwindow
        map = main_window.canvas.Map()

        self.appAnpasser.passeAn( context.application )

        if map.HasLayers() == True:
            layerMitPunktdaten = self._findePunktLayer(context)

            if layerMitPunktdaten < 0:
                self._keinPunktDatenLayer( main_window )
                return

            table = MemoryTable([("type", FIELDTYPE_STRING),
                                 ("dreiecknummer", FIELDTYPE_DOUBLE),
                                 ("seite", FIELDTYPE_INT)],[])
            storageObjekt = MemoryShapeStore()
            storageObjekt.setTable( table )
            storageObjekt.bbox = map.layers[layerMitPunktdaten].BoundingBox()

            storeVorbereiter = shapeStoreVorbereiter( map.layers[layerMitPunktdaten].ShapeStore(), storageObjekt )

            Delaunay = self.diagramSpeicher.getDelaunayRaster( map.layers[layerMitPunktdaten], storeVorbereiter.getNormiertesPunktefeld(), context )

            if Delaunay.fehlerIstAufgetreten():
                self._meldeFehler( main_window, Delaunay.getFehler() )
                return

            storeVorbereiter.zeichneDiagram( Delaunay.getGeordnetesPunktefeld(), Delaunay.getKantenfeld() )

            titel = _("Delaunay Netz")
            layer = customLayer( titel, storageObjekt, typ=typenMap.getDelaunayTyp())
            classification = layer.GetClassification()
            classification.SetDefaultFill(Color(1,0,0))
            classification.SetDefaultLineColor(Color(1,0,0))
            layer.SetClassification(classification)

            storageObjekt.setTitle( titel )
            map.AddLayer(layer)
            context.session.AddShapeStore( storageObjekt )

        else:
            main_window.RunMessageBox(_('missing base data layer'), _('Please add a layer consisting of point data, first'))


    def _voronoi(self, context):
        """
        Zeichnen eines Delaunay-Rasters
        """
        main_window = context.mainwindow
        map = main_window.canvas.Map()

        self.appAnpasser.passeAn( context.application )

        if map.HasLayers() == True:
            layerMitPunktdaten = self._findePunktLayer(context)

            if layerMitPunktdaten < 0:
                self._keinPunktDatenLayer( main_window )
                return


            storageObjekt = MemoryShapeStore()
            storageObjekt.bbox = map.layers[layerMitPunktdaten].BoundingBox()

            storeVorbereiter = shapeStoreVorbereiter( map.layers[layerMitPunktdaten].ShapeStore(), storageObjekt )
            voronoi = self.diagramSpeicher.getVoronoiDiagram( map.layers[layerMitPunktdaten], storeVorbereiter.getNormiertesPunktefeld(), storeVorbereiter.getNormierteBbox(), context )

            if voronoi.fehlerIstAufgetreten():
                self._meldeFehler( main_window, voronoi.getFehler() )
                return

            print "- zeichne diagram"
            storeVorbereiter.zeichneDiagram( voronoi.getGeordnetesPunktefeld(), voronoi.getKantenfeld() )

            titel = _("Voronoi Diagramm")
            layer = customLayer( titel, storageObjekt, typ=typenMap.getVoronoiTyp())
            #layer.SetClassificationColumn( storeVorbereiter.auswertungsAttribut )

            layer.SetClassification( storeVorbereiter.classification )

            storageObjekt.setTitle( titel )
            map.AddLayer(layer)
            context.session.AddShapeStore( storageObjekt )

        else:
            main_window.RunMessageBox(_('missing base data layer'), _('Please add a layer consisting of point data, first'))


    def _farbBereiche(self, context):
        """
        Zeichnen eines Delaunay-Rasters
        """
        main_window = context.mainwindow
        map = main_window.canvas.Map()

        self.appAnpasser.passeAn( context.application )

        if map.HasLayers() == True:
            layerMitPunktdaten = self._findePunktLayer(context)

            if layerMitPunktdaten < 0:
                self._keinPunktDatenLayer( main_window )
                return

            cols = map.layers[layerMitPunktdaten].ShapeStore().Table().Columns()
            colNames = []

            for i in cols:
                colNames.append( i.name )

            diagram_titel = "Angaben zu den Farb-Bereichen"

            v_dialog = attributAuswahl( main_window, colNames, diagram_titel )
            userEingabe = v_dialog.RunDialog()
            if userEingabe == None:
                return

            #while map.layers[layerMitPunktdaten].ShapeStore().Table().HasColumn( userEingabe["attr"] )==False or userEingabe["numKath"].isdigit()==False:
            while map.layers[layerMitPunktdaten].ShapeStore().Table().HasColumn( userEingabe["attr"] )==False:
                if map.layers[layerMitPunktdaten].ShapeStore().Table().HasColumn( userEingabe["attr"] )==False:
                    v_dialog = attributAuswahl( main_window, colNames, diagram_titel, "dieses Attribut existiert nicht" )

                userEingabe = v_dialog.RunDialog()
                if userEingabe == None:
                    return

            storageObjekt = MemoryShapeStore()
            storageObjekt.bbox = map.layers[layerMitPunktdaten].BoundingBox()
            gesuchtesAttribut = userEingabe["attr"]

            storeVorbereiter = shapeStoreVorbereiter( map.layers[layerMitPunktdaten].ShapeStore(), storageObjekt )
            storeVorbereiter.setzeAuswertungsAttribut( gesuchtesAttribut )

            voronoi = self.diagramSpeicher.getVoronoiDiagram( map.layers[layerMitPunktdaten], storeVorbereiter.getNormiertesPunktefeld(), storeVorbereiter.getNormierteBbox(), context )

            if voronoi.fehlerIstAufgetreten():
                self._meldeFehler( main_window, voronoi.getFehler() )
                return

            layerName = "Farb-Bereiche "+gesuchtesAttribut
            layer = customLayer( _(layerName), storageObjekt, typ=typenMap.getVoronoiTyp())
            layer.SetClassificationColumn( storeVorbereiter.auswertungsAttribut )

            #layer.SetClassification(storeVorbereiter.classification)
            self._klassenEinteilungAnpassen( main_window, "Klasseneinteilung anpassen", map, layer )

            storeVorbereiter.classification = layer.GetClassification()
            storeVorbereiter.zeichneFarbBereiche( voronoi.getGeordnetesPunktefeld(), voronoi.getKantenfeld() )

            storageObjekt.setTitle( layerName )
            map.AddLayer(layer)
            context.session.AddShapeStore( storageObjekt )
        else:
            main_window.RunMessageBox(_('missing base data layer'), _('Please add a layer consisting of point data, first'))

    def _klassenEinteilungAnpassen( self, main_window, title, map, layer ):
        klassifizierer = Classifier( main_window, "Klasseneinteilung anpassen", map, layer, group = None)
        klassifizierer._OnGenClass(None)
        klassifizierer.Show(True)

    def _farbFlaechen(self, context):
        """
        Zeichnen eines Delaunay-Rasters
        """
        main_window = context.mainwindow
        map = main_window.canvas.Map()

        self.appAnpasser.passeAn( context.application )

        if map.HasLayers() == True:
            layerMitPunktdaten = self._findePunktLayer(context)

            if layerMitPunktdaten < 0:
                self._keinPunktDatenLayer( main_window )
                return

            cols = map.layers[layerMitPunktdaten].ShapeStore().Table().Columns()
            colNames = []

            for i in cols:
                colNames.append( i.name )

            diagram_titel = "Angaben zu den Farbflaechen"

            v_dialog = attributAuswahl( main_window, colNames, diagram_titel )
            userEingabe = v_dialog.RunDialog()
            if userEingabe == None:
                return

            while map.layers[layerMitPunktdaten].ShapeStore().Table().HasColumn( userEingabe["attr"] )==False:
                if map.layers[layerMitPunktdaten].ShapeStore().Table().HasColumn( userEingabe["attr"] )==False:
                    v_dialog = attributAuswahl( main_window, colNames, diagram_titel, "dieses Attribut existiert nicht" )

                userEingabe = v_dialog.RunDialog()
                if userEingabe == None:
                    return

            storageObjekt = MemoryShapeStore()
            storageObjekt.bbox = map.layers[layerMitPunktdaten].BoundingBox()
            gesuchtesAttribut = userEingabe["attr"]

            storeVorbereiter = shapeStoreVorbereiter( map.layers[layerMitPunktdaten].ShapeStore(), storageObjekt )
            storeVorbereiter.setzeAuswertungsAttribut( gesuchtesAttribut )

            voronoi = self.diagramSpeicher.getVoronoiDiagram( map.layers[layerMitPunktdaten], storeVorbereiter.getNormiertesPunktefeld(), storeVorbereiter.getNormierteBbox(), context )

            if voronoi.fehlerIstAufgetreten():
                self._meldeFehler( main_window, voronoi.getFehler() )
                return

            storeVorbereiter.faerbeFlaechen( voronoi.getGeordnetesPunktefeld(), voronoi.getKantenfeld() )

            layerName = "Farb-Flaechen "+gesuchtesAttribut
            layer = customLayer( _(layerName), storageObjekt, typ=typenMap.getVoronoiTyp() )
            layer.SetClassificationColumn( gesuchtesAttribut )

            self._klassenEinteilungAnpassen( main_window, "Klasseneinteilung anpassen", map, layer )

            storageObjekt.setTitle( layerName )
            map.AddLayer(layer)
            context.session.AddShapeStore( storageObjekt )

        else:
            main_window.RunMessageBox(_('missing base data layer'), _('Please add a layer consisting of point data, first'))

    def _preparePirolDB(self,context):
        if self.dbfact==None:
            connections = context.session.DBConnections()
            if len (connections) == 0:
                 context.mainwindow.RunMessageBox(_('Keine Datenbankverbindung vorhanden'), _('Fuegen sie die Pirol - Datenbankverbindung hinzu'))
                 return False
            #conn = connections[0].connection
            conn = connections[0]
            writer = dbWriter.dbWriter.getDbWriter(conn)
            self.dbfact = DbFactory.getDbFactory( writer )
            self.diagrammBuilder = DiagrammBuilder(self.dbfact,context.mainwindow)
            self.diagramSpeicher.setDiagrammBuilder(self.diagrammBuilder)
            return True
        return True




    def _save(self,context):
        main_window = context.mainwindow
        map = main_window.canvas.Map()
        self.appAnpasser.passeAn( context.application )
        if map.HasLayers() == True:
            layerMitPunktdaten = self._findePunktLayer(context)
            if layerMitPunktdaten < 0:
                self._keinPunktDatenLayer( main_window )
                return
            if not self._preparePirolDB(context):
                return
            zerleger = ShapeStoreZerleger(self.dbfact,map.layers[layerMitPunktdaten].ShapeStore(),main_window)
        if not zerleger.saveShapeFile():
                main_window.RunMessageBox(_('Fehler beim Speichern der Punktdaten'), _('Das solte nicht vorkommen'))
                return
        print " -- Layer in Datenbank exportiert"
        # Ok die Originale Karte entfernen
        map.RemoveLayer(map.layers[layerMitPunktdaten])
        theRootGeom = zerleger.theRootGeom
        gruppe = zerleger.gruppe
        builder = ShapeStoreBuilder(self.dbfact,main_window,None,gruppe.gruppenid)
        if builder != None :

            newStorageObject = builder.getShapeStore()
            if newStorageObject == None:
                 return
            Name= "%s, %s"%(zerleger.schlag.name,zerleger.gruppe.name)
            layer = customLayer(Name,newStorageObject, typenMap.getPirolDB() )
            layer.rolle.schlagid = builder.schlagChoise
            layer.rolle.tiefenid = builder.tiefenChoise
            layer.rolle.gruppenid= builder.gruppenChoise
            map.AddLayer(layer)
            layer.ShapeStore().setTitle(layer.Title())
            context.session.AddShapeStore( layer.ShapeStore() )

            main_window.canvas.FitMapToWindow()

        else:
            main_window.RunMessageBox(_('missing base data layer'), _('Please add a layer consisting of point data, first'))

    def _load(self,context):
        main_window = context.mainwindow
        map = main_window.canvas.Map()

        self.appAnpasser.passeAn( context.application )
        if not self._preparePirolDB(context):
            return
        builder = ShapeStoreBuilder(self.dbfact,main_window)
        if builder == None :
            return
        newStorageObject = builder.getShapeStore()
        if newStorageObject == None:
            return
        layer = customLayer(_(newStorageObject.shapeStoreName),newStorageObject, typenMap.getPirolDB() )

        layer.rolle.schlagid = builder.schlagChoise
        layer.rolle.tiefenid = builder.tiefenChoise
        layer.rolle.gruppenid= builder.gruppenChoise

        map.AddLayer(layer)
        newStorageObject.setTitle( newStorageObject.shapeStoreName )
        context.session.AddShapeStore( newStorageObject )
        main_window.canvas.FitMapToWindow()
        main_window.canvas.full_redraw()

    def passeApplikationAn( self, context ):
        main_window = context.mainwindow
        self.appAnpasser.passeAn( context.application )
        context.session.AddDBConnection( PostGISConnection( "dbname", host="dbhost", user="username", password="passwd", dbtype="", port="" ) )

        #main_window.RunMessageBox(_('Vielen Dank!'), _('Die Pirol-Erweiterungen koennen nun im vollen Umfang genutzt werden!'))

# -----------------------
    def _meldeFehler( main_window, fehler ):
        if fehler != None:
            main_window.RunMessageBox(_('Fehler'), _('Es ist folgender Fehler aufgetreten:\n'+str(fehler)))

    _meldeFehler = staticmethod( _meldeFehler )

# -----------------------
    def _loadcsv(self,context, reversed=False):
        main_window = context.mainwindow
        map = main_window.canvas.Map()
        self.appAnpasser.passeAn( context.application )
        csvreader = None
        storageObjekt =None

        dlg = wxFileDialog(main_window, _("Open CSVFile"), ".", "","Pirol CSV File (*.csv)|*.csv|Textfiles (*.txt)|*.txt",  wxOPEN)
        if dlg.ShowModal() == wxID_OK:
            try:
               name = dlg.GetPath()

               csvreader = csvReader(name,"\t", 0, 1, justFetchRows = True)
               firstline = csvreader.lineone
               fourthline = csvreader.getForthLineAsString()

               dialog = CSV_Dialog( main_window, firstline, fourthline )
               result = dialog.RunDialog()

               if ( result == None ):
                   print "Abbruch!"
                   return

               csvreader = csvReader(name,result["sep"], result["ost"], result["nord"])
               storageObjekt = csvreader.getShapeStore()
            except csvError,e:
               main_window.RunMessageBox(_('Fehler'),_(e.message))
               return
        else:
            dlg.Destroy()
            return

        dlg.Destroy()
        layer = customLayer(_(dlg.GetFilename()),storageObjekt,typ=typenMap.getCSVTyp() )
        map.AddLayer(layer)
        storageObjekt.setTitle( layer.Title() )
        context.session.AddShapeStore( storageObjekt )
        main_window.canvas.FitMapToWindow()
        main_window.canvas.full_redraw()



PrecisionFarming = precisionFarming(registry, main_toolbar, main_menu)
