from Thuban import _
from Thuban.Model.table import FIELDTYPE_INT,FIELDTYPE_STRING,FIELDTYPE_DOUBLE

from customLayer import *
from utilities import *
from polyTool import *

from diagrammDialog import attributAuswahl

# schleifen beschleunigen
try:
    from psyco.classes import __metaclass__
except:
    pass

class rasterTool( polyTool ):
    def __init__( self, view, context, findePunktLayer, appAnpasser, meldeFehler ):
        polyTool.__init__( self, view, context, findePunktLayer, appAnpasser )
        
        self.rasterlayerName = "Basis-Raster Nr."
        self.basisRaster = None
        
        self.layerName = "Basis-Flaeche Nr."
        
        self._meldeFehler = meldeFehler

        spurBreite = self._erfrageSpurBreite( )
            
        self.spurBreite = spurBreite
        self.storageObjekt.setTable( self._erstelleLeereTabelle() )
        
        self.polygone = 0
        self.rasterFlaechenBegrenzung = None
        
    def _erfrageSpurBreite( self ):
        main_window = self.context.mainwindow
        map = main_window.canvas.Map()
        diagram_titel = _("Bitte geben Sie die gewuenschte Spur-Breite ein.")
        text = _("Spur Breite (m)")
        message = _("Gesucht ist die Dezimal- oder Kommazahl, die die Spur-Breite angibt.\nEinige Werte sind bereits eingetragen, Sie koennen aber auch eigene Werte eingeben.")
        
        v_dialog = attributAuswahl( main_window, ["5.0","6.0","10.0","12.0","15.0","16.0","20.0"], diagram_titel, text=text, message=message )
        userEingabe = v_dialog.RunDialog()
        if userEingabe == None:
            return -1

        while not istZahl(userEingabe["attr"]):
            v_dialog = attributAuswahl( main_window, colNames, diagram_titel, text=text, message=message )
            userEingabe = v_dialog.RunDialog()
            
        return float( userEingabe["attr"] )
        
    def Name(self):
        return "RasterWerkzeug"
        
    def _erstelleLeereTabelle( self ):
        table = MemoryTable([
                            ("Funktion", FIELDTYPE_STRING),
                            ("Spurbreite", FIELDTYPE_DOUBLE)
                            ],[])
        table.SetTitle( self.layerName )
        return table

    def _fuegeLinieEinZwischen( self, event, punkt1, punkt2 ):
        main_window = self.context.mainwindow
        map = main_window.canvas.Map()
        self.appAnpasser.passeAn( self.context.application )

        if map.HasLayers() == True and punkt1 != None and punkt2 != None:

            distanz = self._berechneEntfernung( punkt1, punkt2 )

            if ( distanz <= 0.005 ):
                if ( self.closing == 1 ):
                    return False
                if ( self.teilstuecke > 0 ):
                    # Schliessen des LinienZuges
                    self.closing = 1
                    self._fuegeLinieEinZwischen( None, self.punkte[ len(self.punkte)-1 ], self.ausgangsPunkt() )
                    self._resetTool()
                    return True
                else:
                    return False
            else:
                self._aktualisiereTeilstuecke(distanz)

                self.storageObjekt.addLine( [punkt1.x(),punkt2.x()], [punkt1.y(),punkt2.y()] )
                if ( self.teilstuecke == 1 ):
                    self.storageObjekt.appendAttributeData( ["richtungsgebend",self.spurBreite] )
                else:
                    self.storageObjekt.appendAttributeData( ["begrenzend",self.spurBreite] )

                if ( self.teilstuecke == 1 ):
                    map.AddLayer(self.layer)
                    self.context.session.AddShapeStore( self.storageObjekt )
                    
                self.layer.issue(LAYER_CHANGED, self.layer)
                return True
    

    def _resetTool( self ):
        """
        Wird nur beim Schliessen eines Polygons aufgerufen...!?
        """
        main_window = self.context.mainwindow
        map = main_window.canvas.Map()
        
        # Erstellen der neuen Raster Ebene!
        nummer = self._sucheLayerNummer( self.context.mainwindow.canvas.Map(), self.rasterlayerName )
        self.basisRaster = customLayer( _(self.rasterlayerName+str(nummer)), MemoryShapeStore(), typenMap.getRasterTyp() )
        
        self.basisRaster.ShapeStore().setTable( MemoryTable([
                             ("Index", FIELDTYPE_INT)],[] )  )
        
        if ( self.erstelleRaster() ):
            main_window = self.context.mainwindow
            map = main_window.canvas.Map()
            self.basisRaster.ShapeStore().Table().SetTitle( _(self.rasterlayerName+str(nummer)) )
            map.AddLayer(self.basisRaster)
            self.context.session.AddShapeStore( self.basisRaster.ShapeStore() )
        elif self.rasterFlaechenBegrenzung.fehlerIstAufgetreten():
            self._meldeFehler( main_window, self.rasterFlaechenBegrenzung.getFehler() )
            return False
            
        
        # Eigentlicher Reset
        self.polygone = 0
        polyTool._resetTool( self )
        
    def erstelleRaster( self ):
        main_window = self.context.mainwindow
        map = main_window.canvas.Map()
        
        kantenfeld = []
        
        for i in range( len(self.punkte) ):
            anfang = self.punkte[i]
            
            if ( i+1 < len(self.punkte) ):
                ende = self.punkte[i+1]
            else:
                ende = self.punkte[0]
            
            # x_anfang, y_anfang, x_ende, y_ende, dr, dl, inf_anfang = False, inf_ende = False):
            kantenfeld.append( kante( anfang.x(), anfang.y(), ende.x(), ende.y(), -1, -1 ) )
        
        self.rasterFlaechenBegrenzung = polygon( kantenfeld )
        
        if self.rasterFlaechenBegrenzung.fehlerIstAufgetreten():
            self._meldeFehler( main_window, self.rasterFlaechenBegrenzung.getFehler() )
            return False
        
        # test:
        #print "nee: punkt( [1., 2.]):\n",rasterFlaechenBegrenzung.liegtPunktDrin( punkt( [1., 2.]) )
        #print "ja : punkt( [3436094.432, 5797951.4]):\n",self.rasterFlaechenBegrenzung.liegtPunktDrin( punkt( [3436094.432, 5797951.4]) )
        #print "nee: punkt( [3436091.929, 5798008.95]):\n",rasterFlaechenBegrenzung.liegtPunktDrin( punkt( [3436091.929, 5798008.95]) )
        #print "ja : punkt( [3436046., 5798001.]):\n",self.rasterFlaechenBegrenzung.liegtPunktDrin( punkt( [3436046., 5798001.]) )

        rechtEckMacher = rechteckErzeuger(self.spurBreite,self.spurBreite)
        
        ausgangsKante = self.rasterFlaechenBegrenzung.kanten[0].getNormalenKante( self.spurBreite )
        allerErsteKante = self.rasterFlaechenBegrenzung.kanten[0].getNormalenKante( self.spurBreite )
        
        fertig = False
        ersteKante = True
        gegenRichtungUeberprueft = False
        
        rechtsEnde = False
        linksEnde = False
        
        ausgangsKanteLinks = ausgangsKante
        ausgangsKanteRechts = ausgangsKante
        
        sicherheitabstand = self.spurBreite * 1.45
        
        minX, minY, maxX, maxY = self.layer.ShapeStore().BoundingBox()
        
        minX -= sicherheitabstand
        minY -= sicherheitabstand
        
        maxX += sicherheitabstand
        maxY += sicherheitabstand
        
        letztesLPoly = None
        letztesRPoly = None
        
        counter = 0
        while( not fertig and not self.rasterFlaechenBegrenzung.fehlerIstAufgetreten()):
            if ( not linksEnde ):
                nPoly = rechtEckMacher.erzeugeRechtEck(ausgangsKanteLinks)
                if ( self._fuegePolyInRasterLayerEin( nPoly ) ):
                    # nach links
                    letztesLPoly = nPoly
                ausgangsKanteLinks = nPoly.kanten[2]
                tmpPunkt = ausgangsKanteLinks.anfangspunkt
                ausgangsKanteLinks.anfangspunkt = ausgangsKanteLinks.endpunkt
                ausgangsKanteLinks.endpunkt = tmpPunkt
                    
                mnX, mnY, mxX, mxY = nPoly.BoundingBox()
                #if ((mnX < minX or mnY < minY or mxX > maxX or mxY > maxY) and not ersteKante):
                if ((mnX < minX or mnY < minY or mxX > maxX or mxY > maxY)):
                    linksEnde = True
                    
            if ( not rechtsEnde ):
                if ( ersteKante ):
                    tmpPunkt = ausgangsKanteRechts.anfangspunkt
                    ausgangsKanteRechts.anfangspunkt = ausgangsKanteRechts.endpunkt
                    ausgangsKanteRechts.endpunkt = tmpPunkt
                nPoly = rechtEckMacher.erzeugeRechtEck(ausgangsKanteRechts)
                if ( self._fuegePolyInRasterLayerEin( nPoly ) ):
                    # nach rechts
                    letztesRPoly = nPoly
                ausgangsKanteRechts = nPoly.kanten[2]
                tmpPunkt = ausgangsKanteRechts.anfangspunkt
                ausgangsKanteRechts.anfangspunkt = ausgangsKanteRechts.endpunkt
                ausgangsKanteRechts.endpunkt = tmpPunkt
                
                mnX, mnY, mxX, mxY = nPoly.BoundingBox()
                #if ((mnX < minX or mnY < minY or mxX > maxX or mxY > maxY) and not ersteKante):
                if ((mnX < minX or mnY < minY or mxX > maxX or mxY > maxY) ):
                    rechtsEnde = True

            ersteKante = False

            if (rechtsEnde and linksEnde):
                #neue Reihe
                if ( letztesRPoly == None and letztesLPoly == None):
                    print "keine neue Reihe!"
                    # TODO: Gegenrichtung ueberpruefen!!
                    if (gegenRichtungUeberprueft==True):
                        fertig = True
                    else:
                        #fertig = True
                        gegenRichtungUeberprueft = True
                        print "neue Reihe in Gegenrichtung!"
                        ersteKante = True
                        rechtsEnde = False
                        linksEnde = False
                        
                        diffX = allerErsteKante.endpunkt.x() - allerErsteKante.anfangspunkt.x()
                        diffY = allerErsteKante.endpunkt.y() - allerErsteKante.anfangspunkt.y()
                        
                        allerErsteKante.anfangspunkt.setX( allerErsteKante.anfangspunkt.x() )
                        allerErsteKante.endpunkt.setX( allerErsteKante.endpunkt.x()-2.0*diffX )
                        
                        allerErsteKante.anfangspunkt.setY( allerErsteKante.anfangspunkt.y() )
                        allerErsteKante.endpunkt.setY( allerErsteKante.endpunkt.y()-2.0*diffY )
                        
                        ausgangsKanteRechts = allerErsteKante
                        ausgangsKanteLinks = allerErsteKante
                        
                        letztesLPoly = None
                        letztesRPoly = None
                        
                else:
                    ersteKante = True
                    rechtsEnde = False
                    linksEnde = False
#                    if (not gegenRichtungUeberprueft):
                    if ( letztesLPoly != None ):
                        neueKante = letztesLPoly.kanten[1]
                    else:
                        neueKante = letztesRPoly.kanten[3]
#                    else:
#                        if ( letztesLPoly != None ):
#                            neueKante = letztesLPoly.kanten[1]
#                        else:
#                            neueKante = letztesRPoly.kanten[3]
                            
                    tmpPunkt = neueKante.anfangspunkt
                    neueKante.anfangspunkt = neueKante.endpunkt
                    neueKante.endpunkt = tmpPunkt
                    neueKante = neueKante.getNormalenKante( self.spurBreite )
                    ausgangsKanteLinks = neueKante
                    ausgangsKanteRechts = neueKante
                    print "neue Reihe!", neueKante
                        
                    letztesLPoly = None
                    letztesRPoly = None
        
        if (self.rasterFlaechenBegrenzung.fehlerIstAufgetreten()):
            print "Fehler:", self.rasterFlaechenBegrenzung.getFehler()
                  
        return fertig
#        return True
            
    def _fuegePolyInRasterLayerEin( self, poly ):
        
        if ( self.rasterFlaechenBegrenzung.liegtPolygonDrin(poly) ):
            store = self.basisRaster.ShapeStore()
            xs = []
            ys = []
            
            for pkt in poly.getPunkte():
                xs.append( pkt.x() )
                ys.append( pkt.y() )
            
            # Polygon schliessen
            xs.append( xs[0] )
            ys.append( ys[0] )
    
            store.appendAttributeData([self.polygone])
            self.polygone += 1
            store.addPolygon(xs,ys)
            
            #print "fuege Polygon ein"
            return True
        #print "polygon nicht eingefuegt"
        return False
        