
import shapelib
import shptree
import dbflib

from Thuban.Model.data import SHAPETYPE_POLYGON, SHAPETYPE_ARC, SHAPETYPE_POINT, shapelib_shapetypes
from Thuban.Model.table import FIELDTYPE_INT, FIELDTYPE_STRING, FIELDTYPE_DOUBLE

class shapeFileSpeicherer:
    def __init__( self, pfad, layer ):
        """
        dateiName - (war redundant - is jetzt weg) name der ziel datei INKLUSIVE der erweiterung (.shp)
        pfad      - pfad der ziel datei INKLUSIVE des dateinamens (s.o.)
        layer     - halt das layer mit den Daten zum Speichern...
        """
        #self.filename = dateiName
        self.path = pfad
        self.dbfPath = pfad[:pfad.rfind(".shp")] + ".dbf"
        self.layer = layer
        
        self.counterDict = {}
        
        self.dbfDict = {
                        FIELDTYPE_INT:dbflib.FTInteger,
                        FIELDTYPE_DOUBLE:dbflib.FTDouble,
                        FIELDTYPE_STRING:dbflib.FTString
                        }
        
        keys = shapelib_shapetypes.keys()
        vals = shapelib_shapetypes.values()
        
        for index in range( len(vals)):
            self.counterDict[vals[index]] = keys[index]
            
        #print self.counterDict
        
        self.shapefile = None
        self.dbf       = None
        self.shapelibShapeTyp = self.counterDict[self.layer.ShapeType()]
        
        #print "->", pfad, dateiName
        
       
    def speichern( self ):
        if ( not self.layer.HasShapes() ):
            print "gibt nix zu Speichern!"
            return
        self.shapefile = shapelib.create(self.path,self.shapelibShapeTyp)
        
        numPunkte = 0
        shapeNr = 0
        for shape in self.layer.ShapeStore().AllShapes():
            points = shape.Points()

            #print "-->", self.shapelibShapeTyp, shape.ShapeID(), points 
            SHPObj = shapelib.SHPObject(self.shapelibShapeTyp, shapeNr, points )
            self.shapefile.write_object( -1, SHPObj )
            shapeNr += 1
            
        #print "-----------> ", self.shapefile.info()
        del self.shapefile
        
        #print "---------Start DBf-------------"
        

        self.dbf = dbflib.create(self.dbfPath)
        
        ## table head
        table = self.layer.ShapeStore().Table()
        for column in table.Columns():
            column.name = column.name[:10]
            prec = 0
            width = 0
            type = self.dbfDict[column.type]
            if ( column.type == FIELDTYPE_INT ):
                width = 10
            elif ( column.type == FIELDTYPE_DOUBLE ):
                width = 10
                prec = 4
            elif ( column.type == FIELDTYPE_STRING ):
                width = 80
            
            #print "clo: ", column.name, type, width, prec
            
            self.dbf.add_field(column.name, type, width, prec)
    
        del self.dbf
        #print "---------DBf data-------------"
        self.dbf = dbflib.open(self.dbfPath, "r+b")
 
        # Records can be added as a dictionary...
        for rowNr in range( table.NumRows() ):
            #print "data: ",rowNr, table.ReadRowAsDict(rowNr)
            self.dbf.write_record(rowNr, table.ReadRowAsDict(rowNr, row_is_ordinal = 1 ) )

        #print "---------Done DBf-------------"
        