# Author: Ole Rahn
# Project: PIROL

from Thuban.Model.table import *
from Thuban.Model.classification import *

from utilities import *
from gruppen import *


# schleifen beschleunigen
try:
    from psyco.classes import __metaclass__
except:
    pass


class shapeStoreVorbereiter:
    def __init__(self, store, targetStore = None):
        self.store = store
        self.storeTable = store.Table()
        self.punktefeld = []
        self.hoehenlinienWert = 0
        
        boundingBox = store.BoundingBox()
        
        self.mengeKathegorien = 7
        self.auswertungsAttribut = None
        self.attributRechts = None
        self.attributLinks = None
        
        self.classification = Classification()
        self.classification.SetDefaultFill(Color(0,0,1))
        self.classification.SetDefaultLineColor(Color(0,0,1))
        
        self.genauigkeit = 1.0/1024
        self.debug = False
        
        try:
            # DB Version des Shape-Auslesens
            print "- shapeStoreVorbereiter(): versuche Datenbank zu lesen"
            alleShapes = store.AllShapes()

            for i in alleShapes:
                punkteTuple = i.Points()
                icks, uepsilon = punkteTuple[0][0]
                self.punktefeld.append( punkt( [icks, uepsilon], i.ShapeID() ) )
        except:
            print "- shapeStoreVorbereiter(): scheint keine DB zu sein, versuche ShapeFile auszulesen"

            try:
                for i in range(store.NumShapes()):
                    punkteTuple = store.Shape(i).Points()
                    icks, uepsilon = punkteTuple[0][0]
                    self.punktefeld.append( punkt( [icks, uepsilon], store.Shape(i).ShapeID() ) )
            except:
                print "- shapeStoreVorbereiter(): tut auch nicht!"
                return
            
        self.datenWerkzeug = datenTool( self.punktefeld, boundingBox )
        if targetStore != None:
            self.targetStore = targetStore
            self.targetStore.bbox = self.datenWerkzeug.getBbox()
    
    def setzeAnzahlDerKathegorien( self, zahl ):
        self.mengeKathegorien = zahl
        
    def setzeAuswertungsAttribut( self, behandeltesAttribut ):
        self.auswertungsAttribut = behandeltesAttribut
        self.attributRechts = behandeltesAttribut + "_rechts"
        self.attributLinks = behandeltesAttribut + "_links"
        
    def setztePunkteFeld( self, punktefeld, boundingBox, vorbereitenZumAnnaehern = True ):
        self.datenWerkzeug = datenTool( punktefeld, boundingBox, vorbereitenZumAnnaehern )
        self.punktefeld = punktefeld
        self.targetStore.bbox = self.datenWerkzeug.getBbox()
        
    def statistischeAuswertung( self ):
        cols = self.storeTable.Columns()
        memTabHeadLine = []
        
        for i in cols:
            memTabHeadLine.append( (i.name, i.type) )
        
        memTable = MemoryTable(memTabHeadLine,[])
        
        for i in self.punktefeld:
            neueRow = []
            aktRowDict = self.storeTable.ReadRowAsDict( i.index )
            
            for j in cols:
                neueRow.append( aktRowDict[ j.name ] )
            memTable.data.append( tuple( neueRow ) )
        
        resultat = {}
        for i in cols:
            minWert = None
            minGID = -1
            maxWert = None
            maxGID = -1
            mittelwert = 0
            stdAbw = 0
            werteZaehlung = {}
            for j in range( memTable.NumRows() ):
                aktWert = memTable.ReadValue( j, i.name )
                if i.type == FIELDTYPE_INT or i.type == FIELDTYPE_DOUBLE:
                    mittelwert += aktWert
                else:
                    if werteZaehlung.has_key( aktWert ) == True:
                        werteZaehlung[ aktWert ] += 1
                    else:
                        werteZaehlung[ aktWert ] = 1
                    
                if minWert == None or minWert > aktWert:
                    minWert = aktWert
                    minGID = self.punktefeld[ j ].index
                if maxWert == None or maxWert < aktWert:
                    maxWert = aktWert
                    maxGID = self.punktefeld[ j ].index
            
            if i.type == FIELDTYPE_INT or i.type == FIELDTYPE_DOUBLE:
                mittelwert /= memTable.NumRows()
                for j in range( memTable.NumRows() ):
                    aktWert = memTable.ReadValue( j, i.name )
                    if i.type == FIELDTYPE_INT or i.type == FIELDTYPE_DOUBLE:
                        stdAbw += math.pow( aktWert - mittelwert, 2 )
                stdAbw /= memTable.NumRows()
                stdAbw = math.sqrt( stdAbw )
                
            else:
                menge = 0
                for j in werteZaehlung.keys():
                    if werteZaehlung[j] > menge:
                        menge = werteZaehlung[j]
                        mittelwert = j
                minWert= "N.A."
                maxWert= "N.A."
                stdAbw = "N.A."
            resultat[ i.name ] = [ minWert, mittelwert, maxWert, stdAbw, minGID, maxGID ]
            
        return resultat
            
    def identifiziereAusreisser( self ):
        attr = self.auswertungsAttribut
        
        self.setzeAnzahlDerKathegorien( 2 )
        
        werte = []
        minWert = None
        maxWert = None
        mittelwert = 0
        standardAbweichung = 0
        mengeWerte = len(self.punktefeld)
    
        zielSpalte = self.store.Table().Column(attr)
        self.targetStore.setTable( MemoryTable([
                             ("interner Punkt Ind.", FIELDTYPE_INT),
                             (attr, zielSpalte.type)],[]) )

        for i in self.punktefeld:
            aktRowDict = self.storeTable.ReadRowAsDict( i.index )
            wert = aktRowDict[ attr ]
            werte.append( wert )
            
            mittelwert += wert
            
            if minWert == None or minWert > wert:
                minWert = wert
            if maxWert == None or maxWert < wert:
                maxWert = wert
        
        mittelwert /= mengeWerte
        
        for i in self.punktefeld:
            aktRowDict = self.storeTable.ReadRowAsDict( i.index )
            wert = aktRowDict[ attr ]
        
            standardAbweichung += math.pow( wert - mittelwert ,2)
        
        standardAbweichung /= mengeWerte
        standardAbweichung = math.sqrt( standardAbweichung )
        
        dreiSigma = 3.0 * standardAbweichung
        
        untereSchranke = mittelwert - dreiSigma
        obereSchranke  = mittelwert + dreiSigma
        
        ausreisserGefunden = False
        
        for i in self.punktefeld:
            aktRowDict = self.storeTable.ReadRowAsDict( i.index )
            wert = aktRowDict[ attr ]
            
            if wert < untereSchranke or wert > obereSchranke:
                ausreisserGefunden = True
                self.targetStore.addPoint( i.x(), i.y() )
                self.targetStore.appendAttributeData( (i.index, wert) )
        
        self.generiereKathegorien( self.storeTable, minWert, maxWert, schwarzerRand=True )
        
        return ausreisserGefunden

            
    def setzeGesuchtenWertFuerHoehenlinie( self, wert ):
        self.hoehenlinienWert = wert

    def getNormierteBbox(self):
        return self.datenWerkzeug.getNormierteBbox()
        
    def getNormiertesPunktefeld(self):
        return self.datenWerkzeug.getNormiertesPunktefeld()
        
    def getRohPunktefeld( self ):
        return self.punktefeld

    def faerbeFlaechen( self, punktefeldEin, kantenfeldEin ):
        punktefeld = self.datenWerkzeug.hebeNormierungAuf( punktefeldEin )
        kantenfeld = self.datenWerkzeug.unnormieresKantenfeld( kantenfeldEin )
        print "- uebernehme Attribute..."
        origColumns = self.storeTable.Columns()
        origColumn_map = {}
        
        for i in range( len(origColumns) ):
            origColumn_map[origColumns[i].name] = self.storeTable.Column(i) 
        print "- Attribute uebernommen."

        nummer = 0
        attr = self.auswertungsAttribut
        
        zielSpalte = origColumn_map[attr]
        self.targetStore.setTable( MemoryTable([
                             ("interner Punkt Ind.", FIELDTYPE_INT),
                             (attr, zielSpalte.type)],[]) )
        if zielSpalte.type == FIELDTYPE_INT or zielSpalte.type == FIELDTYPE_DOUBLE:
            minWert = None
            maxWert = None
            aktuellerWert = 0
            for i in punktefeld:
                aktuellerDatensatz = self.storeTable.ReadRowAsDict(i.index)
                aktuellerWert = aktuellerDatensatz[attr]
                if minWert == None or aktuellerWert<minWert:
                    minWert = aktuellerWert
                if maxWert == None or aktuellerWert>maxWert:
                    maxWert = aktuellerWert
    
            #self.generiereKathegorien( self.storeTable, minwert=minWert,maxwert=maxWert)
        else:
            disWerte = []
            for i in punktefeld:
                aktuellerDatensatz = self.storeTable.ReadRowAsDict(i.index)
                aktuellerWert = aktuellerDatensatz[attr]
                
                if not (aktuellerWert in disWerte):
                    disWerte.append( aktuellerWert )
            
            #self.generiereKathegorien(self.storeTable, diskreteSpalten=disWerte)
        
        print "- finde Kanten zu Punkten"
        kantenZuPunkten = self._findeKantenZuPunkten( punktefeld, kantenfeld )
        
        print "- ordne und zeichne Shapes"
        for j in kantenZuPunkten:
            if kantenZuPunkten.index(j) % 100 == 0:
                print "- shape", kantenZuPunkten.index(j)
            fertig = False
            aktKantenzug = []
            xs = []
            ys = []
            counter = 0
            #while fertig==False and counter < 10:
            while fertig==False:
                counter += 1
                #print "- neuer durchgang"
                fertig = True
                for i in j:
                    if i.markiert == True:
                        continue
                    #print "- kante", j.index(i)
                    if len( aktKantenzug ) == 0:
                        aktKantenzug.append( i.anfangspunkt )
                        aktKantenzug.append( i.endpunkt )
                        i.markiert = True
                    else:
                        #print i.anfangspunkt.x, i.anfangspunkt.y, "|", i.endpunkt.x, i.endpunkt.y, "|", aktKantenzug[0].x, aktKantenzug[0].y, "|", aktKantenzug[len(aktKantenzug)-1].x, aktKantenzug[len(aktKantenzug)-1].y
                        if i.anfangspunkt.abstandZu( aktKantenzug[0] ) < self.genauigkeit:
                            aktKantenzug.insert( 0, i.endpunkt )
                            i.markiert = True
                        elif i.endpunkt.abstandZu( aktKantenzug[0] ) < self.genauigkeit:
                            aktKantenzug.insert( 0, i.anfangspunkt )
                            i.markiert = True
                        elif i.anfangspunkt.abstandZu( aktKantenzug[len(aktKantenzug)-1] ) < self.genauigkeit:
                            aktKantenzug.append( i.endpunkt )
                            i.markiert = True
                        elif i.endpunkt.abstandZu( aktKantenzug[len(aktKantenzug)-1] ) < self.genauigkeit:
                            aktKantenzug.append( i.anfangspunkt )
                            i.markiert = True
                        else:
                            fertig = False
                            
                        if len(aktKantenzug) > 2 and aktKantenzug[0].abstandZu( aktKantenzug[len(aktKantenzug)-1] ) < self.genauigkeit:
                            fertig == True
                            
                if fertig == True:
                    #print "- hebe Markierungen auf"
                    for i in j:
                        i.markiert = False
            
                if counter >= pow( len(j), 2 ):
                    break
            if len(aktKantenzug)<3:
                continue
            if aktKantenzug[0].abstandZu( aktKantenzug[len(aktKantenzug)-1] ) > self.genauigkeit:
                aktKantenzug.append( aktKantenzug[0] )
                
            for i in aktKantenzug:
                xs.append( i.x() )
                ys.append( i.y() )
                
            punktIndex = punktefeld[ kantenZuPunkten.index(j) ].index
            main_row = self.storeTable.ReadRowAsDict( punktIndex )
            self.targetStore.addPolygon( xs, ys )
            self.targetStore.appendAttributeData( (punktIndex, main_row[attr]) )
            
    def generiereKathegorien( self, table, minwert=0., maxwert=10., diskreteSpalten=[], schwarzerRand=False ):
        if table == []:
            print "generiereKathegorien(): Zuerst Datentabelle fuellen!"
            return
        
        spalte = table.Column(self.auswertungsAttribut)
        if spalte.type == FIELDTYPE_STRING:
            if diskreteSpalten != []:
                diskreteSpalten.sort()
                colorStep = 1./(len(diskreteSpalten)+1.)
                for i in range( len(diskreteSpalten) ):
                    label = "Kat.:"+str(diskreteSpalten[i])
                    props = ClassGroupProperties()
                    if schwarzerRand == False:
                        props.SetLineColor( Color( (colorStep*(1.+i)), 1.-(colorStep*(1.+i)), 0) )
                    else:
                        props.SetLineColor( Color( 0, 0, 0) )
                    props.SetFill( Color( (colorStep*(i+1.)), 1.-(colorStep*(i+1.)), 0) )
                    self.classification.AppendGroup( ClassGroupSingleton( diskreteSpalten[i], props, label ) )
        
            print "diskrete Werte - bis jetzt ungetestet, scheint aber zu funktionieren"
            return
        else:
            minWert = minwert
            maxWert = maxwert
            
            if minWert == maxWert:
                print "generiereKathegorien(): range ist null - kann keine Kathegorien generieren"
                return
            
            genauigkeit = .0001          
            maxWert += genauigkeit
            minWert -= genauigkeit
            Range = maxWert - minWert
            step = Range/float(self.mengeKathegorien)
            #colorStep = 1./float(self.mengeKathegorien+1.)
            colorStep = 1./float(self.mengeKathegorien-1.)
            
            props = []
            label = None
            
            text_unten = " "
            text_oben = " "
            
            for i in range( int(self.mengeKathegorien) ):
                text_unten = " " + (repr(minWert+i*step))
                text_oben = " " + (repr(minWert+(i+1)*step))
            
                label = "Kat.:"+text_unten[:7]+"-"+text_oben[:7]
                props = ClassGroupProperties()
                
                if schwarzerRand == False:
                    props.SetLineColor( Color( (colorStep*i), 1.-(colorStep*i), 0) )
                else:
                    props.SetLineColor( Color( 0, 0, 0) )
                
                props.SetFill( Color( (colorStep*i), 1.-(colorStep*i), 0) )
                self.classification.AppendGroup( ClassGroupRange( (minWert+i*step, minWert+(i+1)*step), props, label ) )
            
            
            
    def _findeKantenZuPunkten( self, punktefeld, kantenfeld ):
        kantenZuPunkten = []
        for i in punktefeld:
            kantenZuPunkten.append([])
        
        for i in kantenfeld:
            if i.ungueltig == True:
                continue
            kantenZuPunkten[ i.datenPunktRechts ].append(i)
            kantenZuPunkten[ i.datenPunktLinks ].append(i)
            
        return kantenZuPunkten
            
    def zeichneFarbBereiche( self, punktefeldEin, kantenfeldEin ):
        punktefeld = self.datenWerkzeug.hebeNormierungAuf( punktefeldEin )
        kantenfeld = self.datenWerkzeug.unnormieresKantenfeld( kantenfeldEin )
        print "uebernehme Attribute..."
        print len(punktefeld), len(kantenfeld)
        
        origColumns = self.storeTable.Columns()
        origColumn_map = {}
        
        for i in range( len(origColumns) ):
            origColumn_map[origColumns[i].name] = self.storeTable.Column(i) 
        print "Attribute uebernommen."

        nummer = 0
        attr = self.auswertungsAttribut
        attr_R = self.attributRechts
        attr_L = self.attributLinks
        
        zielSpalte = origColumn_map[attr]
        self.targetStore.setTable( MemoryTable([
                             ("LinienNummer", FIELDTYPE_INT),
                             (attr_R, zielSpalte.type),
                             (attr_L, zielSpalte.type),
                             (attr, zielSpalte.type)],[]) )
          
        # wertebereich fuer kathegorien feststellen - die kathegorien werden in den folgenden funktionen gebraucht           
#================================================================================
#        minWert = None
#        maxWert = None
#
#        if zielSpalte.type == FIELDTYPE_INT or zielSpalte.type == FIELDTYPE_DOUBLE:
#            for i in punktefeld:
#                aktuellerDatensatz = self.storeTable.ReadRowAsDict(i.index)
#                aktuellerWert = aktuellerDatensatz[attr]
#                if minWert == None or aktuellerWert<minWert:
#                    minWert = aktuellerWert
#                if maxWert == None or aktuellerWert>maxWert:
#                    maxWert = aktuellerWert
#                    
#            self.generiereKathegorien( self.storeTable,minwert=minWert,maxwert=maxWert)
#        
#        else:
#            disWerte = []
#            for i in punktefeld:
#                aktuellerDatensatz = self.storeTable.ReadRowAsDict(i.index)
#                aktuellerWert = aktuellerDatensatz[attr]
#                
#                if not (aktuellerWert in disWerte):
#                    disWerte.append( aktuellerWert )
#            
#            self.generiereKathegorien(self.storeTable,diskreteSpalten=disWerte)
#================================================================================
        
        # rausfischen der zusammengehoerigen punkte
        GruppenBilder = gruppenBilder( self.storeTable.NumRows() )
        for i in kantenfeld:
            print "kante geprueft"
            if i.ungueltig == True:
                print "ungueltig"
                continue
            
            row_left = self.storeTable.ReadRowAsDict( punktefeld[i.datenPunktLinks].index )
            row_right = self.storeTable.ReadRowAsDict( punktefeld[i.datenPunktRechts].index )
            
            class_links = self.classification.FindGroup( row_left[attr] )
            class_rechts =  self.classification.FindGroup( row_right[attr] )
            
            if class_links.Matches(row_right[attr]) and class_rechts.Matches(row_left[attr]):
                i.ungueltig = True
                GruppenBilder.gehoerenZusammen( i.datenPunktLinks, i.datenPunktRechts )
                print "kein Uebergang"
                continue
            

        for i in kantenfeld:
            if i.ungueltig == True:
                #kantenfeld.remove( i )
                continue
            
            row_left = self.storeTable.ReadRowAsDict( punktefeld[i.datenPunktLinks].index )
            row_right = self.storeTable.ReadRowAsDict( punktefeld[i.datenPunktRechts].index )
            main_row = row_left

            if GruppenBilder.wievielMitgliederHatGruppe(i.datenPunktRechts) < GruppenBilder.wievielMitgliederHatGruppe(i.datenPunktLinks):
                main_row = row_right
            
            xs = [i.anfangspunkt.x(),i.endpunkt.x()]
            ys = [i.anfangspunkt.y(),i.endpunkt.y()]
            self.targetStore.addLine( xs, ys )
            self.targetStore.appendAttributeData( (nummer, row_right[attr], row_left[attr], main_row[attr]) )
            nummer += 1
            print "kante zugefuegt!"
        
    
    def zeichneDiagram( self, punktefeldEin, kantenfeldEin ):
        punktefeld = self.datenWerkzeug.hebeNormierungAuf( punktefeldEin )
        kantenfeld = self.datenWerkzeug.unnormieresKantenfeld( kantenfeldEin )
        nummer = 0
        
        self.targetStore.setTable( MemoryTable([
                             ("LinienNummer", FIELDTYPE_INT),
                             ("Punkt_rechts", FIELDTYPE_INT),
                             ("Punkt_links", FIELDTYPE_INT)],[]) )
          
        # zeichnen des voronoi-diagramms  
        for i in kantenfeld:
            if i.ungueltig == True:
                continue
                
            xs = [i.anfangspunkt.x(),i.endpunkt.x()]
            ys = [i.anfangspunkt.y(),i.endpunkt.y()]
            self.targetStore.addLine( xs, ys )
            #self.targetStore.table.data.append( (nummer, punktefeld[i.datenPunktRechts].index, punktefeld[i.datenPunktLinks].index) )
            self.targetStore.appendAttributeData( (nummer, i.datenPunktRechts, i.datenPunktLinks) )
            #print i.anfangspunkt.x, i.anfangspunkt.y, "|", i.endpunkt.x, i.endpunkt.y
            nummer += 1

    
    def zeichneHoehenLinien( self, punktefeldEin, kantenfeldEin, wert=None, x=None, y=None ):
        punktefeld = self.datenWerkzeug.hebeNormierungAuf( punktefeldEin )
        kantenfeld = self.datenWerkzeug.unnormieresKantenfeld( kantenfeldEin )
        print "uebernehme Attribute..."
        origColumns = self.storeTable.Columns()
        origColumn_map = {}
        
        for i in range( len(origColumns) ):
            origColumn_map[origColumns[i].name] = self.storeTable.Column(i)
        print "Attribute uebernommen."
        
        attr = self.auswertungsAttribut
        
        interpolierer = delaunay_interpolierer( attr, self.storeTable, punktefeld, kantenfeld )
        
        if ( wert == None ):
            punkteInDerUmgebung = None

            # step nicht zu klein waehlen, da thuban sonst rundet und komische ergebnisse produziert...
            step = 1.0
            boxRadius = step
            
            while (punkteInDerUmgebung == None or len(punkteInDerUmgebung) == 0) and step < 50:
                #(minx, miny, maxx, maxy)
                punkteInDerUmgebung = list(self.store.ShapesInRegion( (x-boxRadius, y-boxRadius, x+boxRadius, y+boxRadius) ))
                if self.debug == True:        
                    print "box", (x-boxRadius, y-boxRadius, x+boxRadius, y+boxRadius)
                boxRadius += step
            
            ausgangsPunkt = punkt( [x, y] )
            self.hoehenlinienWert = interpolierer.berechneWertAnPunkt( ausgangsPunkt, punkteInDerUmgebung )
        else:
            self.hoehenlinienWert = wert
            interpolierer.setHoehenlinienWert(wert)
        
        zielSpalte = origColumn_map[attr]
        self.targetStore.setTable( MemoryTable([(attr, zielSpalte.type)],[]) )
        anfangsPunktIndex = 0
        endPunktIndex = 0
        
        neueKanten = []
        
        for i in kantenfeld:
            if i.ungueltig == True or i.markiert == True:
                continue
            anfangsPunktIndex = i.datenPunktRechts
            endPunktIndex = i.datenPunktLinks
            row_anfang = self.storeTable.ReadRowAsDict(anfangsPunktIndex)
            row_ende = self.storeTable.ReadRowAsDict(endPunktIndex)
            
            wertAnfang = row_anfang[attr]
            wertEnde = row_ende[attr]
            
            if wertAnfang >= interpolierer.getHoehenlinienWert() and wertEnde <= interpolierer.getHoehenlinienWert():
                if self.debug == True:
                    print "1)", wertAnfang, wertEnde, i
                interpolierer.findeLinienBeginnendAnKante( i, neueKanten )
            elif wertAnfang <= interpolierer.getHoehenlinienWert() and wertEnde >= interpolierer.getHoehenlinienWert():
                if self.debug == True:
                    print "2)", wertAnfang, wertEnde, i
                interpolierer.findeLinienBeginnendAnKante( i, neueKanten )
            i.markiert = True

       
        
        for i in neueKanten:
            xs = [i.anfangspunkt.x(),i.endpunkt.x()]
            ys = [i.anfangspunkt.y(),i.endpunkt.y()]
            self.targetStore.addLine( xs, ys )
            self.targetStore.appendAttributeData( [self.hoehenlinienWert] )
            
        for i in kantenfeld:
            i.markiert = False
        
        minWert, maxWert = self.storeTable.ValueRange( attr )
        werteRange = float(maxWert-minWert)
        aktWert = self.hoehenlinienWert
        
        tmpWert = aktWert - minWert
        faktor = 1.0*tmpWert/werteRange
                    
        self.classification.SetDefaultLineColor(Color( faktor, 1.-faktor, 0))
        self.classification.SetDefaultLineWidth(2)
            
        return self.hoehenlinienWert
            
                

class delaunay_interpolierer:
    def __init__(self, attr, table, punktefeld, kantenfeld ):
        self.auswertungsAttribut = attr
        self.hoehenlinienWert = None
        self.storeTable = table
        self.debug = False
        
        self.punktefeld = punktefeld
        self.kantenfeld = kantenfeld
        
    def getHoehenlinienWert( self ):
        return self.hoehenlinienWert
        
    def setHoehenlinienWert( self, wert ):
        self.hoehenlinienWert = wert
        
    def interpoliereWert( self, kleinerWert, grosserWert, gesuchterWert ):
        wertDifferenz = grosserWert - kleinerWert
        faktorGrundlage = gesuchterWert - kleinerWert
        return (faktorGrundlage/wertDifferenz)
        
    def sucheWeitereKante( self, punkt1GID, punkt2GID ):
        # punkte als GID indexe
        gefunden1 = False
        gefunden2 = False
        punkt1 = -1
        punkt2 = -1
        for i in range(len(self.punktefeld)):
            if self.punktefeld[i].index == punkt1GID:
                gefunden1 = True
                punkt1 = self.punktefeld[i]
            if self.punktefeld[i].index == punkt2GID:
                gefunden2 = True
                punkt2 = self.punktefeld[i]
            if gefunden1==True and gefunden2==True:
                break
                
        punkt1_punkte = punkt1.verbundenePunkte()
        punkt2_punkte = punkt2.verbundenePunkte()
        
        gemeinsamePunkte = []
        
        for i in punkt1_punkte:
            if i in punkt2_punkte:
                gemeinsamePunkte.append(i)
                
        if self.debug == True:
            print "gemeinsame Punkte gefunden:", len( gemeinsamePunkte ), ":", gemeinsamePunkte
        
        suspekteKanten = []
        for i in self.kantenfeld:
            #if i.ungueltig == True or i.markiert == True:
            if i.ungueltig == True:
                continue
            # vorauswahl
            if i.datenPunktRechts in gemeinsamePunkte or i.datenPunktLinks in gemeinsamePunkte:
                if self.debug == True:
                    print i
                if i.datenPunktRechts == punkt1.index or i.datenPunktRechts == punkt2.index or \
                    i.datenPunktLinks == punkt1.index or i.datenPunktLinks == punkt2.index:
                    suspekteKanten.append(i)
        
        if self.debug == True:
            print "-", len(suspekteKanten), "kanten kommen in frage"
        passendeKanten = []
        
        for i in suspekteKanten:
            for j in suspekteKanten:
                if i == j:
                    continue
                if ( j.datenPunktRechts == i.datenPunktRechts and j.datenPunktRechts != punkt1.index and j.datenPunktRechts != punkt2.index ) or\
                    ( j.datenPunktLinks == i.datenPunktLinks and j.datenPunktLinks != punkt1.index and j.datenPunktLinks != punkt2.index) or\
                    ( j.datenPunktLinks == i.datenPunktRechts and j.datenPunktLinks != punkt1.index and j.datenPunktLinks != punkt2.index) or\
                    ( j.datenPunktRechts == i.datenPunktLinks and j.datenPunktRechts != punkt1.index and j.datenPunktRechts != punkt2.index):
                    if not ( i in passendeKanten ):
                        passendeKanten.append(i)
                    if not ( j in passendeKanten ):
                        passendeKanten.append(j)
                    
        
        return passendeKanten
        
    def findeInterpolationsPunkt( self, kante ):
        anfangsPunktIndex = kante.datenPunktRechts
        endPunktIndex = kante.datenPunktLinks
        row_anfang = self.storeTable.ReadRowAsDict(anfangsPunktIndex)
        row_ende = self.storeTable.ReadRowAsDict(endPunktIndex)
        attr = self.auswertungsAttribut
        wertAnfang = row_anfang[attr]
        wertEnde = row_ende[attr]
        iPunkt = []
        
        if max(wertAnfang, wertEnde)<self.hoehenlinienWert or min(wertAnfang, wertEnde)>self.hoehenlinienWert:
            return iPunkt
        elif wertAnfang==self.hoehenlinienWert and wertEnde==self.hoehenlinienWert:
            iPunkt.append( kante.anfangspunkt )
            iPunkt.append( kante.endpunkt )
            if self.debug == True:
                print "beide Punkte???"
        elif wertAnfang==self.hoehenlinienWert:
            iPunkt.append( kante.anfangspunkt )
        elif wertEnde==self.hoehenlinienWert:
            iPunkt.append( kante.endpunkt )
        else:
            kleinerWert = self.hoehenlinienWert
            grosserWert = self.hoehenlinienWert
            if wertAnfang<self.hoehenlinienWert and wertEnde>self.hoehenlinienWert:
                a = kante.anfangspunkt
                b = kante.endpunkt
                kleinerWert = wertAnfang
                grosserWert = wertEnde
            else:
                a = kante.endpunkt
                b = kante.anfangspunkt
                kleinerWert = wertEnde
                grosserWert = wertAnfang
            
            ab = punkt( [b.x() - a.x(), b.y() - a.y()] )
            abFaktor = self.interpoliereWert( kleinerWert, grosserWert, self.hoehenlinienWert )
            iPunkt.append( punkt( [a.x() + abFaktor * ab.x(), a.y() + abFaktor * ab.y()] ) )
        return iPunkt
        
    def berechneWertAnPunkt( self, ausgangsPunkt, punkteInDerUmgebung ):
        attr = self.auswertungsAttribut
    
        kleinsterAbstand = -1
        naechsterPunkt = None
        
        for i in punkteInDerUmgebung:
            aktX, aktY = i.Points()[0][0]
            aktPunkt = punkt( [aktX, aktY], i.ShapeID() )
            #print i.ShapeID(), i.Points()[0]
            aktAbstand = ausgangsPunkt.abstandZu( aktPunkt )
            if kleinsterAbstand < 0 or aktAbstand < kleinsterAbstand:
                kleinsterAbstand = aktAbstand
                naechsterPunkt = aktPunkt
                
        if self.debug == True:        
            print "naechsterPunkt", naechsterPunkt
        if naechsterPunkt == None:
            print "- konnte nicht den naechsten Punkt bestimmen, da offenbar keine Punktdaten fuer den gesuchten Bereich ermittelt werden konnten (DB Problem?)"
            print punkteInDerUmgebung
            for i in punkteInDerUmgebung:
                print i
            return
        
        
        evtlBeteiligteKanten = []
        for i in self.kantenfeld:
            if i.datenPunktRechts == naechsterPunkt.index or i.datenPunktLinks == naechsterPunkt.index:
                evtlBeteiligteKanten.append(i)
        
        kante1 = None
        kante2 = None
        kante3 = None        
        for i in evtlBeteiligteKanten:
            for j in evtlBeteiligteKanten:
                kante1 = None
                kante2 = None
                kante3 = None
                if i == j:
                    continue
                kante3 = self._findeVerbindendeKante( i, j )
                if kante3 != None:
                    #print "3. kante gefunden"
                    if self._punktLiegtInDreieck( ausgangsPunkt, i, j, kante3 ) == True:
                        kante1 = i
                        kante2 = j
                        break
                    
            if kante1 != None and kante2 != None:
                break
        
        if kante1 == None or kante2 == None or kante3 == None:
            print "- berechneWertAnPunkt(): konnte kein dreieck konstruieren - uebernehme wert des naechsten punktes"
            if self.debug == True:        
                print kante1, kante2, kante3
            wertAmNaechstenPunkt = self.storeTable.ReadValue( naechsterPunkt.index, attr )
            self.hoehenlinienWert = wertAmNaechstenPunkt
        else:
            print "- berechneWertAnPunkt(): interpoliere Wert am Punkt"
            
            self.hoehenlinienWert = self._interpoliereWertInDreieck( ausgangsPunkt, kante1, kante2, kante3 )
            
        return self.hoehenlinienWert
            
            
    def _interpoliereWertInDreieck( self, testPunkt, kante1, kante2, kante3 ):
        #  ----->       -------->      -------->   -------->
        #  punkt1 + a * punkt_1_2 + b* punkt_2_3 = testPunkt
        # a und b fuer x und y loesen, dann 
        
        punkt1 = kante1.anfangspunkt
        punkt1.index = kante1.datenPunktRechts
        punkt2 = kante1.endpunkt
        punkt2.index = kante1.datenPunktLinks
        
        if kante2.datenPunktRechts == kante3.datenPunktRechts:
            punkt3 = kante2.anfangspunkt
            punkt3.index = kante2.datenPunktRechts
        elif kante2.datenPunktRechts == kante3.datenPunktLinks:
            punkt3 = kante2.anfangspunkt
            punkt3.index = kante2.datenPunktRechts
        elif kante2.datenPunktLinks == kante3.datenPunktRechts:
            punkt3 = kante2.endpunkt
            punkt3.index = kante2.datenPunktLinks
        elif kante2.datenPunktLinks == kante3.datenPunktLinks:
            punkt3 = kante2.endpunkt
            punkt3.index = kante2.datenPunktLinks
        else:
            print "punktLiegtInDreieck(): kann kein dreieck konstruieren, ", kante2, kante3
            return False
        a = 0.
        b = 0.
        
        punkt_1_2 = punkt( [punkt2.x() - punkt1.x(), punkt2.y() - punkt1.y()] )
        punkt_2_3 = punkt( [punkt3.x() - punkt2.x(), punkt3.y() - punkt2.y()] )
        
        if punkt_1_2.x() != 0 and punkt_2_3.y() != (punkt_2_3.x()*punkt_1_2.y()/punkt_1_2.x()):
            b = (punkt1.y()-testPunkt.y()-((punkt1.x()-testPunkt.x())*punkt_1_2.y()/punkt_1_2.x()))/(-punkt_2_3.y()+(punkt_2_3.x()*punkt_1_2.y()/punkt_1_2.x()))
        elif punkt_2_3.y() != 0:
            b = ( punkt1.x() - testPunkt.x() )/(-punkt_2_3.x())
        else:
            print "punkt_1_2.x == 0 oder (-N.y+(N.x*punkt_1_2.y/punkt_1_2.x)==0"
        
        if punkt_1_2.x() != 0:
            a = (punkt1.x()+b*punkt_2_3.x()-testPunkt.x())/(-punkt_1_2.x())
        elif punkt_1_2.y() != 0 and (-punkt_2_3.x())*punkt_2_3.y() != testPunkt.y():
            a = (punkt1.y()+(punkt1.x()-testPunkt.x())/(-punkt_2_3.x())*punkt_2_3.y()-testPunkt.y() )/(-punkt_1_2.y())
        else:
            print "punkt_1_2.x == 0 oder (-N.y+(N.x*punkt_1_2.y/punkt_1_2.x)==0"
            
        # hole werte:
        attr = self.auswertungsAttribut
        wert1 = self.storeTable.ReadValue( punkt1.index, attr )
        wert2 = self.storeTable.ReadValue( punkt2.index, attr )
        wert_1_2 = wert2 - wert1
        wert3 = self.storeTable.ReadValue( punkt3.index, attr )
        wert_2_3 = wert3 - wert2
        
        return ( wert1 + a*wert_1_2 + b*wert_2_3 )
        
        
    def _findeVerbindendeKante( self, kante1, kante2 ):
        punkt1 = -1
        punkt2 = -1
        
        if kante1.datenPunktRechts == kante2.datenPunktRechts:
            punkt1 = kante1.datenPunktLinks
            punkt2 = kante2.datenPunktLinks
        elif kante1.datenPunktRechts == kante2.datenPunktLinks:
            punkt1 = kante1.datenPunktLinks
            punkt2 = kante2.datenPunktRechts
        elif kante1.datenPunktLinks == kante2.datenPunktRechts:
            punkt1 = kante1.datenPunktRechts
            punkt2 = kante2.datenPunktLinks
        elif kante1.datenPunktLinks == kante2.datenPunktLinks:
            punkt1 = kante1.datenPunktRechts
            punkt2 = kante2.datenPunktRechts
        else:
            return None
            
        for i in self.kantenfeld:
            if (i.datenPunktRechts == punkt1 and i.datenPunktLinks == punkt2) or (i.datenPunktRechts == punkt2 and i.datenPunktLinks == punkt1):
                return i
                
        return None
    
    def _punktLiegtInDreieck( self, testPunkt, kante1, kante2, kante3 ):
        eckpunkt1 = kante1.anfangspunkt
        eckpunkt2 = kante1.endpunkt
        
        #vektor_1_2 = punkt( eckpunkt2.x - eckpunkt1.x, eckpunkt2.y - eckpunkt1.y )
        vektor_1_2 = kante( eckpunkt1.x(), eckpunkt1.y(), eckpunkt2.x(), eckpunkt2.y(), -1, -1 )
        
        if kante2.datenPunktRechts == kante3.datenPunktRechts:
            eckpunkt3 = kante2.anfangspunkt
        elif kante2.datenPunktRechts == kante3.datenPunktLinks:
            eckpunkt3 = kante2.anfangspunkt
        elif kante2.datenPunktLinks == kante3.datenPunktRechts:
            eckpunkt3 = kante2.endpunkt
        elif kante2.datenPunktLinks == kante3.datenPunktLinks:
            eckpunkt3 = kante2.endpunkt
        else:
            if self.debug == True:        
                print "punktLiegtInDreieck(): kann kein dreieck konstruieren, ", kante1, kante2, kante3
            return False
            
        #vektor_2_3 = punkt( eckpunkt3.x - eckpunkt2.x, eckpunkt3.y - eckpunkt2.y )
        vektor_2_3 = kante( eckpunkt2.x(), eckpunkt2.y(), eckpunkt3.x(), eckpunkt3.y(), -1, -1 )
        
        #vektor_3_1 = punkt( eckpunkt1.x - eckpunkt3.x, eckpunkt1.y - eckpunkt3.y )
        vektor_3_1 = kante( eckpunkt3.x(), eckpunkt3.y(), eckpunkt1.x(), eckpunkt1.y(), -1, -1 )
        
        faktor1 = self._bestimmeYFaktor( vektor_1_2, testPunkt )
        faktor2 = self._bestimmeYFaktor( vektor_2_3, testPunkt )
        faktor3 = self._bestimmeYFaktor( vektor_3_1, testPunkt )
        
        faktoren = [ faktor1, faktor2, faktor3 ]
        
        if 0 in faktoren:
            return True
        elif max( faktoren ) < 0:
            return True
        elif min( faktoren ) > 0:
            return True
            
        return False
        
        
    def _bestimmeYFaktor( self, kante, zuPruefenderPunkt ):
        # a + x* ab + y * N = c
        y = 0.
        
        a = kante.anfangspunkt
        b = kante.endpunkt
        c = zuPruefenderPunkt
        
        ab = punkt( [b.x() - a.x(), b.y() - a.y()] )
        # positive N-Richtung (y) heisst links von ab
        N = punkt( [-ab.y(), ab.x()] )
        
        if ab.x() != 0:
            y = (a.y()-c.y()-((a.x()-c.x())*ab.y()/ab.x()))/(-N.y()+(N.x()*ab.y()/ab.x()))
            #x = (a.x+y*N.x-c.x)/(-ab.x)
        elif ab.y() != 0:
            y = ( a.x() - c.x() )/(-N.x())
            #x = (a.y+(a.x-c.x)/(-N.x)*N.y-c.y )/(-ab.y)
        else:
            print "ab.x == 0 oder (-N.y+(N.x*ab.y/ab.x)==0"
        
        return y    
    
    def findeLinienBeginnendAnKante( self, testKante, neueKanten ):
        anfangsPunktIndex = testKante.datenPunktRechts
        endPunktIndex = testKante.datenPunktLinks
        row_anfang = self.storeTable.ReadRowAsDict(anfangsPunktIndex)
        row_ende = self.storeTable.ReadRowAsDict(endPunktIndex)
        attr = self.auswertungsAttribut
        wertAnfang = row_anfang[attr]
        wertEnde = row_ende[attr]
        
        neueLinie_anfangspunkt = None
        neueLinie_endpunkt = None
        endPunktGefunden = False
        moeglichePunkte = self.findeInterpolationsPunkt( testKante )
        
        if len(moeglichePunkte) == 2:
            neueLinie_anfangspunkt = moeglichePunkte[0]
            neueLinie_endpunkt = moeglichePunkte[1]
        elif len(moeglichePunkte) == 1:
            neueLinie_anfangspunkt = moeglichePunkte[0]
            suspekteKanten = self.sucheWeitereKante( anfangsPunktIndex, endPunktIndex )
            #print "-", len(suspekteKanten),"suspekte kanten gefunden"
            moeglicheEndpunkte = []
            for i in suspekteKanten:
                moeglicheEndpunkteTmp = []
                moeglicheEndpunkteTmp = self.findeInterpolationsPunkt( i )
                if len( moeglicheEndpunkteTmp )==2:
                    neueKante = kante( moeglicheEndpunkteTmp[0].x, moeglicheEndpunkteTmp[0].y, moeglicheEndpunkteTmp[1].x, moeglicheEndpunkteTmp[1].y, -1, -1 )
                    neueKanten.append( neueKante )
                else:
                    # kann im Grunde nur 1 Punkt sein...
                    for j in moeglicheEndpunkteTmp:
                        moeglicheEndpunkte.append(j)
            if moeglicheEndpunkte == []:
                if self.debug == True:
                    print "keine Endpunkte"
                return
            elif len( moeglicheEndpunkte ) >= 1:
                if self.debug == True:
                    print len( moeglicheEndpunkte ), "Endpunkte gefunden"
                for i in range( len(moeglicheEndpunkte) ):
                    neueKante = kante( neueLinie_anfangspunkt.x(), neueLinie_anfangspunkt.y(), moeglicheEndpunkte[i].x(), moeglicheEndpunkte[i].y(), -1, -1 )
                    neueKanten.append( neueKante )

        return
