import math

# schleifen beschleunigen
try:
    from psyco.classes import __metaclass__
except:
    pass

# statistik modul
from statistik_module import *
from fehlerhaftesObjekt import *

class statistischeTests(potentiellFehlerhaftesObjekt):
    """
    Interface fuer statistische Tests, z.B. Chi-Quadrat-Unabhaengigkeit und die Varianzanalyse
    """
    def __init__(self):
        potentiellFehlerhaftesObjekt.__init__(self)
        self._pruefSumme = 0
        self._pruefsummeBerechnet = False
        return

    def getPruefsumme( self ):
        if self._pruefsummeBerechnet == False:
            self._pruefSumme = self._berechnePruefsumme()
            self._pruefsummeBerechnet = True
        return self._pruefSumme
        
    def _berechnePruefsumme():
        self.fehler = fehler( "_berechnePruefsumme(): Diese Funktion muss von der abgeleiteten Klasse implementiert werden" )
        return
        
    def signifikanzZurPruefsumme( self ):
        self.fehler = fehler("signifikanzZurPruefsumme(): Diese Funktion muss von der abgeleiteten Klasse implementiert werden")
        return 0.

class VarianzAnalyse(statistischeTests):
    def __init__( self, werte ):
        statistischeTests.__init__( self )
        """
        werte - matrix mit den zu anlysierenden stichproben
        """
        self.werte = werte
        self.X_quer = 0
        self.n = 0
        self.k = len( werte )
        
        self.X_quer_i = []
        self.ni = []
        
        self._SST = 0
        self._SSE = 0
        
        for i in werte:
            self.n += len(i)
            self.ni.append( len(i) )
            
        self._dfnum = (self.k-1)
        self._dfden = (self.n - self.k)
            
    def _berechnePruefsumme( self ):
        self.X_quer = self._berechneXQuer( self.werte, self.n )
        self.X_quer_i = self._berechneXQuer_i( self.werte, self.n )
        self._SST = self._berechneSST( self.werte, self.ni, self.X_quer_i, self.X_quer )
        self._SSE = self._berechneSSE( self.werte, self.X_quer_i )
        
        return (( self._SST / self._dfnum ) / ( self._SSE / self._dfden ))
        
    def _berechneSST( self, werte, ni, X_quer_i, X_quer ):
        SST = 0
        
        for i in range( len( werte ) ):
            SST += ni[i] * pow((X_quer_i[i]-X_quer),2)
            
        return SST
        
    def _berechneSSE( self, werte, X_quer_i ):
        SSE = 0
        for i in range(len( werte )):
            for j in range( len( werte[i] ) ):
                SSE += pow( werte[i][j] - X_quer_i[i], 2 )
                
        return SSE
            
        
    def _berechneXQuer_i( self, werte, n ):
        X_quer_i = []
        for i in range(len( werte )):
            tmp_summe = 0
            for j in range( len( werte[i] ) ):
                tmp_summe += werte[i][j]
                
            tmp_summe /= len( werte[i] )
            X_quer_i.append( tmp_summe )
                
        return X_quer_i
        
    def _berechneXQuer( self, werte, n ):
        X_quer = 0.0
        for i in range(len( werte )):
            for j in range( len( werte[i] ) ):
                X_quer += werte[i][j]
                
        return ( 1.0 * X_quer / n )
        
    def signifikanzZurPruefsumme( self ):
        return lfprob( self._dfnum, self._dfden, self.getPruefsumme() )
        
    def getXYFreiheitsgrade( self ):
        return [ self._dfnum, self._dfden ]

class ChiQuadratTest(statistischeTests):
    def __init__( self, werte, mengeWerte, mins, maxs, kathLists = None ):
        statistischeTests.__init__(self)

        self.n = mengeWerte
        self.mins = mins
        self.maxs = maxs
        self.werte = werte
        self.numkath = [ 0, 0 ]

        freiheit = int(round(math.sqrt(math.sqrt( self.n ))))

        self.diskreteWerte = [ False, False ]
        if kathLists != None:
            if kathLists["0"] != []:
                self.diskreteWerte[0] = True
                self.numkath[0] = len( kathLists["0"] )
            else:
                self.numkath[0] = freiheit + 1
            if kathLists["1"] != []:
                self.diskreteWerte[1] = True
                self.numkath[1] = len( kathLists["1"] )
            else:
                self.numkath[1] = freiheit + 1
        else:
            self.numkath = [ freiheit + 1, freiheit + 1 ]

        self.freiheitsGrade = (self.numkath[0]-1)*(self.numkath[1]-1)

        # haeufigkeiten
        
        if self.diskreteWerte[0] == False:
            self.a = self.zaehleHaeufigkeitenAus( werte[0], self.mins[0], self.maxs[0], self.numkath[0] )
        else:
            self.a = self.zaehleHaeufigkeitenAus( werte[0], diskreteWerteListe = kathLists["0"] )
        if self.diskreteWerte[1] == False:
            self.b = self.zaehleHaeufigkeitenAus( werte[1], self.mins[1], self.maxs[1], self.numkath[1] )
        else:
            self.b = self.zaehleHaeufigkeitenAus( werte[1], diskreteWerteListe = kathLists["1"] )
        self.c = self.erstelleHaeufigkeitenMatrix( werte,  self.mins, self.maxs, self.numkath, kathLists, self.diskreteWerte )

        self._pruefsummeBerechnet = False
    
    def signifikanzZurPruefsumme( self ):
        return lchisqprob(self.getPruefsumme(), self.freiheitsGrade)

    def getFreiheitsgrade( self ):
        return self.freiheitsGrade

    def getXYFreiheitsgrade( self ):
        return self.numkath

    def _berechnePruefsumme( self ):
        D = 0
        
        for k in range( self.numkath[0] ):
            for l in range( self.numkath[1] ):
                if (self.a[k]) != 0 and (self.b[l]) != 0:
                    erwarteteAnzahl = (1.0 * self.a[k] * self.b[l]/self.n)
                    D += pow( self.c[k][l] - (erwarteteAnzahl), 2 ) / erwarteteAnzahl
                else:
                    # wenn eine spalte null ist, muss automatisch auch der andere spaltenwert
                    # und der wert in der kreuztabelle null sein
                    continue

        return D

    def zaehleHaeufigkeitenAus( self, stichproben, minWert = 0, maxWert = 0, mengeKath = 0, diskreteWerteListe = None ):
        haeufigkeiten = []
        
        if diskreteWerteListe == None:
            Range = maxWert - minWert
            step = Range / float(mengeKath)
        else:
            Range = 0
            step = 0
            mengeKath = len( diskreteWerteListe )
    
        for i in range( mengeKath ):
           haeufigkeiten.append(0)
    
        for i in stichproben:
            kath = self.bestimmeKathegorie( Range, minWert, i, step, mengeKath, diskreteWerteListe )
            haeufigkeiten[ kath ] += 1
            
        return haeufigkeiten
        
    def bestimmeKathegorie( self, Range, minWert, wert, step, numkath, diskreteWerteListe = None ):
        if diskreteWerteListe == None:
            ergebnis = (wert-minWert)/step
            if ergebnis == numkath:
                ergebnis -= 1
        else:
            if wert in diskreteWerteListe:
                ergebnis = diskreteWerteListe.index(wert)
                #print diskreteWerteListe, wert, ergebnis
            else:
                self.fehler = fehler( "Achtung: diskrete Kathegorie fuer den Wert"+ wert +"konnte nicht ermittelt werden!" )
                ergebnis = -1

        return int(ergebnis)

    def erstelleHaeufigkeitenMatrix( self, stichproben, minWert, maxWert, mengeKath, kathLists = None, diskreteWerte = None ):
        if diskreteWerte == None or diskreteWerte[0] == False:
            Range1 = maxWert[0] - minWert[0]
            step1 = Range1 / float(mengeKath[0])
        else:
            Range1 = 0
            step1 = 0
        
        if diskreteWerte == None or diskreteWerte[1] == False:
            Range2 = maxWert[1] - minWert[1]
            step2 = Range2 / float(mengeKath[1])
        else:
            Range2 = 0
            step2 = 0

        c = []
        # matrix initialisieren
        for k in range( mengeKath[0] ):
            c.append( [] )
            for l in range( mengeKath[1] ):
                c[k].append(0)
    
        # matrix fuellen
        for i in range( min( len(stichproben[0]), len(stichproben[1]) )):
            if diskreteWerte == None or diskreteWerte[0] == False:
                kath1 = self.bestimmeKathegorie( Range1, minWert[0], stichproben[0][i], step1, mengeKath[0] )
            else:
                kath1 = self.bestimmeKathegorie( Range1, minWert[0], stichproben[0][i], step1, mengeKath[0], kathLists["0"] )
            if diskreteWerte == None or diskreteWerte[1] == False:
                kath2 = self.bestimmeKathegorie( Range2, minWert[1], stichproben[1][i], step2, mengeKath[1] )
            else:
                kath2 = self.bestimmeKathegorie( Range2, minWert[1], stichproben[1][i], step2, mengeKath[1], kathLists["1"] )
            c[kath1][kath2] += 1

        return c


    
