# Author: Ole Rahn
# Project: PIROL

from feldsortierer import *
from fehlerhaftesObjekt import *
from fehler import FEHLER_POLYGON_OFFEN
import os

from time import clock
from wxPython.wx import wxBell

import math

# schleifen beschleunigen
try:
    from psyco.classes import __metaclass__
except:
    pass

from Thuban.Model.data import SHAPETYPE_POINT


def istZahl( string ):
    if ( string.isdigit() ):
        return True
    elif ( (string.startswith("-") or string.startswith("+")) and string[1:].isdigit ):
        return True
    elif( string.count(".") == 1 ):
        
        ersterTeil = string[ :string.find(".") ]
        zweiterTeil = string[ string.find(".")+1: ]
        
        if ( 
            (len(ersterTeil)==0 or ersterTeil.isdigit() or ((string.startswith("-") or string.startswith("+")) and ersterTeil[1:].isdigit)) and
            (len(zweiterTeil)==0 or zweiterTeil.isdigit() ) 
            ):
            return True
    
    print "keine Zahl:", string
    return False
        

def piep():
    #print "- piep():", os.system("echo -en \"\\007\"")
    wxBell()

class datenTool:
    """
    Klasse mit Hilfsfunktionen zur Vor-, Nachbearbeitung der Daten
    """

    def __init__(self, punktefeld, bbox, vorbereitenZumAnnaehern = True):
        self.punktefeld = punktefeld
        self.multiplikator = 100.0

        self.sicherheitsAbstand = 100.0

        if vorbereitenZumAnnaehern == True:
            self.bboxErweiterung = 10.0
        else:
            self.bboxErweiterung = 0

        xOffset = bbox[0]
        yOffset = bbox[1]

        self.origBbox = bbox

        self.minX = bbox[0]-self.bboxErweiterung
        self.minY = bbox[1]-self.bboxErweiterung
        self.maxX = bbox[2]+self.bboxErweiterung
        self.maxY = bbox[3]+self.bboxErweiterung


        self.offset = xOffset
        if xOffset > yOffset:
            self.offset = yOffset

        self.KantenSindMultipliziert = False
        self.PunkteSindMultipliziert = False
        self.BboxIstMultipliziert = False

        self.KantenSindNormiert = False
        self.PunkteSindNormiert = False
        self.BboxIstNormiert = False

    def getBbox( self ):
        self.kehreMultiplizierungUm()
        return [ self.minX, self.minY, self.maxX, self.maxY ]

    def getNormierteBbox( self ):
        if self.BboxIstMultipliziert == False and self.BboxIstNormiert == False:
            self.normiereUndMultipliziere()
        return [ self.minX, self.minY, self.maxX, self.maxY ]

    def normiereUndMultipliziere( self ):
        if self.BboxIstMultipliziert == False and self.BboxIstNormiert == False:
            self.minX = (self.origBbox[0]-self.bboxErweiterung-self.offset)*self.multiplikator
            self.minY = (self.origBbox[1]-self.bboxErweiterung-self.offset)*self.multiplikator
            self.maxX = (self.origBbox[2]+self.bboxErweiterung-self.offset)*self.multiplikator
            self.maxY = (self.origBbox[3]+self.bboxErweiterung-self.offset)*self.multiplikator

        self.KantenSindMultipliziert = True
        self.PunkteSindMultipliziert = True
        self.BboxIstMultipliziert = True

        self.KantenSindNormiert = True
        self.PunkteSindNormiert = True
        self.BboxIstNormiert = True

        for i in self.punktefeld:
            i.normiert = True
            i.multipliziert = True
            # abschneiden eines gemeinsamen offsets, um die Zahlen fuer spaetere Berechnungen moeglichst klein zu halten
            if i.normiert == True:
                i.setX( i.x() - self.offset )
                i.setY( i.y() - self.offset )
            # spreizen der abstaende, um kleinste abstaende zu vergroessern
            if i.multipliziert == True:
                i.setX( i.x() * self.multiplikator )
                i.setY( i.y() * self.multiplikator )

    def getNormiertesPunktefeld( self ):
        self.normiereUndMultipliziere()
        return self.punktefeld

    def getUnnormiertesPunktefeld( self ):
        self.kehreMultiplizierungUm()
        self.kehreNormierungUm()
        return self.punktefeld

    def unnormieresKantenfeld( self, kantenfeld ):
        for i in kantenfeld:
            if i.anfangspunkt.multipliziert == True:
                i.anfangspunkt.setX( i.anfangspunkt.x() / self.multiplikator )
                i.anfangspunkt.setY( i.anfangspunkt.y() / self.multiplikator )
                i.anfangspunkt.multipliziert = False
            if i.endpunkt.multipliziert == True:
                i.endpunkt.setX( i.endpunkt.x() / self.multiplikator )
                i.endpunkt.setY( i.endpunkt.y() / self.multiplikator )
                i.endpunkt.multipliziert = False
        for i in kantenfeld:
            if i.anfangspunkt.normiert == True:
                i.anfangspunkt.setX( i.anfangspunkt.x() + self.offset )
                i.anfangspunkt.setY( i.anfangspunkt.y() + self.offset )
                i.anfangspunkt.normiert = False
            if i.endpunkt.normiert == True:
                i.endpunkt.setX( i.endpunkt.x() + self.offset )
                i.endpunkt.setY( i.endpunkt.y() + self.offset )
                i.endpunkt.normiert = False
        return kantenfeld

    def hebeNormierungAuf( self, punktefeld ):
        for i in punktefeld:
            if i.multipliziert == True:
                i.setX( i.x() / self.multiplikator )
                i.setY( i.y() / self.multiplikator )
                i.multipliziert = False
            if i.normiert == True:
                i.setX( i.x() + self.offset )
                i.setY( i.y() + self.offset )
                i.normiert = False

        return punktefeld

    def kehreMultiplizierungUm( self ):
        if self.PunkteSindMultipliziert == True:
            for i in self.punktefeld:
                if i.multipliziert == True:
                    i.setX( i.x() / self.multiplikator )
                    i.setY( i.y() / self.multiplikator )
                    i.multipliziert = False

            self.PunkteSindMultipliziert = False
        if self.BboxIstMultipliziert == True:
            self.minX /= self.multiplikator
            self.minY /= self.multiplikator
            self.maxX /= self.multiplikator
            self.maxY /= self.multiplikator
            self.BboxIstMultipliziert = False

    def kehreNormierungUm( self ):
        if self.PunkteSindNormiert == True:
            for i in self.punktefeld:
                if i.normiert == True:
                    i.x += self.offset
                    i.y += self.offset
                    i.normiert = False

            self.PunkteSindNormiert = False
        if self.BboxIstNormiert == True:
            self.minX += self.offset
            self.minY += self.offset
            self.maxX += self.offset
            self.maxY += self.offset
            self.BboxIstNormiert = False

# -------------------- Punkte ---------------------------
class punkt:
    def __init__(self, koords, index = -1 ):
        self.index = index
        self.koord = []

        for i in koords:
            self.koord.append(i)

        self.dimension = len( self.koord )

        self.normiert = False
        self.multipliziert = False
        self.markiert = False

    def markiere(self):
        self.markiert = True

    def unmarkiere(self):
        self.markiert = False

    def x(self):
        return self.koord[0]

    def y(self):
        return self.koord[1]

    def setX( self, neuerWert ):
        self.koord[0] = neuerWert

    def setY( self, neuerWert ):
        self.koord[1] = neuerWert

    def __str__(self):
        SKoord = ""
        for i in self.koord:
            if self.koord.index(i) != 0:
                SKoord = SKoord + ", %1.2f" % (i)
            else:
                SKoord = SKoord + "%1.2f" % (i)

        if self.index >= 0:
            return "P" + str(self.index) + "[" + SKoord + "] "
        else:
            return "P" + "[" + SKoord + "]"


    def komponente(self, index):
        return self.koord[index]

    def getXYTuple(self):
        return (self.x(), self.y())

    def getKoordinaten( self ):
        return self.koord

    def abstandZu( self, punkt ):
        quadratSumme = 0

        if self.dimension == 2:
            # 90% der faelle
            # schneller?
            x_diff = punkt.koord[0] - self.koord[0]
            y_diff = punkt.koord[1] - self.koord[1]
            quadratSumme = x_diff * x_diff + y_diff * y_diff
        else:
            for i in range( min( self.dimension, punkt.dimension ) ):
                quadratSumme +=  math.pow( punkt.koord[i] - self.koord[i], 2 )

        return math.sqrt( quadratSumme )

class clusterPunkt(punkt):
    def __init__( self, koords, index = -1, anzahlCluster=0):
        punkt.__init__( self, koords, index )

        self.anzahlCluster = anzahlCluster

        self.zugehoerigkeiten = []

        for i in range( self.anzahlCluster ):
            self.zugehoerigkeiten.append( 0.0 )

    def GetZugehoerigkeit( self, zentrum ):
        return float(self.zugehoerigkeiten[zentrum])

    def SetZugehoerigkeit( self, zentrum, zug ):
        self.zugehoerigkeiten[zentrum] = zug
        
    def GetPrimaereZugehoerigkeit( self ):
        max = 0.0
        ind = 0
        
        for i in range( len(self.zugehoerigkeiten) ):
            akt = self.zugehoerigkeiten[i]
            if ( akt > max ):
                max = akt
                ind = i
                
        return ind


class delaunayPunkt( punkt ):
    def __init__(self, koords, index = -1):
        punkt.__init__( self, koords, index )
        self.endpunkt = False
        self.gehoertZuDreieck = []
        self.schonVerbundenMit = []

    def _gehoertZu( self, nummer ):
        if nummer in self.gehoertZuDreieck:
            return True
        return False

    def verbindenMit( self, nummer ):
        if not nummer in self.schonVerbundenMit:
            self.schonVerbundenMit.append( nummer )

    def verbundenePunkte( self ):
        return self.schonVerbundenMit

    def istVerbundenMit( self, nummer ):
        for i in self.schonVerbundenMit:
            if i == nummer:
                return True
        return False

    def gehortZuMehrAlsEinemDreieckMit( self, punkt ):
        zaehler = 0
        for i in self.gehoertZuDreieck:
            if punkt._gehoertZu(i)==True:
                zaehler += 1

        if zaehler < 2:
            return False
        else:
            return True


# -----------------------------------------------

class kante:

    def __init__(self, x_anfang, y_anfang, x_ende, y_ende, dr, dl, inf_anfang = False, inf_ende = False):
        self.anfangspunkt = punkt( [x_anfang, y_anfang] )
        self.anfangspunkt.normiert = True
        self.anfangspunkt.multipliziert = True
        self.endpunkt = punkt( [x_ende, y_ende] )
        self.endpunkt.normiert = True
        self.endpunkt.multipliziert = True

        self.datenPunktRechts = dr
        self.datenPunktLinks = dl

        self.infinite_anfang = inf_anfang
        self.infinite_ende = inf_ende
        self.ungueltig = False

        self.markiert = False
        # added by bernd for more precision
        self.orig_xanfang = repr(x_anfang)
        self.orig_yanfang = repr(y_anfang)
        self.orig_xende = repr(x_ende)
        self.orig_yende = repr(y_ende)
        
    def getLaenge( self ):
        return self.anfangspunkt.abstandZu( self.endpunkt )
        
    def bestimmeSchnittpunkt( self, kante ):
        kante1 = self
        kante2 = kante
        # loesen der gleichung
        # ->       ->   ->       ->
        # a  + n * ab = c  + m * cd
        a = kante1.anfangspunkt
        ab = punkt( [kante1.endpunkt.x() - a.x(), kante1.endpunkt.y() - a.y()] )
        c = kante2.anfangspunkt
        cd = punkt( [kante2.endpunkt.x() - c.x(), kante2.endpunkt.y() - c.y()] )
        
        if a.abstandZu(kante1.endpunkt) == 0:
            # or c.abstandZu(kante2.endpunkt)==0:
            print "Nullkante"
            return None
            
        # faktor m errechnen
        m = 0.0
        if ab.x() == 0 and cd.x()!=0:
            m = (a.x()-c.x())/cd.x()
        elif ab.x()==0 and cd.x()==0:
            print "Achtung: sonderfall ab.x == 0 and cd.x == 0, kanten", kante1.datenPunktRechts, kante1.datenPunktLinks, "|", kante2.datenPunktRechts, kante2.datenPunktLinks
            if a.x() != c.x():
                return None
            elif (c.y() >= a.y() and c.y() <= a.y() + ab.y()) or (c.y() <= a.y() and c.y() >= a.y() + ab.y()):
                return punkt( [a.x(), c.y()] )
            elif (a.y() >= c.y() and a.y() <= c.y() + cd.y()) or (a.y() <= c.y() and a.y() >= c.y() + cd.y()):
                return punkt( [a.x(), a.y()] )
            else:
                return None
                 
        elif ab.y()==0 and cd.y() != 0:
            m = (a.y()-c.y())/cd.y()
        elif ab.y()==0 and cd.y()==0:
            # dieser fall sollte nicht vorkommen
            print "falsch!"
            #m = (a.x-c.x+(-a.x+c.x)/ab.x)/(cd.x-cd.x/ab.x)
            m = 0
        elif ab.x()!=0 and (cd.x()*ab.y()/ab.x())!=cd.y():
            # scheint ok zu sein
            m = ( a.y()-c.y()+((c.x()-a.x())*ab.y())/ab.x() )/( cd.y()-(cd.x()*ab.y()/ab.x()) )
        else:
            print "M##>", ab, "|", cd
            #self.setFehler( code = FEHLER_ALLG_RECHENFEHLER )
            return None
        
        if kante2.infinite_anfang!=True and m <= 0 :
            # kanten schneiden sich nicht zwischen Anfangs- und Endpunkt oder sowieso gar nicht
            #print "M weicht ab!"
            #if self.debug == True:
                #print "M: Kanten schneiden sich nicht"
            return None
        if kante2.infinite_ende!=True and m >= 1 :
            # kanten schneiden sich nicht zwischen Anfangs- und Endpunkt oder sowieso gar nicht
            #print "M weicht ab!"
            #if self.debug == True:
                #print "M2: Kanten schneiden sich nicht"
            return None


        # sicherheitshalber n berechnen und spaeter ergebnisse vergleichen
        n = 0.0
        if ab.x()==0 and ab.y()!=0 and cd.x()!=0:
            n = (a.y()-c.y()-((a.x()-c.x())*cd.y()/cd.x()))/(-ab.y())
        elif ab.x()!=0:
            n = (c.x()+m*cd.x()-a.x())/ab.x()
        elif cd.y()==0 and ab.y() != 0:
            n = (c.y()-a.y())/ab.y()
        else:
            print "N##>", ab, "|", cd
            #self.setFehler( code = FEHLER_ALLG_RECHENFEHLER )
            return None

        if kante1.infinite_anfang!=True and n <= 0 :
            # kanten schneiden sich nicht zwischen Anfangs- und Endpunkt oder sowieso gar nicht
            #print "n weicht ab!"
            #if self.debug == True:
                #print "N: kanten schneiden sich nicht"
            return None
        if kante1.infinite_ende!=True and n >= 1 :
            # kanten schneiden sich nicht zwischen Anfangs- und Endpunkt oder sowieso gar nicht
            #print "n weicht ab!"
            #if self.debug == True:
                #print "N2: Kanten schneiden sich nicht"
            return None


        schnittPunkt1 = punkt( [a.x() + n * ab.x(), a.y() + n * ab.y()] )
        
        schnittPunkt1.normiert = True
        schnittPunkt1.multipliziert = True
        return schnittPunkt1

        
    def berechneNormalenFaktor( self, pkt ):
        punkt1 = self.anfangspunkt
        punkt2 = self.endpunkt
        punkt3 = pkt
        #           (C)
        #          / |  \
        #        /   |     \
        #      /     |y*N     \
        #   (A)______|_________(B)
        # ab ist vektor punkt1 -> punkt2 = a -> b
        ab = punkt( [punkt2.x()-punkt1.x(), punkt2.y()-punkt1.y()], -1 )
        # N ist normale zu ab
        N = punkt( [ab.y(), -ab.x()], -2 )
        # berechne faktor y mit dem N multipliziert werden muss, um punkt3 zu erreichen

        try:
            ergebnis = (((punkt1.y()- punkt3.y())-( (punkt1.x() - punkt3.x()) * ab.y() / ab.x() ))/ (N.x() * ab.y() / ab.x() - N.y()))
        except:
            # faengt division durch null fehler ab, die hier aber nicht auftauchen koennen sollten
            ergebnis = 1
            print "berechneNormalenFaktor(): Fehler", ",", (ab.x() - N.y()), ",", ab.x(), ",", ab.y()

        return ergebnis
        
    def getNormalenKante( self, laenge=1.0 ):
        punkt1 = self.anfangspunkt
        punkt2 = self.endpunkt

        ab = punkt( [punkt2.x()-punkt1.x(), punkt2.y()-punkt1.y()], -1 )

        N = punkt( [ab.y(), -ab.x()], -2 )

        nullPunkt = punkt( [0.0, 0.0] )
        nLaenge = nullPunkt.abstandZu( N )
        try:
            faktor = laenge / nLaenge * -1.0
        except:
            faktor = -1.0
        
        N.setX( N.x()*faktor )
        N.setY( N.y()*faktor )
        
        nKante = kante( punkt2.x(), punkt2.y(), punkt2.x()+N.x(), punkt2.y()+N.y(), -1, -1 )
        
        #print "laenge neue Kante:", nKante.getLaenge(), ", vorgegeben:", laenge
        
        return nKante
        
    def __str__(self):
        return "kante(" + str(self.datenPunktRechts)+ "," + str(self.datenPunktLinks)+ "," + str(self.anfangspunkt)+ "," + str(self.endpunkt) + ")"

# -----------------------------------------------

class polygon(potentiellFehlerhaftesObjekt):
    """
    besteht aus einem Feld aus Kanten, die einen geschlossenen Linien Zug bilden
    """
    
    def __init__( self, kanten ):
        potentiellFehlerhaftesObjekt.__init__( self )
        self.kanten = kanten
        self.numKanten = len(kanten)
        
        self.debug = False
        
        self.pi = math.pi
        
        self.minX = None
        self.minY = None
        self.maxX = None
        self.maxY = None
        
        self.index = -1
        
        for knte in self.kanten:
            if (knte.getLaenge() == 0):
                print "Nullkante raus"
                self.kanten.remove(knte)
        
        if ( not self._istPolygonGeschlossen() ):
            self.setFehler( code=FEHLER_POLYGON_OFFEN )
        else:
            self._setBbox()
            
    def setIndex( self, index ):
        self.index = index
        
    def getIndex( self ):
        return self.index

    def BoundingBox( self ):
        return ( self.minX, self.minY, self.maxX, self.maxY )
        
    def _setBbox( self ):
        for pkt in self.getPunkte():
            if ( self.minX == None or pkt.x() < self.minX ):
                self.minX = pkt.x()
            if ( self.minY == None or pkt.y() < self.minY ):
                self.minY = pkt.y()
            if ( self.maxX == None or pkt.x() > self.maxX ):
                self.maxX = pkt.x()
            if ( self.maxY == None or pkt.y() > self.maxY ):
                self.maxY = pkt.y()

    def _istPolygonGeschlossen( self ):
        start = self.kanten[0].anfangspunkt
        ende  = self.kanten[len(self.kanten)-1].endpunkt
        
        if ( start.x() == ende.x() and start.y() == ende.y() ):
            return True
        else:
            return False
        
    def __str__( self ):
        string = "---- Polygon, "+repr(self.numKanten)+" Kanten:\n"
        
        for kante in self.kanten:
            string += str( kante )+"\n"
            
        string += "----"
        
        return string
        
    def getPunkte( self ):
        punkte = []
        for kante in self.kanten:
            punkte.append( kante.anfangspunkt )
            
        return punkte
        
    def _wirdKanteGeschnitten( self, k ):
        for knte in self.kanten:
            if ( knte.bestimmeSchnittpunkt( k) != None ):
                return True
        return False
        
    def liegtPunktDrin( self, pkt ):
        """
        Idee:
            - Setze punkt als Ursprung (verschiebe punkt + polygon)
            - Summiere Winkel zu den Punkten d. Polygons auf
                - wenn 0 -> punkt liegt nicht drin
                - wenn 360 -> punkt liegt drin
        """
        
        # zuerst mal schauen, ob der punkt ueberhaupt drin liegen koennte...
        if ( pkt.x() < self.minX or pkt.y() < self.minY or pkt.x() > self.maxX or pkt.y() > self.maxY ):
            # ... noe
            return False
        
        # ... koennte sein

        eckpunkte = []
        # anti-callByReference: kopieren der Koordinaten aus den "punkt"-Objekten
        # gleichzeitig: verschieben
        for epunkt in self.getPunkte():
            neuX = (epunkt.x()-pkt.x())
            neuY = (epunkt.y()-pkt.y())
            neuPunkt = punkt( [neuX, neuY] )
            eckpunkte.append(neuPunkt)
            
        winkel = .0
        
        ursprung = punkt( [.0,.0] )
        
        if ( len(eckpunkte) < 3 ):
            print "nur 2 oder weniger Eckpunkte!!"
            return False

        for i in range( len(eckpunkte) ):
            aktuellerPunkt = eckpunkte[i]
            if ( i+1 < len(eckpunkte) ):
                naechsterPunkt = eckpunkte[i+1]
            else:
                naechsterPunkt = eckpunkte[0]
                
            knte = kante( aktuellerPunkt.x(), aktuellerPunkt.y(), naechsterPunkt.x(), naechsterPunkt.y(), -1, -1 )
            
            c = aktuellerPunkt.abstandZu( ursprung )
            b = naechsterPunkt.abstandZu( ursprung )
            a = aktuellerPunkt.abstandZu( naechsterPunkt )
            
            if ( a == 0 ):
                continue
            
            faktor = knte.berechneNormalenFaktor( ursprung )
            
            if ( math.fabs(faktor) < 0.000001 ):
                #print "faktor==0"
                return False
            
            try:
                oben = (a*a-b*b-c*c)
                unten = (-2.0*b*c)
                if ( oben == 0 and unten == 0 ):
                    #return True
                    return False
                elif ( math.fabs( math.fabs(oben)-math.fabs(unten)) < 0.000001 ):
                    # wenn divident und divisor praktisch gleich sind (divident/divisor=1) und es aufgrund von 
                    # float darstellungsfehlern zu bereichsueberschreitungen kommt, werden hier bei erhaltung der vorzeichen
                    # beide auf 1 reduziert -> keine bereichsueberschreitung mehr...
                    oben /= math.fabs(oben)
                    unten /= math.fabs(unten)

                aktWinkel = math.acos( oben/unten )
            except:
                self.setFehler( "Fehler bei der Berechnung des Winkels, Divisor:"+repr(-2.0*b*c)+", cosinus:"+repr(a*a-b*b-c*c)+"/Divisor" );
                return False
            
            if ( faktor < 0. ):
                aktWinkel = -aktWinkel
            
            #print "teilwinkel", i, ":", (aktWinkel/self.pi*180.0), ",", faktor
            
            winkel += aktWinkel
        
        if ( self.debug ):    
            print "gesamtwinkel:", (winkel/self.pi*180.0)
        
        if ( math.fabs( math.fabs(winkel) - 2.0*self.pi) < 0.000001 ):
            if ( self.debug ):    
                print "punkt liegt drin!"
            return True
            
        if ( self.debug ):    
            print "punkt liegt nicht drin!"
        return False

    def schneidetKante( self, potSchneider ):
        for kante in self.kanten:
            if ( kante.bestimmeSchnittpunkt(potSchneider) != None ):
                return True
        return False
        
    def liegtPolygonDrin( self, poly ):
        # ausgehend von der basis-flaeche pruefen, ob Quadrat drin liegt?!
        punkte = poly.getPunkte()
        
        for pkt in punkte:
            if ( self.liegtPunktDrin(pkt)==True ):
                return True
                
        for knte in poly.kanten:
            if ( knte.getLaenge() == 0 ):
                continue
            if ( self._wirdKanteGeschnitten( knte ) ):
                return True
            
        return False
# -----------------------------------------------

class rechteckErzeuger:
    def __init__( self, kantenLaengeSenkr, kantenLaengePar ):
        self.laengeS = kantenLaengeSenkr
        self.laengeP = kantenLaengePar
        
    def erzeugeRechtEck( self, ausgangskante ):
        """
        die Idee ist: am Ende der Kante immer in positiver Normalen-Richtung eine neue Kante anzuhaengen,
        bis man - nach drei neuen kanten - am anfangspunkt der ausgangskante ankommt
        
        am Ende alles als Polygon zurueckgeben
        """
        kanten = []
        kanten.append( ausgangskante )
        
        for i in range(3):
            aktKante = kanten[ len(kanten)-1 ]
            
            laenge = 0.0
            
            if ( i%2 == 0 ):
                laenge = self.laengeP
            else:
                laenge = self.laengeS
            
            if ( i < 2 ):
                kanten.append( aktKante.getNormalenKante( laenge ) )
            else:
                neueKante = aktKante.getNormalenKante( laenge )
                neueKante.endpunkt.setX( kanten[0].anfangspunkt.x() )
                neueKante.endpunkt.setY( kanten[0].anfangspunkt.y() )
                kanten.append( neueKante )
                
        return polygon( kanten )
                
            

# -----------------------------------------------

class stopuhr:
    def __init__( self ):
        self.startzeit = 0
        self.stopzeit = 0
        self.differenz = 0


    def start( self ):
        self.startzeit = clock()

    def stop( self ):
        self.stopzeit = clock()
        self.differenz = self.stopzeit - self.startzeit
        return self.differenz

    def getDauer( self ):
        return self.differenz
