# Author: Ole Rahn
# Project: PIROL

import math
from feldsortierer import *
from utilities import *

from diagramm_oberklasse import *

# schleifen beschleunigen
try:
    from psyco.classes import __metaclass__
except:
    pass


class voronoi_diagram(diagramm):
    def __init__( self, punktefeld, bbox ):
        diagramm.__init__(self)

        self.genauigkeit = 1.0/4096.0
        self.multiplikator = 100.0
        
        self.verbose = False
        if len( punktefeld ) > 500:
            self.verbose = True
        
        self.sicherheitsAbstand = 100.0
        self.bboxErweiterung = 10.0
        
        self.minX = bbox[0]
        self.minY = bbox[1]
        self.maxX = bbox[2]
        self.maxY = bbox[3]
                
        self.punktefeld = punktefeld
        
        self.mengePunkte = len( self.punktefeld )
       
        self.geloeschteKanten = 0
        
        self.berechnetFuer = 0
        self.fertigBerechnetFuer = 0
        self.berechnungAbgeschlossen = False

    def start( self ):
        self.berechneDiagramm()
        return 0
        
    def getProzentFertig( self ):
        if self.fehlerIstAufgetreten():
            return 100
        prozent = int( round(100.0 * self.berechnetFuer / float(self.mengePunkte)))
        return prozent

    def getKantenfeld( self ):
        resultierendesKantenfeld = []
        
        aktKante = None
        for i in self.kantenfeld:
            if i.ungueltig == True:
                continue
            aktKante = kante( i.anfangspunkt.x(), i.anfangspunkt.y(),  i.endpunkt.x(),  i.endpunkt.y(), i.datenPunktRechts, i.datenPunktLinks, i.infinite_anfang, i.infinite_ende)
            resultierendesKantenfeld.append(aktKante)
            
        return resultierendesKantenfeld


    def bestimmeBegrenzungsLinien( self, wo, annaehern = True ):
        """
            wo:
            0 = oben,
            1 = rechts,
            2 = unten,
            3 = links
        """
        # a + x*ab + y*N = c
        if self.debug == True:
            print "bestimmeBegrenzungsLinien()"
        
        if annaehern == True:
            # fuer genauere Linienbegrenzung...
            minX = self.minX - (self.bboxErweiterung * self.multiplikator)
            maxX = self.maxX + (self.bboxErweiterung * self.multiplikator)
            minY = self.minY - (self.bboxErweiterung * self.multiplikator)
            maxY = self.maxY + (self.bboxErweiterung * self.multiplikator)
        else:
            # ... die ist allerdings nicht geeignet fuer kleine Punktmengen
            minX = self.minX
            maxX = self.maxX
            minY = self.minY
            maxY = self.maxY
        
        medY = (minY + maxY) / 2
        medX = (minX + maxX) / 2
        
        step = .7 * self.bboxErweiterung * self.multiplikator
        
        if wo == 0:
            print "- bestimme obere Begrenzung"
            step *= -1
            a = punkt( [minX, maxY] )
            b = punkt( [maxX, maxY] )
        elif wo == 2:
            print "- bestimme untere Begrenzung"
            b = punkt( [minX, minY] )
            a = punkt( [maxX, minY] )
        elif wo == 1:
            print "- bestimme rechte Begrenzung"
            step *= -1
            b = punkt( [maxX, minY] )
            a = punkt( [maxX, maxY] )
        elif wo == 3:
            print "- bestimme linke Begrenzung"
            a = punkt( [minX, minY] )
            b = punkt( [minX, maxY] )
        
        if annaehern == True:
            if wo == 0 or wo == 2:
                durchgang = 1
                while durchgang < 3:
                    grenzeUeberschritten = False
                    while grenzeUeberschritten == False:
                        if durchgang == 1:
                            a.setY( a.y() + step )
                        else:
                            b.setY( b.y() + step )
                        for c in self.punktefeld:
                            if wo == 0 and c.y() < medY - medY/4 :
                                continue
                            elif wo == 2 and c.y() > medY + medY/4:
                                continue
                            ab = punkt( [b.x() - a.x(), b.y() - a.y()] )
                            N = punkt( [-ab.y(), ab.x()] )
                            y = 0
                            if ab.x() != 0:
                                y = (a.y()-c.y()-((a.x()-c.x())*ab.y()/ab.x()))/(-N.y()+(N.x()*ab.y()/ab.x()))
                            elif ab.y() != 0:
                                y = (a.x() - c.x())/-N.x()
                            else:
                                print "ab ist (0,0) ->", a_punkt, "=", b_punkt, ", SCHWERER FEHLER!"
                                continue
                            if y > 0:
                                if durchgang == 1:
                                    a.setY( a.y() - step )
                                    durchgang += 1
                                else:
                                    b.setY( b.y() - step )
                                    durchgang += 1
                                grenzeUeberschritten = True
                                break
            
            elif wo == 1 or wo == 3:
                durchgang = 1
                while durchgang < 3:
                    grenzeUeberschritten = False
                    while grenzeUeberschritten == False:
                        if durchgang == 1:
                            a.setX( a.x() + step )
                        else:
                            b.setX( b.x() + step )
                        for c in self.punktefeld:
                            if wo == 0 and c.x() < medX - medX/4:
                                continue
                            elif wo == 2 and c.x() > medX + medX/4:
                                continue
                            ab = punkt( [b.x() - a.x(), b.y() - a.y()] )
                            N = punkt( [-ab.y(), ab.x()] )
                            y = 0
                            if ab.x() != 0:
                                y = (a.y()-c.y()-((a.x()-c.x())*ab.y()/ab.x()))/(-N.y()+(N.x()*ab.y()/ab.x()))
                            elif ab.y() != 0:
                                y = (a.x() - c.x())/-N.x()
                            else:
                                print "ab ist (0,0) ->", a_punkt, "=", b_punkt, ", SCHWERER FEHLER!"
                                continue
                            if y > 0:
                                if durchgang == 1:
                                    a.setX( a.x() - step )
                                    durchgang += 1
                                else:
                                    b.setX( b.x() - step )
                                    durchgang += 1
                                grenzeUeberschritten = True
                                break

        return kante( a.x(), a.y(), b.x(), b.y(), -1, -1, True, True )
                 
    def bestimmeYFaktor( self, kante, zuPruefenderPunkt ):
        # a + x* ab + y * N = c
        y = 0.
        
        a = kante.anfangspunkt
        b = kante.endpunkt
        c = zuPruefenderPunkt
        
        ab = punkt( [b.x() - a.x(), b.y() - a.y()] )
        # positive N-Richtung (y) heisst links von ab
        N = punkt( [-ab.y(), ab.x()] )
        
        if ab.x() != 0:
            y = (a.y()-c.y()-((a.x()-c.x())*ab.y()/ab.x()))/(-N.y()+(N.x()*ab.y()/ab.x()))
            #x = (a.x+y*N.x-c.x)/(-ab.x)
        elif ab.y() != 0:
            y = ( a.x() - c.x() )/(-N.x())
            #x = (a.y+(a.x-c.x)/(-N.x)*N.y-c.y )/(-ab.y)
        else:
            print "ab.x == 0 oder (-N.y+(N.x*ab.y/ab.x)==0"
            self.setFehler( meldung = "schwerer Fehler bei der Berechnung des Diagramms")
        
        return y            
           
    
    def begrenzeLinien( self, annaehern = True ):
        """
        abschneiden der "ueberstehenden" linien enden
        """
        if len( self.punktefeld )==0:
            print "   keine Punkte zur Begrenzung da..."
            return
            
        if self.debug == True:
            print "- berechne Begrenzungslinien"

        #linkerRand = kante( minX, maxY, minX, minY, -1, -1, False, False )
        linkerRand = self.bestimmeBegrenzungsLinien( 3, annaehern )
        #rechterRand = kante( maxX, minY, maxX, maxY, -1, -1, False, False )
        rechterRand = self.bestimmeBegrenzungsLinien( 1, annaehern )
        #obererRand = kante( maxX, maxY, minX, maxY, -1, -1, False, False )
        obererRand = self.bestimmeBegrenzungsLinien( 0, annaehern )
        #untererRand = kante( minX, minY, maxX, minY, -1, -1, False, False )
        untererRand = self.bestimmeBegrenzungsLinien( 2, annaehern )

        for i in self.kantenfeld:
            if i.ungueltig == True:
                continue
            elif max(i.anfangspunkt.x(), i.endpunkt.x()) < min(linkerRand.anfangspunkt.x(), linkerRand.endpunkt.x()) or \
                 min(i.anfangspunkt.x(), i.endpunkt.x()) > max(rechterRand.anfangspunkt.x(), rechterRand.endpunkt.x()) or \
                 min(i.anfangspunkt.y(), i.endpunkt.y()) > max(obererRand.anfangspunkt.y(), obererRand.endpunkt.y()) or \
                 max(i.anfangspunkt.y(), i.endpunkt.y()) < min(untererRand.anfangspunkt.y(), untererRand.endpunkt.y()):
                i.ungueltig = True
                continue
            
            if min(i.endpunkt.x(), i.anfangspunkt.x()) < max(linkerRand.anfangspunkt.x(), linkerRand.endpunkt.x()):
                schnittpunkt = self.bestimmeSchnittpunkt( linkerRand, i )
                if schnittpunkt != None:
                    neuerPunkt = punkt( [schnittpunkt.x(), schnittpunkt.y()] )
                    neuerPunkt.normiert = True
                    neuerPunkt.multipliziert = True
                    if i.anfangspunkt.x() < i.endpunkt.x():
                        i.anfangspunkt = neuerPunkt
                        i.infinite_anfang = False
                    else:
                        i.endpunkt = neuerPunkt
                        i.infinite_ende = False
            if max(i.anfangspunkt.x(), i.endpunkt.x()) > min(rechterRand.anfangspunkt.x(), rechterRand.endpunkt.x()):
                schnittpunkt = self.bestimmeSchnittpunkt( rechterRand, i )       
                if schnittpunkt != None:
                    neuerPunkt = punkt( [schnittpunkt.x(), schnittpunkt.y()] )
                    neuerPunkt.normiert = True
                    neuerPunkt.multipliziert = True
                    if i.anfangspunkt.x() > i.endpunkt.x():
                        i.anfangspunkt = neuerPunkt
                        i.infinite_anfang = False
                    else:
                        i.endpunkt = neuerPunkt
                        i.infinite_ende = False                        
            if max(i.anfangspunkt.y(), i.endpunkt.y()) > min(obererRand.anfangspunkt.y(), obererRand.endpunkt.y()):
                schnittpunkt = self.bestimmeSchnittpunkt( obererRand, i )       
                if schnittpunkt != None:
                    neuerPunkt = punkt( [schnittpunkt.x(), schnittpunkt.y()] )
                    neuerPunkt.normiert = True
                    neuerPunkt.multipliziert = True
                    if i.anfangspunkt.y() > i.endpunkt.y():
                        i.anfangspunkt = neuerPunkt
                        i.infinite_anfang = False
                    else:
                        i.endpunkt = neuerPunkt
                        i.infinite_ende = False
            if min(i.anfangspunkt.y(), i.endpunkt.y()) < max(untererRand.anfangspunkt.y(), untererRand.endpunkt.y()):
                schnittpunkt = self.bestimmeSchnittpunkt( untererRand, i )
                if schnittpunkt != None:
                    neuerPunkt = punkt( [schnittpunkt.x(), schnittpunkt.y()] )
                    neuerPunkt.normiert = True
                    neuerPunkt.multipliziert = True
                    if i.anfangspunkt.y() < i.endpunkt.y():
                        i.anfangspunkt = neuerPunkt
                        i.infinite_anfang = False
                    else:
                        i.endpunkt = neuerPunkt
                        i.infinite_ende = False
                    
        for i in self.kantenfeld:
            if i.ungueltig == True:
                #self.kantenfeld.remove(i)
                continue

            y1 = 0.
            y2 = 0.
            
            if min(i.anfangspunkt.x(), i.endpunkt.x()) < max(linkerRand.anfangspunkt.x(), linkerRand.endpunkt.x()):
                y1 = self.bestimmeYFaktor( linkerRand, i.anfangspunkt )
                y2 = self.bestimmeYFaktor( linkerRand, i.endpunkt ) 
            if (y1 <=0 or y2 <= 0) and max(i.anfangspunkt.x(), i.endpunkt.x()) > min(rechterRand.anfangspunkt.x(), rechterRand.endpunkt.x()):
                y1 = self.bestimmeYFaktor( rechterRand, i.anfangspunkt )
                y2 = self.bestimmeYFaktor( rechterRand, i.endpunkt )
            if (y1 <=0 or y2 <= 0) and max(i.anfangspunkt.y(), i.endpunkt.y()) > min(obererRand.anfangspunkt.y(), obererRand.endpunkt.y()):
                y1 = self.bestimmeYFaktor( obererRand, i.anfangspunkt )
                y2 = self.bestimmeYFaktor( obererRand, i.endpunkt )
            if (y1 <=0 or y2 <= 0) and min(i.anfangspunkt.y, i.endpunkt.y()) < max(untererRand.anfangspunkt.y(), untererRand.endpunkt.y()):
                y1 = self.bestimmeYFaktor( untererRand, i.anfangspunkt )
                y2 = self.bestimmeYFaktor( untererRand, i.endpunkt ) 

            if y1 > 0 and y2 > 0:
                #self.kantenfeld.remove(i)
                i.ungueltig = True
            
    def berechneDiagramm( self ):
        if len( self.punktefeld )==0:
            self.fehler = fehler("Es liegen keine Ausgangspunktedaten vor.")
            return
        print "- berechne Diagramm fuer", len( self.punktefeld ), "Punkte"
        sortierer = feld_sortierer( self.punktefeld )
        self.punktefeld = sortierer.sortiertesFeld()
        self.erstelleDiagramm( 0, len( self.punktefeld )-1 )
        print "-", self.geloeschteKanten, "geloescht"
        self.berechnungAbgeschlossen = True
        
    def fuelleDiagrammMitDatenbankDaten( self, kantenfeld ):
        diagramm.fuelleDiagrammMitDatenbankDaten( self, kantenfeld )

    def erstelleDiagramm( self, von, bis ):
        if von >= bis or self.fehlerIstAufgetreten():
            return
        trenner = int( math.floor( (von+bis)/2 ))
        self.erstelleDiagramm( von, trenner )
        self.erstelleDiagramm( trenner+1, bis )
        
        geschnitteneKanten = []
        neuerKantenZug = []
        
        if self.fehlerIstAufgetreten() != True:
            startPunkte = self.bestimmeObereTangentialPunkte( von, trenner, bis )
        
        if self.fehlerIstAufgetreten() or startPunkte==[] or startPunkte[0]<0 or startPunkte[1]<0:
            print "keine startpunkte!!!"
            self.setFehler( meldung = "es konnten keine Tangentialpunkte fuer die Erstellung eines Teil-Diagramms gefunden werden - ist die Projektion Gauss-Krueger?" )
            return

        done = False
        anfang = True
        a = startPunkte[0]
        b = startPunkte[1]
        letzterSchnittpunkt = None
        letzteSchnittKante = -1
        letzteKante = -1
        

        while done==False and self.fehlerIstAufgetreten()==False:
            # vektor von a nach b
            if self.debug == True:
                print "- von:", von, "bis", bis, "a: ", a, "b:", b
            ab = punkt( [self.punktefeld[b].x() - self.punktefeld[a].x(), self.punktefeld[b].y() - self.punktefeld[a].y()] )
            # punkt genau zwischen a und b
            mittePunkt = punkt( [self.punktefeld[a].x() + ab.x()/2, self.punktefeld[a].y() + ab.y()/2] )
            # normale zu ab
            normale = punkt( [ab.y(), -ab.x()] )
            
            if normale.y() == 0:
                if self.debug == True:
                    print "waagerecht!!!"

                if letzterSchnittpunkt == None:
                    anfang = False
                    
                    if self.punktefeld[a].y() > self.punktefeld[b].y():
                        neueKante = kante( self.maxX, mittePunkt.y(), self.minX, mittePunkt.y(), a, b, True, True )
                    else:
                        neueKante = kante( self.minX, mittePunkt.y(), self.maxX, mittePunkt.y(), a, b, True, True )
                else:
                    if normale.x() > 0:
                        if letzterSchnittpunkt.x() < self.maxX:
                            neueKante = kante( letzterSchnittpunkt.x(), letzterSchnittpunkt.y(), self.maxX, letzterSchnittpunkt.y(), a, b, False, True )
                        else:
                            neueKante = kante( letzterSchnittpunkt.x(), letzterSchnittpunkt.y(), letzterSchnittpunkt.x()+(self.sicherheitsAbstand), letzterSchnittpunkt.y(), a, b, False, True )
                    else:
                        if letzterSchnittpunkt.x() > self.minX:
                            neueKante = kante( letzterSchnittpunkt.x(), letzterSchnittpunkt.y(), self.minX, letzterSchnittpunkt.y(), a, b, False, True )
                        else:
                            neueKante = kante( letzterSchnittpunkt.x(), letzterSchnittpunkt.y(), letzterSchnittpunkt.x()-(self.sicherheitsAbstand), letzterSchnittpunkt.y(), a, b, False, True )

                    
            else:
                # faktor fuer normale zum schnittpunkt mit der unteren grenzkante
                if letzterSchnittpunkt == None or letzterSchnittpunkt.y() > self.minY:
                    faktor_unten = (self.minY - mittePunkt.y()) / normale.y()
                else:
                    faktor_unten = (letzterSchnittpunkt.y() - mittePunkt.y() - (self.sicherheitsAbstand)) / normale.y()
                # unterer Tangentenpunkt:
                u_tangente = punkt( [mittePunkt.x() + faktor_unten * normale.x(), mittePunkt.y() + faktor_unten * normale.y()] )

                if anfang == True:
                    anfang = False
                    # faktor fuer normale zum schnittpunkt mit der oberen grenzkante
                    faktor_oben = (self.maxY - mittePunkt.y()) / normale.y()
                    # oberer Tangentenpunkt:
                    o_tangente = punkt( [mittePunkt.x() + faktor_oben * normale.x(), mittePunkt.y() + faktor_oben * normale.y()] )
                    neueKante = kante( o_tangente.x(), o_tangente.y(), u_tangente.x(), u_tangente.y(), a, b, True, True )
                else:
                    # resultierende kante
                    neueKante = kante( letzterSchnittpunkt.x(), letzterSchnittpunkt.y(), u_tangente.x(), u_tangente.y(), a, b, False, True )
                    
            akt_schnittpunkt = None
            schnittpunkt = None
            schnittkante = -1
            anfangOderEnde = 0
            
            
            # Beschleunigung: der neue Schnittpunkt wird nach unten nur berechnet, wenn er noch im "interessanten" Bereich liegt,
            # da das Voronoi Diagram nach unten nicht vollstaendig sein muss - das fuehrt allerdings dazu, dass kanten geloescht werden,
            # die in Richtung der Trennkante nach unten auslaufen und deren Schnittpunkt (weil nicht berechnet) sie nicht mehr eindeutig einer Seite
            # zuordnet...
            if letzterSchnittpunkt==None or letzterSchnittpunkt.y() > self.minY:
                for i in range( len(self.kantenfeld) ):
    
                    if self.fehlerIstAufgetreten():
                        return
                    if self.kantenfeld[i].ungueltig == True:
                        continue
                    if self.kantenfeld[i].datenPunktRechts < von or self.kantenfeld[i].datenPunktRechts > bis:
                        continue
                    # gesuchte Kante muss entweder punkt a oder punkt b auf einer seite haben
                    if not (self.kantenfeld[i].datenPunktRechts == a or\
                            self.kantenfeld[i].datenPunktRechts == b or \
                            self.kantenfeld[i].datenPunktLinks == a or\
                            self.kantenfeld[i].datenPunktLinks == b):
                        continue
                    if i == letzteKante:
                        continue
                    if self.kantenfeld[i].infinite_ende == False and self.kantenfeld[i].infinite_anfang == False and\
                        neueKante.infinite_anfang == False and min( self.kantenfeld[i].anfangspunkt.y(), self.kantenfeld[i].endpunkt.y()) > neueKante.anfangspunkt.y():
                        # die aktuelle Kante liegt nicht im Bereich einer moeglichen Schnittkante (oberhalb eines moeglichen Schnittpunktes)
                        continue
                    akt_schnittpunkt = self.bestimmeSchnittpunkt( neueKante, self.kantenfeld[i] )
                    if akt_schnittpunkt == None:
                        continue
    
                    if schnittpunkt==None or akt_schnittpunkt.y() > schnittpunkt.y():
                        schnittpunkt = akt_schnittpunkt
                        schnittkante = i
                        

            # print "-> i:", schnittkante
            if schnittpunkt == None or i < 0:
                done = True
                self.kantenfeld.append( neueKante )
                neueKantenNr = self.kantenfeld.index( neueKante )
                        
                neuerKantenZug.append(neueKantenNr)
                        
                self.bereinigeKantenfeld( von, trenner, bis, neuerKantenZug, geschnitteneKanten)
                
                if self.debug == True:
                    print "letzter Punkt d. Kantenzugs gefunden"
                    
                if von == 0 or von == math.floor(self.mengePunkte/2)+1:
                    if bis != self.mengePunkte - 1:
                        self.fertigBerechnetFuer = bis
                    elif von == 0:
                        self.fertigBerechnetFuer = bis
                if von == self.fertigBerechnetFuer + 1:
                    spanne = bis - von + 1
                    self.berechnetFuer = self.fertigBerechnetFuer + 1 + spanne
                    
                    #print "- abschliessende Berechnung fuer die ersten", self.berechnetFuer, "von", self.mengePunkte, "Punkten"
                
                return
                
            else:
                daten_r = self.kantenfeld[schnittkante].datenPunktRechts
                daten_l = self.kantenfeld[schnittkante].datenPunktLinks
                
                if daten_r < 0 or daten_l < 0:
                    print "datenfehler!!"
                    print "daten_r:", daten_r, "daten_l:", daten_l, "schnittkante:", schnittkante, "len():", len(self.kantenfeld)
                    self.setFehler( "schwere Inkonsistenz der Daten" )
                    return

        
                if daten_r <= trenner and daten_l <= trenner:
                    # links
                    if daten_l == a:
                        a = daten_r
                    elif daten_r == a:
                        a = daten_l
                    else:

                        if self.debug == True:
                            print "a: ", a, "(b);", b, ",", "daten_r:", daten_r, "daten_l:", daten_l, "trenner:", trenner+0.5
                        if self.kantenfeld[schnittkante].infinite_anfang == False and self.kantenfeld[schnittkante].infinite_ende == False:
                            print "a: falscher schnitt!!"
                            self.setFehler( "es wurde eine falsche Kante geschnitten" )
                        if self.punktefeld[ daten_r ].y() < self.punktefeld[ daten_l ].y():
                            a = daten_l
                        elif self.punktefeld[ daten_r ].y() > self.punktefeld[ daten_l ].y():
                            a = daten_r
                        else:
                            print "ach nee!" 
                        print "a neu:", a
                        anfang = True
                        letzteKante = -1
                        continue


                    ab_faktor = self.bestimmeABFaktorZuPunkt( self.kantenfeld[schnittkante], schnittpunkt )
                        
                    if ab_faktor == "fehler":
                        print "falscher schnitt links"
                        self.setFehler("Falscher Kanten-Schnitt im linken Teil-Diagramm")
                        return

                    anfangOderEnde = self.begrenzeKante( schnittkante, neueKante, schnittpunkt )

                    if anfangOderEnde==0:
                        print "unklar ob anfang oder ende zu begrenzen!"
                        self.setFehler("Es konnte nicht festgestellt werden, ob Anfang oder Ende der Kante zu begrenzen sind")

                elif daten_r > trenner and daten_l > trenner:
                    # rechts
                    if daten_l == b:
                        b = daten_r
                    elif daten_r == b:
                        b = daten_l
                    else:
                        print "b: ", b, "(a);", a, ",",  "daten_r:", daten_r, "daten_l:", daten_l, "trenner:", trenner+0.5
                        if self.kantenfeld[schnittkante].infinite_anfang == False and self.kantenfeld[schnittkante].infinite_ende == False:
                            print "b: falscher schnitt!!"
                            self.setFehler( "es wurde eine falsche Kante geschnitten" )
                        if self.punktefeld[ daten_r ].y() > self.punktefeld[ daten_l ].y():
                            b = daten_r
                        elif self.punktefeld[ daten_r ].y() < self.punktefeld[ daten_l ].y():
                            b = daten_l
                        else:
                            print "Noe!!!!"
                        print "b neu:", b
                        anfang = True
                        letzteKante = -1
                        continue

                    # anfang und ende infinit
                    ab_faktor = self.bestimmeABFaktorZuPunkt( self.kantenfeld[schnittkante], schnittpunkt )
                        
                    if ab_faktor == "fehler":
                        print "flascher schnitt rechts"
                        self.setFehler("Falscher Kanten-Schnitt im rechten Teil-Diagramm")
                        return

                    anfangOderEnde = self.begrenzeKante( schnittkante, neueKante, schnittpunkt )

                    if anfangOderEnde == 0:
                        print "unklar ob anfang oder ende zu begrenzen!"
                        self.setFehler("Es konnte nicht festgestellt werden, ob Anfang oder Ende der Kante zu begrenzen sind")

                else:
                    print "----> daten_r", daten_r, "daten_l", daten_l, "trenner", trenner+0.5
                    self.setFehler("schwerer Fehler in der internen Datenhaltung")
                    return
                
                if neueKante.infinite_anfang == True:
                    # richtungsvektor der neuen kante
                    R = punkt( [neueKante.endpunkt.x() - neueKante.anfangspunkt.x(), neueKante.endpunkt.y() - neueKante.anfangspunkt.y()])
                    neuerPunkt = punkt( [schnittpunkt.x() - R.x(), schnittpunkt.y() - R.y()] )
                    neuerPunkt.normiert = True
                    neuerPunkt.multipliziert = True
                    neueKante.anfangspunkt = punkt( [schnittpunkt.x() - R.x(), schnittpunkt.y() - R.y()] )
                    if self.debug == True:
                        print "verschiebe anfang von", neueKante.datenPunktRechts, neueKante.datenPunktLinks

                neueKante.endpunkt = schnittpunkt
                geschnitteneKanten.append( schnittkante )
                
                neueKante.infinite_ende = False
                self.kantenfeld.append( neueKante )
                neueKantenNr = self.kantenfeld.index( neueKante )
                neuerKantenZug.append(neueKantenNr)
                    
                letzteKante = neueKantenNr
                
                letzterSchnittpunkt = schnittpunkt
                letzteSchnittKante = schnittkante
                letzterAnfangOderEnde = anfangOderEnde
                
    print "fertig"
    
    def bestimmeKanteAmYWert( self, kantenfeld, yWert ):
        """
        bestimmt welche kante eines gegebenen kantenzugs am gegebenen yWert liegt
        """
        for i in range( len( kantenfeld ) ):
            if self.kantenfeld[kantenfeld[i]].ungueltig == True:
                continue
            if self.kantenfeld[kantenfeld[i]].anfangspunkt.y() > self.kantenfeld[kantenfeld[i]].endpunkt.y():
                oben = self.kantenfeld[kantenfeld[i]].anfangspunkt.y()
                unten = self.kantenfeld[kantenfeld[i]].endpunkt.y()
                oben_infinite = self.kantenfeld[kantenfeld[i]].infinite_anfang
                unten_infinite = self.kantenfeld[kantenfeld[i]].infinite_ende
            else:
                oben = self.kantenfeld[kantenfeld[i]].endpunkt.y()
                unten = self.kantenfeld[kantenfeld[i]].anfangspunkt.y()
                unten_infinite = self.kantenfeld[kantenfeld[i]].infinite_anfang
                oben_infinite = self.kantenfeld[kantenfeld[i]].infinite_ende
            if oben >= yWert and unten <= yWert:
                return kantenfeld[i]
            elif oben >= yWert and unten_infinite == True:
                return kantenfeld[i]
            elif unten <= yWert and oben_infinite:
                return kantenfeld[i]
                
        print "bestimmeKanteAmYWert(): keine Kante gefunden!"
        return -1
        
               
    def bestimmeObPunktAufDerRichtigenSeiteDerKanteIst( self, trenner, zuPruefendeKante, anfangsOderEndpunkt, kantenfeld ):
        """
        vergleicht anhand der daten punkte (rechts, links) und dem trenner, ob ein gegebener Punkt
        auf der richtigen Seiten einer gegebenen trenn-Kante liegt
        """
        zuPruefenderPunkt = zuPruefendeKante.anfangspunkt
        if anfangsOderEndpunkt > 0:
            zuPruefenderPunkt = zuPruefendeKante.endpunkt
            
        trennKantenIndex = self.bestimmeKanteAmYWert( kantenfeld, zuPruefenderPunkt.y() )
        
        if trennKantenIndex < 0:
            # keine trennkante im kantenfeld gefunden - was nun??
            print "keine trennkante gefunden"
            return False
        trennKante = self.kantenfeld[ trennKantenIndex ]
        
        # a + x* ab + y * N = c
        
        a = trennKante.anfangspunkt
        b = trennKante.endpunkt
        c = zuPruefenderPunkt
        
        ab = punkt( [b.x() - a.x(), b.y() - a.y()] )
        # positive N-Richtung (y) heisst links von ab
        N = punkt( [-ab.y(), ab.x()] )
        
        if ab.x() != 0:
            y = (a.y()-c.y()-((a.x()-c.x())*ab.y()/ab.x()))/(-N.y()+(N.x()*ab.y()/ab.x()))
            #x = (a.x+y*N.x-c.x)/(-ab.x)
            #print "1) ab", ab, "y", y
        elif ab.y() != 0:
            y = ( a.x() - c.x() )/(-N.x())
            #x = (a.y+(a.x-c.x)/(-N.x)*N.y-c.y )/(-ab.y)
            #print "2) ab =", ab, "y", y
        else:
            print "ab.x == 0 oder (-N.y+(N.x*ab.y/ab.x)==0"
            self.setFehler( code = FEHLER_ALLG_RECHENFEHLER )
            return False
            
        if y > 0:
            #print "c liegt links der trenn-kante", anfangsOderEndpunkt
            if trennKante.datenPunktLinks <= trenner:
                #print "links ist unterer x-bereich"
                if zuPruefendeKante.datenPunktLinks <= trenner and zuPruefendeKante.datenPunktRechts <= trenner:
                    return True
                else:
                    #print "l1, trennKante:", trennKante
                    return False
            else:
                #print "links ist oberer x-bereich"
                if zuPruefendeKante.datenPunktLinks > trenner and zuPruefendeKante.datenPunktRechts > trenner:
                    return True
                else:
                    #print trenner, trennKante
                    #print "l2", a, b, c
                    return False

        elif y < 0:
            #print "c liegt rechts der trenn-kante", anfangsOderEndpunkt
            if trennKante.datenPunktRechts <= trenner:
                #print "rechts ist unterer x-bereich"
                if zuPruefendeKante.datenPunktLinks <= trenner and zuPruefendeKante.datenPunktRechts <= trenner:
                    return True
                else:
                    #print "r1"
                    return False
            else:
                #print "rechts ist oberer x-bereich"
                if zuPruefendeKante.datenPunktLinks > trenner and zuPruefendeKante.datenPunktRechts > trenner:
                    return True
                else:
                    #print "r2"
                    return False
        else:
            if ab.x() != 0:
                x = (a.x()+y*N.x()-c.x())/(-ab.x())
                #print "2.1) ab", ab, "y", y, "x", x
            elif ab.y() != 0:
                x = (a.y()+(a.x()-c.x())/(-N.x())*N.y()-c.y() )/(-ab.y())
                #print "2.2) ab =", ab, "y", y, "x", x
            else:
                print "ab.x == 0 oder (-N.y+(N.x*ab.y/ab.x)==0"
                self.setFehler( code = FEHLER_ALLG_RECHENFEHLER )
                return False
            # c liegt irgendwo auf der (ggf. verlaengerten) Gerade ab
            if self.debug == True:
                print "bestimmeObPunktAufDerRichtigenSeiteDerKanteIst(): y == 0 !!", anfangsOderEndpunkt
            if x >= 0 and x <= 1:
                return True
            elif (x > 1 and ab.x() > 0) or (x < 0 and ab.x() < 0):
                if zuPruefendeKante.datenPunktLinks > trenner and zuPruefendeKante.datenPunktRechts > trenner:
                    return True
                else:
                    #print "s1"
                    return False
            elif (x > 1 and ab.x() < 0) or (x < 0 and ab.x() > 0):
                if zuPruefendeKante.datenPunktLinks <= trenner and zuPruefendeKante.datenPunktRechts <= trenner:
                    return True
                else:
                    #print "s2"
                    return False
            else:
                print "verdammter sonderfall!"
            
    
    def bereinigeKantenfeld( self, von, trenner, bis, kantenzug, geschnitteneKanten ):
        mittelGerade_X = (self.punktefeld[trenner].x() + self.punktefeld[trenner+1].x()) / 2
        
        maxX = mittelGerade_X
        minX = mittelGerade_X
        
        # feststellen der max. x-ausdehnung des neuen kantenzuges
        for i in kantenzug:
            akt_kante = self.kantenfeld[i]
            if akt_kante.endpunkt.x() > maxX:
                maxX = akt_kante.endpunkt.x()
            elif akt_kante.endpunkt.x() < minX:
                minX = akt_kante.endpunkt.x()
                
            if akt_kante.anfangspunkt.x() > maxX:
                maxX = akt_kante.anfangspunkt.x()
            elif akt_kante.anfangspunkt.x() < minX:
                minX = akt_kante.anfangspunkt.x()
        
        index = -1
        # wenn die aktuelle kante...
        for i in range( len(self.kantenfeld) ):
            # - nicht sowieso geloescht ist
            if self.kantenfeld[i].ungueltig == True:
                continue
            
            # - nicht ausserhalb des betrachteten Bereichs liegt
            if self.kantenfeld[i].datenPunktRechts < von or self.kantenfeld[i].datenPunktRechts > bis:
                continue
            elif self.kantenfeld[i].datenPunktLinks < von or self.kantenfeld[i].datenPunktLinks > bis:
                continue
                
            # - keine kante des gerade erzeugten kantenzugs ist
            try:
                index = kantenzug.index(i)
            except ValueError: 
                index = -1
            else:
                continue
            
            # - keine von diesem kantenzug geschnittene kante ist
            try:
                index = geschnitteneKanten.index(i)
            except ValueError: 
                index = -1
            else:
                continue
                
            # - die nicht im sicheren bereich (unter- bzw. oberhalb der min./max. x-ausdehung des kantenzuges liegt
            if (self.kantenfeld[i].anfangspunkt.x() < minX or self.kantenfeld[i].endpunkt.x() < minX) and self.kantenfeld[i].datenPunktRechts <= trenner:
                continue
            elif (self.kantenfeld[i].anfangspunkt.x() > maxX or self.kantenfeld[i].endpunkt.x() > maxX) and self.kantenfeld[i].datenPunktRechts > trenner:
                continue
             
            
            # - liegen anfangs- oder endpunkt auf der falschen seite des unsicheren bereiches ist die kante auf jeden fall faellig...
            if self.kantenfeld[i].anfangspunkt.x() > maxX and self.kantenfeld[i].endpunkt.x() > maxX and self.kantenfeld[i].datenPunktRechts <= trenner:
                #print "1) loesche", self.kantenfeld[i].datenPunktRechts, ",", self.kantenfeld[i].datenPunktLinks
                self.kantenfeld[i].ungueltig = True
                continue
            elif self.kantenfeld[i].anfangspunkt.x() < minX and self.kantenfeld[i].endpunkt.x() < minX and self.kantenfeld[i].datenPunktRechts > trenner:
                #print "2) loesche", self.kantenfeld[i].datenPunktRechts, ",", self.kantenfeld[i].datenPunktLinks
                self.kantenfeld[i].ungueltig = True
                continue
            elif self.kantenfeld[i].anfangspunkt.x() < minX and self.kantenfeld[i].endpunkt.x() < minX and self.kantenfeld[i].datenPunktRechts <= trenner:
                # andererseits kann die Kante auch auf jeden fall ok sein...
                continue
            elif self.kantenfeld[i].anfangspunkt.x() > maxX and self.kantenfeld[i].endpunkt.x() > maxX and self.kantenfeld[i].datenPunktRechts > trenner:
                # andererseits kann die Kante auch auf jeden fall ok sein...
                continue
                
            # - ... bleiben noch die kanten im unsicheren bereich
            anfangsPunktOK = self.bestimmeObPunktAufDerRichtigenSeiteDerKanteIst( trenner, self.kantenfeld[i], -1, kantenzug )
            endPunktOK = self.bestimmeObPunktAufDerRichtigenSeiteDerKanteIst( trenner, self.kantenfeld[i], 1, kantenzug )
            
            if anfangsPunktOK==False or endPunktOK==False:
                #print "3) loesche", self.kantenfeld[i]
                #if not (self.kantenfeld[i].datenPunktRechts == 235 and self.kantenfeld[i].datenPunktLinks == 263):
                if endPunktOK==False and self.kantenfeld[i].endpunkt.y() < self.minY:
                    schnittPunkt = self.bestimmeSchnittpunkt( self.kantenfeld[i], self.kantenfeld[kantenzug[ len(kantenzug)-1 ]] )
                    if schnittPunkt != None and schnittPunkt.y() < self.minY:
                        continue

                self.kantenfeld[i].ungueltig = True
        
        counter = 0        
        for i in self.kantenfeld:
            if i.ungueltig == True:
                counter += 1
                self.kantenfeld.remove(i)

        self.geloeschteKanten += counter
            
                    
                

    def begrenzeKante( self, geschnitteneKanteIndex, schneidendeKante, schnittpunkt):
        # a + x*ab + y*N = c
        # wenn richtung der kante == richtung y*N -> begrenze anfang
        # sonst begrenze ende
        geschnitten = self.kantenfeld[ geschnitteneKanteIndex ]
        
        if schneidendeKante.datenPunktLinks == geschnitten.datenPunktLinks or schneidendeKante.datenPunktLinks == geschnitten.datenPunktRechts:
            angrenzenderPunkt = schneidendeKante.datenPunktRechts
        elif schneidendeKante.datenPunktRechts == geschnitten.datenPunktLinks or schneidendeKante.datenPunktRechts == geschnitten.datenPunktRechts:
            angrenzenderPunkt = schneidendeKante.datenPunktLinks
        else:
            print "---> dritter Punkt kann nicht ermittelt werden!!"
            self.setFehler( meldung="Referenzpunkt zur Ermittlung, ob Anfang oder Ende der Kante begrenzt werden muessen konnte nicht gefunden werden" )
            return 0
        
        #print "angrenzenderPunkt:", angrenzenderPunkt, "dl:", geschnitten.datenPunktLinks, "dr:", geschnitten.datenPunktRechts
            
        a = self.punktefeld[ geschnitten.datenPunktLinks ]
        b = self.punktefeld[ geschnitten.datenPunktRechts ]
        c = self.punktefeld[ angrenzenderPunkt ]
        c2 = geschnitten.anfangspunkt
        ab = punkt( [b.x() - a.x(), b.y() - a.y()] )
        N = punkt( [ab.y(), -ab.x()] )

        if abs(a.x()-b.x()) < self.genauigkeit and abs(a.y()-b.y()) < self.genauigkeit:
            print "NULLKANTE:", self.punktefeld[ geschnitten.datenPunktLinks ].index, self.punktefeld[ geschnitten.datenPunktRechts ].index
            return 0
        
        if ab.x() != 0:
            y = (a.y()-c.y()-((a.x()-c.x())*ab.y()/ab.x()))/(-N.y()+(N.x()*ab.y()/ab.x()))
            #x = (a.x+y*N.x-c.x)/(-ab.x)
        elif ab.y() != 0:
            y = (a.x() - c.x())/(-N.x())
        else:
            print "ab.x == 0 oder (-N.y+(N.x*ab.y/ab.x)==0"
            self.setFehler( code = FEHLER_ALLG_RECHENFEHLER )
            return 0
        
        # vektor von der (ggf. verlaengerten) Linie ab nach c
        yN = punkt( [y*N.x(), y*N.y()] )

        # richtungsvektor der geschnittenen kante
        R = punkt([geschnitten.endpunkt.x() - geschnitten.anfangspunkt.x(), geschnitten.endpunkt.y() - geschnitten.anfangspunkt.y()])
        # liegt der angrenzende punkt c in kantenrichtung?
        zuCInKantenRichtung = True
        #print "N", N.x, N.y, ";y", y, "R", R.x, R.y
        if yN.x() != 0:
            if R.x()/yN.x() < 0:
                zuCInKantenRichtung = False
        else:
            if self.debug == True:
                print "begrenzeKante(): yN.x == 0"
            if yN.y() == 0:
                self.setFehler( "Referenzpunkt liegt auf der (verlaengerten) Kante - kann nicht als Referenz genutzt werden.")
                print "---> yN.x =", yN.x(), "und yN.y", yN.y(), "==0"
                return 0
            if R.y()/yN.y() < 0:
                zuCInKantenRichtung = False

        #print "R faktor:", R.x/yN.x, ",", R.y/yN.y

        if zuCInKantenRichtung == False:
            #print "begrenze anfang"
            self.kantenfeld[ geschnitteneKanteIndex ].anfangspunkt = schnittpunkt
            # sicherstellen, dass ende auch wirklich am ende ist
            if self.kantenfeld[ geschnitteneKanteIndex ].infinite_ende == True:
                if self.debug == True:
                    print "begrenzeKante(): verschiebe ende von", self.kantenfeld[ geschnitteneKanteIndex ].datenPunktRechts, self.kantenfeld[ geschnitteneKanteIndex ].datenPunktLinks
                neuerPunkt = punkt( [schnittpunkt.x() + R.x(), schnittpunkt.y() + R.y()] )
                neuerPunkt.normiert = True
                neuerPunkt.multipliziert = True
                self.kantenfeld[ geschnitteneKanteIndex ].endpunkt = neuerPunkt

            self.kantenfeld[ geschnitteneKanteIndex ].infinite_anfang = False
            return -1

        else:
            #print "begrenze ende"
            self.kantenfeld[ geschnitteneKanteIndex ].endpunkt = schnittpunkt
            # sicherstellen, dass anfang auch wirklich am anfang ist
            if self.kantenfeld[ geschnitteneKanteIndex ].infinite_anfang == True:
                if self.debug == True:
                    print "begrenzeKante(): verschiebe anfang von", self.kantenfeld[ geschnitteneKanteIndex ].datenPunktRechts, self.kantenfeld[ geschnitteneKanteIndex ].datenPunktLinks
                neuerPunkt = punkt( [schnittpunkt.x()-R.x(), schnittpunkt.y()-R.y()] )
                neuerPunkt.normiert = True
                neuerPunkt.multipliziert = True
                self.kantenfeld[ geschnitteneKanteIndex ].anfangspunkt = neuerPunkt
            self.kantenfeld[ geschnitteneKanteIndex ].infinite_ende = False
            return 1


        
    def bestimmeObereTangentialPunkte( self, von, trenner, bis ):
        # die normale der tangente auf den tangential punkten muss einen positive y wert haben.
        # --> jeder punkt bei dem ein positiver faktor fuer die normale herauskommt liegt ueber der tangenten
        # und ist der neue ansatzpunkt, bis es keinen solchen punkt mehr gibt
        
        # a + x*ab + y*N = c
        if self.debug == True:
            print "bestimmeObereTangentialPunkte():", von, trenner, bis
        minY_links = -1.
        maxY_links = -1.
        minY_rechts = -1.
        maxY_rechts = -1.

        minY = self.maxY+1
        maxY = self.minY-1

        for i in range( von, trenner+1 ):
                
            if self.punktefeld[i].y() > maxY:
                maxY = self.punktefeld[i].y()
                maxY_links = i
            elif self.punktefeld[i].y() < minY:
                minY = self.punktefeld[i].y()
                minY_links = i
            
        minY = self.maxY+1
        maxY = self.minY-1

        for i in range( trenner+1, bis+1 ):
            if self.punktefeld[i].y() > maxY:
                maxY = self.punktefeld[i].y()
                maxY_rechts = i
            elif self.punktefeld[i].y() < minY:
                minY = self.punktefeld[i].y()
                minY_rechts = i
                
        a_punkt = maxY_links
        b_punkt = maxY_rechts
        
        a_veraerndert = True
        b_veraerndert = True
        
        old_a = -1
        old_b = -1
        
        y_a = 0.
        y_b = 0.
        
        while a_veraerndert == True or b_veraerndert == True:

            if self.debug == True:
                print "tangential:",a_punkt,b_punkt
            a_veraerndert = False
            b_veraerndert = False
            a = self.punktefeld[a_punkt]
            b = self.punktefeld[b_punkt]
            ab = punkt( [b.x() - a.x(), b.y() - a.y()] )
            N = punkt( [-ab.y(), ab.x()] )
            
            old_a = a_punkt
            y_a = 0.
            for i in range( von, trenner+1 ):
                y = 0.
                if i == a_punkt:
                    continue
                c = self.punktefeld[i]
                if ab.x() != 0:
                    y = (a.y()-c.y()-((a.x()-c.x())*ab.y()/ab.x()))/(-N.y()+(N.x()*ab.y()/ab.x()))
                elif ab.y() != 0:
                    if self.debug == True:
                        print "1) ------------------>"
                    y = (a.x() - c.x())/-N.x()
                else:
                    print "ab ist (0,0) ->", a_punkt, "=", b_punkt, ", SCHWERER FEHLER!"
                    self.setFehler(code = FEHLER_NULLVEKTOR)
                    return []
                if y > 0 and y > y_a:
                    y_a = y
                    if self.debug == True:
                        print "y1:", y
                    a_veraerndert = True
                    a_punkt = i
            
            old_b = b_punkt
            y_b = 0.
            for i in range( trenner+1, bis+1 ):
                y = 0.
                if i == b_punkt:
                    continue
                c = self.punktefeld[i]
                if ab.x() != 0:
                    y = (a.y()-c.y()-((a.x()-c.x())*ab.y()/ab.x()))/(-N.y()+(N.x()*ab.y()/ab.x()))
                elif ab.y() != 0:
                    # evtl falsch
                    # y = (a.x-c.x-((a.y-c.y)*ab.x/ab.y))/(-N.x+(N.y*ab.x/ab.y))
                    # neu
                    if self.debug == True:
                        print "2) ------------------>"
                    y = (a.x() - c.x())/-N.x()
                else:
                    print "ab ist (0,0) ->", a_punkt, "=", b_punkt, ", SCHWERER FEHLER!"
                    self.setFehler(code = FEHLER_NULLVEKTOR)
                    return []
                if y > 0 and y > y_b:
                    y_b = y
                    if self.debug == True:
                        print "y2:", y
                    b_veraerndert = True
                    b_punkt = i
                    
            if old_a == a_punkt and old_b == b_punkt:
                if self.debug == True:
                    print "bestimmeObereTangentialPunkte(): punkte sind unveraendert", a_punkt, b_punkt
                break
        
        
        if self.fehlerIstAufgetreten():
            print "bestimmeObereTangentialPunkte(): schwerer Fehler ist aufgetreten!"
        
        punkt_links = a_punkt
        punkt_rechts = b_punkt
        
        return [punkt_links, punkt_rechts]
            
        
        

    def bestimmeABFaktorZuPunkt( self, kante, c ):
        # a + x * ab = c
        #hilfsVektor = punkt( self.punktefeld[kante.datenPunktRechts].x - self.punktefeld[kante.datenPunktLinks].x, self.punktefeld[kante.datenPunktRechts].y - self.punktefeld[kante.datenPunktLinks].y )
        #a = punkt( self.punktefeld[kante.datenPunktLinks].x + hilfsVektor.x/2, self.punktefeld[kante.datenPunktLinks].y + hilfsVektor.y/2 )
        a = kante.anfangspunkt
        b = kante.endpunkt
        ab = punkt( [b.x() - a.x(), b.y() - a.y()] )
        
        if ab.x() != 0:
            x = (c.x()-a.x())/ab.x()
        elif ab.x() == 0:
            x = (c.y()-a.y())/ab.y()
            
        # ergebnis pruefen
        probePunkt = punkt( [a.x() + x*ab.x(), a.y() + x*ab.y()] )
        
        # runden gegen float darstellungsfehler (z.B. 0.1 in dezimal ist nicht genau darstellbar als dual zahl, aehnlich 1/3 in dezimal)
        if abs(probePunkt.x() - c.x()) < self.genauigkeit and abs(probePunkt.y() - c.y()) < self.genauigkeit:
            return x
        else:
            print repr(probePunkt.x()), ",", repr(c.x()), "gleich:", probePunkt.x() == c.x(), ",", repr(probePunkt.y()), ",", repr(c.y()), "gleich:", probePunkt.y() == c.y()
            return "fehler"
        
                
        
    def bestimmeSchnittpunkt( self, kante1, kante2 ):
        # loesen der gleichung
        # ->       ->   ->       ->
        # a  + n * ab = c  + m * cd
        a = kante1.anfangspunkt
        ab = punkt( [kante1.endpunkt.x() - a.x(), kante1.endpunkt.y() - a.y()] )
        c = kante2.anfangspunkt
        cd = punkt( [kante2.endpunkt.x() - c.x(), kante2.endpunkt.y() - c.y()] )
        
        if a.abstandZu(kante1.endpunkt) == 0:
            # or c.abstandZu(kante2.endpunkt)==0:
            self.setFehler( code = FEHLER_NULLVEKTOR )
            print "Nullkante"
            return None
            
        # faktor m errechnen
        m = 0.0
        if ab.x() == 0 and cd.x()!=0:
            m = (a.x()-c.x())/cd.x()
        elif ab.x()==0 and cd.x()==0:
            print "Achtung: sonderfall ab.x == 0 and cd.x == 0, kanten", kante1.datenPunktRechts, kante1.datenPunktLinks, "|", kante2.datenPunktRechts, kante2.datenPunktLinks
            if a.x() != c.x():
                return None
            elif (c.y() >= a.y() and c.y() <= a.y() + ab.y()) or (c.y() <= a.y() and c.y() >= a.y() + ab.y()):
                return punkt( [a.x(), c.y()] )
            elif (a.y() >= c.y() and a.y() <= c.y() + cd.y()) or (a.y() <= c.y() and a.y() >= c.y() + cd.y()):
                return punkt( [a.x(), a.y()] )
            else:
                return None
                 
        elif ab.y()==0 and cd.y() != 0:
            m = (a.y()-c.y())/cd.y()
        elif ab.y()==0 and cd.y()==0:
            # dieser fall sollte nicht vorkommen
            print "falsch!"
            #m = (a.x-c.x+(-a.x+c.x)/ab.x)/(cd.x-cd.x/ab.x)
            m = 0
        elif ab.x()!=0 and (cd.x()*ab.y()/ab.x())!=cd.y():
            # scheint ok zu sein
            m = ( a.y()-c.y()+((c.x()-a.x())*ab.y())/ab.x() )/( cd.y()-(cd.x()*ab.y()/ab.x()) )
        else:
            print "M##>", ab, "|", cd
            #print "    c:", c.x, ",", c.y, "|", "d:", kante2.endpunkt.x, ",", kante2.endpunkt.y
            self.setFehler( code = FEHLER_ALLG_RECHENFEHLER )
            return None
        
        if kante2.infinite_anfang!=True and m <= 0 :
            # kanten schneiden sich nicht zwischen Anfangs- und Endpunkt oder sowieso gar nicht
            #print "M weicht ab!"
            #if self.debug == True:
                #print "M: Kanten schneiden sich nicht"
            return None
        if kante2.infinite_ende!=True and m >= 1 :
            # kanten schneiden sich nicht zwischen Anfangs- und Endpunkt oder sowieso gar nicht
            #print "M weicht ab!"
            #if self.debug == True:
                #print "M2: Kanten schneiden sich nicht"
            return None


        # sicherheitshalber n berechnen und spaeter ergebnisse vergleichen
        n = 0.0
        if ab.x()==0 and ab.y()!=0 and cd.x()!=0:
            n = (a.y()-c.y()-((a.x()-c.x())*cd.y()/cd.x()))/(-ab.y())
        elif ab.x()!=0:
            n = (c.x()+m*cd.x()-a.x())/ab.x()
        elif cd.y()==0 and ab.y() != 0:
            n = (c.y()-a.y())/ab.y()
        else:
            print "N##>", ab, "|", cd
            self.setFehler( code = FEHLER_ALLG_RECHENFEHLER )
            return None

        if kante1.infinite_anfang!=True and n <= 0 :
            # kanten schneiden sich nicht zwischen Anfangs- und Endpunkt oder sowieso gar nicht
            #print "n weicht ab!"
            #if self.debug == True:
                #print "N: kanten schneiden sich nicht"
            return None
        if kante1.infinite_ende!=True and n >= 1 :
            # kanten schneiden sich nicht zwischen Anfangs- und Endpunkt oder sowieso gar nicht
            #print "n weicht ab!"
            #if self.debug == True:
                #print "N2: Kanten schneiden sich nicht"
            return None


        schnittPunkt1 = punkt( [a.x() + n * ab.x(), a.y() + n * ab.y()] )
        #schnittPunkt2 = punkt( c.x + m * cd.x, c.y + m * cd.y )
        
        schnittPunkt1.normiert = True
        schnittPunkt1.multipliziert = True
        return schnittPunkt1


