from utilities import *

from fehler import *
from fehlerhaftesObjekt import *
import Projection

import Thuban.Model.proj

from customLayer import *

# schleifen beschleunigen
try:
    from psyco.classes import __metaclass__
except:
    pass

gaussKruegerProjName = "Gauss-Krueger OS, Zone3"

from Thuban.Model.data import SHAPETYPE_POINT

projmap={"keine":None,
         "LongLat WGS84":"+proj=latlong +datum=WGS84",
         #gaussKruegerProjName:"+proj=tmerc +ellps=bessel +lat_0=0 +lon_0=9 +k=1.000000 +x_0=3500000 +y_0=0 +towgs84=591.28,81.35,396.39,1.477,-0.0736,-1.458,9.82 +units=m no_defs"}
         gaussKruegerProjName:"+init=epsg:31463 +towgs84=606,23,413"}

#+towgs84=-586,-87,-409,0.52,0.15,-2.82,0.999991
#+towgs84=591.28,81.35,396.39,1.477,-0.0736,-1.458,9.82
#+towgs84=-591.28,-81.35,-396.39,1.477,-0.0736
# dx= - 586 m, dy = - 87 m, dz = -409 m, rx = 0.52 ", ry = 0.15", rz = - 2.82" ,Mastab: = 0.999991)

class ebenenProjezierer(potentiellFehlerhaftesObjekt):

    def __init__(self, projIn, layerIn, projOut=None, filterAbst=-1):
        potentiellFehlerhaftesObjekt.__init__( self )
        self.in_proj=projIn
        
        global gaussKruegerProjName
        global projmap
        
        if ( projOut == None ):
            projOut = projmap[gaussKruegerProjName]
            
        self.out_proj=projOut
        
        if (self.in_proj == projmap["LongLat WGS84"]):
            self.in_proj=None
        
        self.in_layer = layerIn
        self.out_layer = None
        
        self.debug = True
        
        self.filter = True
        if ( filterAbst < 0 ):
            self.filter = False
        
        self.stdAbstand = filterAbst

    def addText(self,text):
        if ( self.fehler == None ):
            self.fehler = fehler(text)
        else:
            self.fehler.meldung = self.fehler.meldung + "\n" + text
            
    def getStdAbstand( self ):
        return self.stdAbstand
        
    def _getProjStrings( self, projName ):
        if ( projName == None ):
            return None
        
        projString = projmap[projName]
        paramStrings = projString.split()
        
        for i_strng in range( len(paramStrings)):
            paramStrings[i_strng] = paramStrings[i_strng].replace("+","")
            
        return paramStrings
        
    def getOutProjStrings( self ):
        if ( self.out_proj == None ):
            return None
        
        paramStrings = self.out_proj.split()
        
        for i_strng in range( len(paramStrings)):
            paramStrings[i_strng] = paramStrings[i_strng].replace("+","")
            
        return paramStrings
        
    def getOutProj( self ):
        strngs = self.getOutProjStrings()
        if ( strngs != None ):
            return Thuban.Model.proj.Projection( strngs, "Gauss-Krueger, Zone 3" )
        else:
            return None

    
    def erstelleProjeziertesLayer(self):
        
        punktekontrollfeld = []
        kontrolleIstBesser = self.filter
        
        newStore = MemoryShapeStore.getMemCopyOfShapeStore( self.in_layer.ShapeStore() )

        if ( newStore.ShapeType() == SHAPETYPE_POINT ):
            kontrolleIstBesser = True

        if self.in_proj != None:
            in_proj = Projection.Projection(self.in_proj.split())
            if (self.debug):
                print ("IN:Proj %s"%self.in_proj)
        elif (self.debug):
            print "Eingangsprojektion: "
        if self.out_proj != None:
            out_proj = Projection.Projection(self.out_proj.split())
            if (self.debug):
                print ("Out:Proj %s"%self.out_proj)
        else:
            self.addText("Keine Ausgangsprojektion")

        toBeRemoved = []
        auslassungen = {}

        for shape in newStore.AllShapes():
            points = shape.Points()[0]
            newPoints = []
            
            for point in points:
                wirdEingefuegt = True
                x,y = point
                
                if ( self.debug and len(punktekontrollfeld)==0):
                    print "---------------------"
                    print "vorher: ", point
                    
                if self.in_proj != None:
                    normpoint = in_proj.Inverse( x, y )
                else:
                    normpoint = point
                
                normx,normy=normpoint
                
                if (self.debug and len(punktekontrollfeld)==0):
                    print "mittendrin: ", normpoint
                
                if self.out_proj != None:
                    outpoint = out_proj.Forward( normx, normy )
                else:
                    outpoint = normpoint
                
                if (self.debug and len(punktekontrollfeld)==0):
                    print "nachher: ", outpoint
                    
                if ( kontrolleIstBesser ):
                    neuerPunkt = punkt( outpoint, shape.ShapeID() )
                    
                    for einPunkt in punktekontrollfeld:
                        abst = neuerPunkt.abstandZu( einPunkt )
                        if ( abst <= self.stdAbstand ):
                            wirdEingefuegt = False
                            if (self.debug and False):
                                print "Abstand:", abst
                                
                            if ( auslassungen.has_key(abst) ):
                                auslassungen[abst] += 1
                            else:
                                auslassungen[abst] = 1
                            break
                        
                    if (wirdEingefuegt):
                        punktekontrollfeld.append(neuerPunkt)

                if (wirdEingefuegt):
                    newPoints.append( tuple(outpoint) )
                else:
                    if (self.debug and False):
                        print "punkt wird ausgelassen!"
                    
            
            if ( len( newPoints ) != 0 ):
                shape.SetPoints( [newPoints] )
            else:
                toBeRemoved.append( shape)
                
        toBeRemoved.reverse()   
        for shape in toBeRemoved:
            newStore.RemoveShape( shape )
            # zweimal die indexe aufforsten macht auch quatsch!!
#            for checked in newStore.AllShapes():
#                if ( checked.ShapeID() > shape.ShapeID() ):
#                    checked.SetShapeID( checked.ShapeID()-1 )
                    
        print "-", len(toBeRemoved),"Punkte ausgelassen:"
        
        print "  Abstand,\t\t Anzahl d. Auslassung"
        for k in auslassungen.keys():
            print "    ", repr(k), ",\t\t", repr(auslassungen[k])
            
        newStore.calculateBbox()
        self.out_layer = customLayer( self.in_layer.Title(), newStore, typ = typenMap.getProjezierterTyp(), visible = self.in_layer.Visible())
                 
        self.out_layer.SetClassificationColumn( self.in_layer.GetClassificationColumn() )
        self.out_layer.SetClassification( self.in_layer.GetClassification() )
        
        # dann rechnet der irgendwas, was er nicht soll...
        #self.out_layer.SetProjection( out_proj )
        # darum...
        #self.out_layer.projection = self.getOutProj()
        # ist aber auch doof, weil alles abkackt, wenn man dann ein shapefile dazu packt...
            
        return self.out_layer


                
            
            
